﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.ComponentModel;
using TSLab.DataSource;
using TSLab.Script;
using TSLab.Script.Handlers;

//namespace Indicators.TSLab.Blocks.FiRt {
namespace ZIPRT{
  class Consts {
    public const string Category = "ZIP.FinInfo.jhgjrht.RT";
  }

  public class FinInfoRtValue<T> where T : struct {
    readonly PropertyDescriptor _pd;

    public FinInfoRtValue() {}
    public FinInfoRtValue(string property) { _pd = TypeDescriptor.GetProperties(typeof(FinInfo))[property]; }

    public IList<T> Execute(ISecurity security) {
      int count = security.Bars.Count;
      if (count == 0)
        return new T[0];
      
      var o = _pd.GetValue(security.FinInfo);
      if (o is T?) {
        var a = (T?) o;
        return new ConstantList<T>(count, a.HasValue ? a.Value : default(T));
      }
      if (o is T)
        return new ConstantList<T>(count, (T) o);
      return new T[count];
    }
  }

  /// <summary>
  /// Цена предложения
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FIR Цена предложения")]
  public class FinInfoRtAsk : FinInfoRtValue<double>, IBar2DoubleHandler {
    public FinInfoRtAsk() : base("Ask") {}
  }

  /// <summary>
  /// Средняя цена за сессию
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FIR Средняя цена")]
  public class FinInfoRtAvgPrice : FinInfoRtValue<double>, IBar2DoubleHandler {
    public FinInfoRtAvgPrice() : base("AvgPrice") { }
  }

  /// <summary>
  /// Цена спроса
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FIR Цена спроса")]
  public class FinInfoRtBid : FinInfoRtValue<double>, IBar2DoubleHandler {
    public FinInfoRtBid() : base("Bid") { }
  }

  /// <summary>
  /// Количество заявок на покупку в очереди торговой системы
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FIR Заявок на покупку")]
  public class FinInfoRtBuyCount : FinInfoRtValue<double>, IBar2DoubleHandler {
    public FinInfoRtBuyCount() : base("BuyCount") { }
  }

  /// <summary>
  /// Гарантийные обязательства покупателя, в рублях
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FIR ГО покупателя")]
  public class FinInfoRtBuyDeposit : FinInfoRtValue<double>, IBar2DoubleHandler {
    public FinInfoRtBuyDeposit() : base("BuyDeposit") { }
  }

  /// <summary>
  /// Объем всех заявок на покупку в очереди торговой системы, выраженный в единицах ценных бумаг
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FIR Объем заявок на покупку")]
  public class FinInfoRtBuySqty : FinInfoRtValue<double>, IBar2DoubleHandler {
    public FinInfoRtBuySqty() : base("BuySqty") { }
  }

  /// <summary>
  /// Изменение (абсолютное)
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FIR Изменение цены")]
  public class FinInfoRtChange : FinInfoRtValue<double>, IBar2DoubleHandler {
    public FinInfoRtChange() : base("Change") { }
  }

  /// <summary>
  /// Изменение (относительное)
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FIR Изменение цены %")]
  public class FinInfoRtChangePercent : FinInfoRtValue<double>, IBar2DoubleHandler {
    public FinInfoRtChangePercent() : base("ChangePercent") { }
  }

  /// <summary>
  /// Цена закрытия сессии (прошлой)
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FIR Цена закрытия")]
  public class FinInfoRtClosePrice : FinInfoRtValue<double>, IBar2DoubleHandler {
    public FinInfoRtClosePrice() : base("ClosePrice") { }
  }

  /// <summary>
  /// Наибольшая цена спроса в течение торговой сессии
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FIR Макс. цена спроса")]
  public class FinInfoRtHighBid : FinInfoRtValue<double>, IBar2DoubleHandler {
    public FinInfoRtHighBid() : base("HighBid") { }
  }

  /// <summary>
  /// Торги по инструменту разрешены или нет
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FIR Торги разрешены")]
  public class FinInfoRtReaderIsTradingEnabled : FinInfoRtValue<bool>, IBar2BoolsHandler {
    public FinInfoRtReaderIsTradingEnabled() : base("IsTradingEnabled") { }
  }

  /// <summary>
  /// Цена последней сделки
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FIR Цена последней")]
  public class FinInfoRtLastPrice : FinInfoRtValue<double>, IBar2DoubleHandler {
    public FinInfoRtLastPrice() : base("LastPrice") { }
  }

  /// <summary>
  /// Время обновления данных в миллисекундах
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FIR Время обновления")]
  public class FinInfoRtLastUpdate : IBar2DoubleHandler {
    public IList<double> Execute(ISecurity security) {
      int count = security.Bars.Count;
      if (count == 0)
        return new double[0];

      //var v = security.FinInfo.LastUpdate.ToOADate();
      var v = new TimeSpan(security.FinInfo.LastUpdate.Ticks).TotalMilliseconds;
      //var v = security.FinInfo.LastUpdate.Ticks;

      return new ConstantList<double>(count, v);
    }
  }

  /// <summary>
  /// Время последней сделки в миллисекундах
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FIR Время последней сделки")]
  public class FinInfoRtLastTradeUpdate : IBar2DoubleHandler {
    public IList<double> Execute(ISecurity security) {
      int count = security.Bars.Count;
      if (count == 0)
        return new double[0];

      if (!security.FinInfo.LastTradeUpdate.HasValue)
        return new ConstantList<double>(count, 0);

      //var v = security.FinInfo.LastTradeUpdate.Value.ToOADate();
      var v = new TimeSpan(security.FinInfo.LastTradeUpdate.Value.Ticks).TotalMilliseconds;
      //var v = security.FinInfo.LastTradeUpdate.Value.Ticks;

      return new ConstantList<double>(count, v);
    }
  }

  /// <summary>
  /// Наименьшая цена предложения в течение торговой сессии
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FIR Мин. цена предложения")]
  public class FinInfoRtLowOffer : FinInfoRtValue<double>, IBar2DoubleHandler {
    public FinInfoRtLowOffer() : base("LowOffer") { }
  }

  /// <summary>
  /// Максимальная цена сделки за сессию
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FIR Макс. цена сделки")]
  public class FinInfoRtMaxDeal : FinInfoRtValue<double>, IBar2DoubleHandler {
    public FinInfoRtMaxDeal() : base("MaxDeal") { }
  }

  /// <summary>
  /// Минимальная цена сделки за сессию
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FIR Мин. цена сделки")]
  public class FinInfoRtMinDeal : FinInfoRtValue<double>, IBar2DoubleHandler {
    public FinInfoRtMinDeal() : base("MinDeal") { }
  }

  /// <summary>
  /// Количество сделок за торговый день
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FIR Сделок")]
  public class FinInfoRtNoDeal : FinInfoRtValue<double>, IBar2DoubleHandler {
    public FinInfoRtNoDeal() : base("NoDeal") { }
  }

  /// <summary>
  /// Открытый интерес
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FIR Открытый интерес")]
  public class FinInfoRtOpenInterest : FinInfoRtValue<double>, IBar2DoubleHandler {
    public FinInfoRtOpenInterest() : base("OpenInterest") { }
  }

  /// <summary>
  /// Цена открытия сессии
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FIR Цена открытия")]
  public class FinInfoRtOpenPrice : FinInfoRtValue<double>, IBar2DoubleHandler {
    public FinInfoRtOpenPrice() : base("OpenPrice") { }
  }

  /// <summary>
  /// Цена последней сделки предыдущего торгового дня
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FIR Цена посл. сделки пред. торг. дня")]
  public class FinInfoRtPrevPrice : FinInfoRtValue<double>, IBar2DoubleHandler {
    public FinInfoRtPrevPrice() : base("PrevPrice") { }
  }

  /// <summary>
  /// Цена последней сделки к оценке предыдущего дня
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FIR Цена посл. сделки к оценке пред. дня")]
  public class FinInfoRtPrice2PrevAvg : FinInfoRtValue<double>, IBar2DoubleHandler {
    public FinInfoRtPrice2PrevAvg() : base("Price2PrevAvg") { }
  }

  /// <summary>
  /// Верхний лимит цены
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FIR Верхний лимит цены")]
  public class FinInfoRtPriceMax : FinInfoRtValue<double>, IBar2DoubleHandler {
    public FinInfoRtPriceMax() : base("PriceMax") { }
  }

  /// <summary>
  /// Нижний лимит цены
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FIR Нижний лимит цены")]
  public class FinInfoRtPriceMin : FinInfoRtValue<double>, IBar2DoubleHandler {
    public FinInfoRtPriceMin() : base("PriceMin") { }
  }

  /// <summary>
  /// Количество заявок на продажу в очереди торговой системы
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FIR Заявок на продажу")]
  public class FinInfoRtSellCount : FinInfoRtValue<double>, IBar2DoubleHandler {
    public FinInfoRtSellCount() : base("SellCount") { }
  }

  /// <summary>
  /// Гарантийные обязательства продавца, в рублях
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FIR ГО продавца")]
  public class FinInfoRtSellDeposit : FinInfoRtValue<double>, IBar2DoubleHandler {
    public FinInfoRtSellDeposit() : base("SellDeposit") { }
  }

  /// <summary>
  /// Объем всех заявок на продажу в очереди торговой системы, выраженный в единицах ценных бумаг
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FIR Объем заявок на продажу")]
  public class FinInfoRtSellSqty : FinInfoRtValue<double>, IBar2DoubleHandler {
    public FinInfoRtSellSqty() : base("SellSqty") { }
  }

  /// <summary>
  /// Стоимость шага цены
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FIR Стоимость шага цены")]
  public class FinInfoRtStepPrice : FinInfoRtValue<double>, IBar2DoubleHandler {
    public FinInfoRtStepPrice() : base("StepPrice") { }
  }

  /// <summary>
  /// Объем совершенных сделок, выраженный в единицах ценных бумаг
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FIR Объем сделок")]
  public class FinInfoRtSumQty : FinInfoRtValue<double>, IBar2DoubleHandler {
    public FinInfoRtSumQty() : base("SumQty") { }
  }

  /// <summary>
  /// Объем совершенных сделок, выраженный в рублях.
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FIR Объем сделок в рублях")]
  public class FinInfoRtSumVolume : FinInfoRtValue<double>, IBar2DoubleHandler {
    public FinInfoRtSumVolume() : base("SumVolume") { }
  }

  /// <summary>
  /// Теоретическая (Расчетная) цена опциона
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FIR Теоретическая цена")]
  public class FinInfoRtTheoreticPrice : FinInfoRtValue<double>, IBar2DoubleHandler {
    public FinInfoRtTheoreticPrice() : base("TheoriticPrice") { }
  }

  /// <summary>
  /// Волатильность финансового инструмента, %
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FIR Волатильность %")]
  public class FinInfoRtVolatility : FinInfoRtValue<double>, IBar2DoubleHandler {
    public FinInfoRtVolatility() : base("Volatility") { }
  }

  /// <summary>
  /// Разница между количеством заявок на покупку и на продажу в очереди торговой системы.
  /// Положительное значение означает, что заявок на покупку больше, чем заявок на продажу 
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FIR Дельта заявок")]
  public class FinInfoRtDeltaCount : IBar2DoubleHandler {
    public IList<double> Execute(ISecurity security) {
      int count = security.Bars.Count;
      if (count == 0)
        return new double[0];
      var fi = security.FinInfo;
      return new ConstantList<double>(count, (fi.BuyCount ?? 0) - (fi.SellCount ?? 0));
    }
  }

  // Истории нет, поэтому посчитать момент нельзя.
  //[HandlerName("FIR Момент дельты заявок")]
  //public class FinInfoRtDeltaCountMoment : IBar2DoubleHandler, IValuesHandlerWithNumber {}

  /// <summary>
  /// Разница между объемом всех заявок на покупку и продажу в очереди торговой системы, выраженная в единицах ценных бумаг
  /// Положительное значение означает, что объем на покупку больше, чем объем на продажу 
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FIR Дельта объема заявок")]
  public class FinInfoRtDeltaSqty : IBar2DoubleHandler {
    public IList<double> Execute(ISecurity security) {
      int count = security.Bars.Count;
      if (count == 0)
        return new double[0];
      var fi = security.FinInfo;
      return new ConstantList<double>(count, (fi.BuySqty ?? 0) - (fi.SellSqty ?? 0));
    }
  }

  // Истории нет, поэтому посчитать момент нельзя.
  //[HandlerName("FIR Момент дельты объема заявок")]
  //public class FinInfoRtDeltaSqtyMoment : IBar2DoubleHandler, IValuesHandlerWithNumber {}

  /// <summary>
  /// Размер средней заявки на покупку выраженный в единицах ценных бумаг
  /// ИстСрПокупка = ИстСумПок / ИстЗаявПок
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FIR Средняя заявка на покупку")]
  public class FinInfoRtAvgBuy : IBar2DoubleHandler {
    public IList<double> Execute(ISecurity security) {
      int count = security.Bars.Count;
      if (count == 0)
        return new double[0];
      var fi = security.FinInfo;
      var v = fi.BuyCount ?? 0;
      if (v != 0)
        v = (fi.BuySqty ?? 0)/v;
      return new ConstantList<double>(count, v);
    }
  }

  /// <summary>
  /// Размер средней заявки на продажу выраженный в единицах ценных бумаг
  /// ИстСрПродажа = ИстСумПрод / ИстЗаявПрод
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FIR Средняя заявка на продажу")]
  public class FinInfoRtAvgSell : IBar2DoubleHandler {
    public IList<double> Execute(ISecurity security) {
      int count = security.Bars.Count;
      if (count == 0)
        return new double[0];
      var fi = security.FinInfo;
      var v = fi.SellCount ?? 0;
      if (v != 0)
        v = (fi.SellSqty ?? 0)/v;
      return new ConstantList<double>(count, v);
    }
  }

  /// <summary>
  ///   Список нужной длины из одного значения
  /// </summary>
  class ConstantList<T> : IList<T> {
    T _value;
    readonly int _count;

    public ConstantList() { }

    public ConstantList(int count, T v) {
      _value = v;
      _count = count;
    }

    public void Add(T item) { throw new NotImplementedException(); }
    public void Clear() { throw new NotImplementedException(); }
    public bool Contains(T item) { throw new NotImplementedException(); }
    public void CopyTo(T[] array, int arrayIndex) { throw new NotImplementedException(); }
    public bool Remove(T item) { throw new NotImplementedException(); }
    public int IndexOf(T item) { throw new NotImplementedException(); }
    public void Insert(int index, T item) { throw new NotImplementedException(); }
    public void RemoveAt(int index) { throw new NotImplementedException(); }

    public int Count { get { return _count; } }
    public bool IsReadOnly { get { return true; } }
    public T this[int index] { get { return _value; } set { _value = value; } }
    public IEnumerator<T> GetEnumerator() { return new Enumerator<T>(_count, _value); }
    IEnumerator IEnumerable.GetEnumerator() { return GetEnumerator(); }

    public class Enumerator<T1> : IEnumerator<T1> {
      int _i;
      readonly int _count;
      readonly T1 _v;

      public Enumerator(int count, T1 v) {
        _count = count;
        _v = v;
      }
      public void Dispose() { }
      public bool MoveNext() { return ++_i <= _count; }
      public void Reset() { _i = 0; }
      public T1 Current { get { return _v; } }
      object IEnumerator.Current { get { return _v; } }
    }
  }

}