﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.IO;
using TSLab.DataSource;
using TSLab.Script;
using TSLab.Script.Handlers;

//[assembly: System.Reflection.AssemblyVersion("2.0.0.0")]
[assembly: System.Reflection.AssemblyFileVersion("2.0.0.0")]

//namespace Indicators.TSLab.Blocks.Fi {
namespace ZIP {
  class Consts {
    public const string Category = "ZIP.FinInfo.jhgjrht";
    public const string DefPath = @"C:\\Data";
    public const string PathName = "Папка";
    public const string FICacheKey = "FICacheKey";
  }

  public class FinInfoValueParams {
    public IContext Context { get; set; }

    [HandlerParameter(true, Consts.DefPath, Name = Consts.PathName, NotOptimized = true)]
    public string DataPath { get; set; }

    //[HandlerParameter(true, "false", Name = "Записывать лог", NotOptimized = true)]
    public bool Logging { get; set; }
  }

  /// <summary>
  /// Блок - накопитель истории 
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FI Накопитель"), OutputsCount(0)]
  public class FinInfoWriter : FinInfoValueParams, IOneSourceHandler, IStreamHandler, ISecurityInputs {
    //[HandlerParameter(true, "false", Name = "Разрешить повторы", NotOptimized = true)]
    public bool All { get; set; }

    public void Execute(ISecurity security) {
      FinInfoCache.WriteFininfo(security, DataPath, All, Logging);
    }
  }

  public class FinInfoReaderValue : FinInfoValueParams {
    readonly PropertyDescriptor _pd;
    
    public FinInfoReaderValue(string property) {
      _pd = TypeDescriptor.GetProperties(typeof(FinInfoBlock))[property];
    }

    public IList<double> Execute(ISecurity security) {
      var count = security.Bars.Count;
      if (count == 0)
        return new double[0];

      //var cache = Context.LoadObject(Consts.FICacheKey) as FinInfoCache;
      var cache = Context.LoadGlobalObject(Consts.FICacheKey) as FinInfoCache;
      if (cache == null)
        Context.StoreGlobalObject(Consts.FICacheKey, cache = new FinInfoCache());

      var values = new double[count];
      cache.GetFininfoBlocks(security, Context, DataPath, Logging, (iBar, fb) => {
        var o = _pd.GetValue(fb);
        if (o is float)
          values[iBar] = (float) o;
      });
      return values;
    }
  }

  /// <summary>
  /// Цена предложения
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FI Цена предложения")]
  public class FinInfoReaderAsk : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderAsk() : base("Ask") {}
  }

  /// <summary>
  /// Средняя цена за сессию
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FI Средняя цена")]
  public class FinInfoReaderAvgPrice : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderAvgPrice() : base("AvgPrice") { }
  }

  /// <summary>
  /// Цена спроса
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FI Цена спроса")]
  public class FinInfoReaderBid : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderBid() : base("Bid") { }
  }

  /// <summary>
  /// Количество заявок на покупку в очереди торговой системы
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FI Заявок на покупку")]
  public class FinInfoReaderBuyCount : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderBuyCount() : base("BuyCount") { }
  }

  /// <summary>
  /// Гарантийные обязательства покупателя, в рублях
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FI ГО покупателя")]
  public class FinInfoReaderBuyDeposit : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderBuyDeposit() : base("BuyDeposit") { }
  }

  /// <summary>
  /// Объем всех заявок на покупку в очереди торговой системы, выраженный в единицах ценных бумаг
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FI Объем заявок на покупку")]
  public class FinInfoReaderBuySqty : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderBuySqty() : base("BuySqty") { }
  }

  /// <summary>
  /// Изменение (абсолютное)
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FI Изменение цены")]
  public class FinInfoReaderChange : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderChange() : base("Change") { }
  }

  /// <summary>
  /// Изменение (относительное)
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FI Изменение цены %")]
  public class FinInfoReaderChangePercent : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderChangePercent() : base("ChangePercent") { }
  }

  /// <summary>
  /// Цена закрытия сессии (прошлой)
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FI Цена закрытия")]
  public class FinInfoReaderClosePrice : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderClosePrice() : base("ClosePrice") { }
  }

  /// <summary>
  /// Наибольшая цена спроса в течение торговой сессии
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FI Макс. цена спроса")]
  public class FinInfoReaderHighBid : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderHighBid() : base("HighBid") { }
  }

  /// <summary>
  /// Торги по инструменту разрешены или нет
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FI Торги разрешены")]
  public class FinInfoReaderIsTradingEnabled : FinInfoValueParams, IBar2BoolsHandler, IContextUses {
    public IList<bool> Execute(ISecurity security) {
      var cache = Context.LoadGlobalObject(Consts.FICacheKey) as FinInfoCache;
      if (cache == null)
        Context.StoreGlobalObject(Consts.FICacheKey, cache = new FinInfoCache());

      var values = new bool[security.Bars.Count];
      cache.GetFininfoBlocks(security, Context, DataPath, Logging, (bar, fb) => { 
        values[bar] = fb.IsTradingEnabled;
      });
      return values;
    }
  }

  /// <summary>
  /// Цена последней сделки
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FI Цена последней")]
  public class FinInfoReaderLastPrice : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderLastPrice() : base("LastPrice") { }
  }

  /// <summary>
  /// Время обновления данных
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FI Время обновления")]
  public class FinInfoReaderLastUpdate : FinInfoValueParams, IBar2DoubleHandler, IContextUses {
    public IList<double> Execute(ISecurity security) {
      var cache = Context.LoadGlobalObject(Consts.FICacheKey) as FinInfoCache;
      if (cache == null)
        Context.StoreGlobalObject(Consts.FICacheKey, cache = new FinInfoCache());

      var values = new double[security.Bars.Count];
      cache.GetFininfoBlocks(security, Context, DataPath, Logging, (bar, fb) => {
        //values[bar] = fb.LastUpdate.ToOADate();
        values[bar] = new TimeSpan(fb.LastUpdate.Ticks).TotalSeconds;
        //values[bar] = fb.LastUpdate.Ticks;
      });
      return values;
    }
  }

  /// <summary>
  /// Время обновления данных локальное
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FI Время обновления локальное")]
  public class FinInfoReaderLocalDate : FinInfoValueParams, IBar2DoubleHandler, IContextUses {
    public IList<double> Execute(ISecurity security) {
      var cache = Context.LoadGlobalObject(Consts.FICacheKey) as FinInfoCache;
      if (cache == null)
        Context.StoreGlobalObject(Consts.FICacheKey, cache = new FinInfoCache());

      var values = new double[security.Bars.Count];
      cache.GetFininfoBlocks(security, Context, DataPath, Logging, (bar, fb) => {
        //values[bar] = fb.LocalDate.ToOADate();
        values[bar] = new TimeSpan(fb.LocalDate.Ticks).TotalSeconds;
        //values[bar] = fb.LocalDate.Ticks;
      });
      return values;
    }
  }
  
  /// <summary>
  ///   Время последней сделки
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FI Время последней сделки")]
  public class FinInfoReaderLastTradeUpdate : FinInfoValueParams, IBar2DoubleHandler, IContextUses {
    public IList<double> Execute(ISecurity security) {
      var cache = Context.LoadGlobalObject(Consts.FICacheKey) as FinInfoCache;
      if (cache == null)
        Context.StoreGlobalObject(Consts.FICacheKey, cache = new FinInfoCache());

      var values = new double[security.Bars.Count];
      cache.GetFininfoBlocks(security, Context, DataPath, Logging, (bar, fb) => {
        //values[bar] = fb.LocalDate.ToOADate();
        values[bar] = new TimeSpan(fb.LastTradeUpdate.Ticks).TotalSeconds;
        //values[bar] = fb.LocalDate.Ticks;
      });
      return values;
    }
  }

  /// <summary>
  /// Наименьшая цена предложения в течение торговой сессии
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FI Мин. цена предложения")]
  public class FinInfoReaderLowOffer : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderLowOffer() : base("LowOffer") { }
  }

  /// <summary>
  /// Максимальная цена сделки за сессию
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FI Макс. цена сделки")]
  public class FinInfoReaderMaxDeal : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderMaxDeal() : base("MaxDeal") { }
  }

  /// <summary>
  /// Минимальная цена сделки за сессию
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FI Мин. цена сделки")]
  public class FinInfoReaderMinDeal : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderMinDeal() : base("MinDeal") { }
  }

  /// <summary>
  /// Количество сделок за торговый день
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FI Сделок")]
  public class FinInfoReaderNoDeal : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderNoDeal() : base("NoDeal") { }
  }

  /// <summary>
  ///   Открытый интерес
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FI Открытый интерес")]
  public class FinInfoReaderOpenInterest : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderOpenInterest() : base("OpenInterest") { }
  }

  /// <summary>
  /// Цена открытия сессии
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FI Цена открытия")]
  public class FinInfoReaderOpenPrice : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderOpenPrice() : base("OpenPrice") { }
  }

  /// <summary>
  /// Цена последней сделки предыдущего торгового дня
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FI Цена посл. сделки пред. торг. дня")]
  public class FinInfoReaderPrevPrice : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderPrevPrice() : base("PrevPrice") { }
  }

  /// <summary>
  /// Цена последней сделки к оценке предыдущего дня
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FI Цена посл. сделки к оценке пред. дня")]
  public class FinInfoReaderPrice2PrevAvg : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderPrice2PrevAvg() : base("Price2PrevAvg") { }
  }

  /// <summary>
  /// Верхний лимит цены
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FI Верхний лимит цены")]
  public class FinInfoReaderPriceMax : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderPriceMax() : base("PriceMax") { }
  }

  /// <summary>
  /// Нижний лимит цены
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FI Нижний лимит цены")]
  public class FinInfoReaderPriceMin : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderPriceMin() : base("PriceMin") { }
  }

  /// <summary>
  /// Количество заявок на продажу в очереди торговой системы
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FI Заявок на продажу")]
  public class FinInfoReaderSellCount : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSellCount() : base("SellCount") { }
  }

  /// <summary>
  /// Гарантийные обязательства продавца, в рублях
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FI ГО продавца")]
  public class FinInfoReaderSellDeposit : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSellDeposit() : base("SellDeposit") { }
  }

  /// <summary>
  /// Объем всех заявок на продажу в очереди торговой системы, выраженный в единицах ценных бумаг
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FI Объем заявок на продажу")]
  public class FinInfoReaderSellSqty : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSellSqty() : base("SellSqty") { }
  }

  /// <summary>
  /// Стоимость шага цены
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FI Стоимость шага цены")]
  public class FinInfoReaderStepPrice : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderStepPrice() : base("StepPrice") { }
  }

  /// <summary>
  /// Объем совершенных сделок, выраженный в единицах ценных бумаг
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FI Объем сделок")]
  public class FinInfoReaderSumQty : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSumQty() : base("SumQty") { }
  }

  /// <summary>
  /// Объем совершенных сделок, выраженный в рублях.
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FI Объем сделок в рублях")]
  public class FinInfoReaderSumVolume : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSumVolume() : base("SumVolume") { }
  }

  /// <summary>
  /// Теоретическая (Расчетная) цена опциона
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FI Теоретическая цена")]
  public class FinInfoReaderTheoreticPrice : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderTheoreticPrice() : base("TheoreticPrice") { }
  }

  /// <summary>
  /// Волатильность финансового инструмента, %
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FI Волатильность %")]
  public class FinInfoReaderVolatility : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderVolatility() : base("Volatility") { }
  }

  /// <summary>
  /// Разница между количеством заявок на покупку и на продажу в очереди торговой системы.
  /// Положительное значение означает, что заявок на покупку больше, чем заявок на продажу 
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FI Дельта заявок")]
  public class FinInfoReaderDeltaCount : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderDeltaCount() : base("DeltaCount") { }
  }

  /// <summary>
  /// Моментум разницы между количеством всех заявок на покупку и на продажу в очереди торговой системы.
  /// Положительное значение означает, что количество заявок на покупку растет
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FI Момент дельты заявок")]
  public class FinInfoReaderDeltaCountMoment : FinInfoValueParams, IBar2DoubleHandler, IContextUses {
    [HandlerParameter(true, "1", Min = "1", Max = "10", Name = "Период")]
    public int Period { get; set; }

    public IList<double> Execute(ISecurity security) {
      var cache = Context.LoadGlobalObject(Consts.FICacheKey) as FinInfoCache;
      if (cache == null)
        Context.StoreGlobalObject(Consts.FICacheKey, cache = new FinInfoCache());

      int count = security.Bars.Count;
      var values = new double[count];
      cache.GetFininfoBlocks(security, Context, DataPath, Logging, (bar, fb) => {
        values[bar] = fb.BuyCount - fb.SellCount;
      });
      int p = Period;
      for (int i = count-1; i >= 0; i--)
        if (i >= p)
          values[i] -= values[i - p];
      return values;
    }
  }

  /// <summary>
  /// Разница между объемом всех заявок на покупку и продажу в очереди торговой системы, выраженная в единицах ценных бумаг
  /// Положительное значение означает, что объем на покупку больше, чем объем на продажу 
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FI Дельта объема заявок")]
  public class FinInfoReaderDeltaSqty : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderDeltaSqty() : base("DeltaSqty") { }
  }

  /// <summary>
  /// Моментум разницы между объемом всех заявок на покупку и продажу в очереди торговой системы.
  /// Положительное значение означает, что объем заявок на покупку растет
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FI Момент дельты объема заявок")]
  public class FinInfoReaderDeltaSqtyMoment : FinInfoValueParams, IBar2DoubleHandler, IContextUses {
    [HandlerParameter(true, "1", Min = "1", Max = "10", Name = "Период")]
    public int Period { get; set; }

    public IList<double> Execute(ISecurity security) {
      var cache = Context.LoadGlobalObject(Consts.FICacheKey) as FinInfoCache;
      if (cache == null)
        Context.StoreGlobalObject(Consts.FICacheKey, cache = new FinInfoCache());

      int count = security.Bars.Count;
      var values = new double[count];
      cache.GetFininfoBlocks(security, Context, DataPath, Logging, (bar, fb) => {
        values[bar] = fb.BuySqty - fb.SellSqty;
      });
      int p = Period;
      for (int i = count-1; i >= 0; i--)
        if (i >= p)
          values[i] -= values[i - p];
      return values;
    }
  }

  /// <summary>
  /// Размер средней заявки на покупку выраженный в единицах ценных бумаг
  /// ИстСрПокупка = ИстСумПок / ИстЗаявПок
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FI Средняя заявка на покупку")]
  public class FinInfoReaderAvgBuy : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderAvgBuy() : base("AvgBuy") { }
  }

  /// <summary>
  /// Размер средней заявки на продажу выраженный в единицах ценных бумаг
  /// ИстСрПродажа = ИстСумПрод / ИстЗаявПрод
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FI Средняя заявка на продажу")]
  public class FinInfoReaderAvgSell : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderAvgSell() : base("AvgSell") { }
  }


  /// <summary>
  ///   Объем лучшего предложения
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FI Объем лучшего предложения")]
  public class FinInfoReaderAskVolume : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderAskVolume() : base("AskVolume") { }
  }
  
  /// <summary>
  ///   Объем лучшего спроса
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FI Объем лучшего спроса")]
  public class FinInfoReaderBidVolume : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderBidVolume() : base("BidVolume") { }
  }

  /// <summary>
  ///   Цена из стакана из строки с наибольшим объемом на продажу
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FI Цена из стакана наибольшего объема на продажу")]
  public class FinInfoReaderPriceOfHiAskVolume : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderPriceOfHiAskVolume() : base("PriceOfHiAskVolume") { }
  }

  /// <summary>
  ///   Цена из стакана из строки с наибольшим объемом на покупку
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FI Цена из стакана наибольшего объема на покупку")]
  public class FinInfoReaderPriceOfHiBidVolume : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderPriceOfHiBidVolume() : base("PriceOfHiBidVolume") { }
  }

  /// <summary>
  ///   Наибольший объем на продажу из стакана
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FI Наибольший объем на продажу из стакана")]
  public class FinInfoReaderVolumeOfHiAskVolume : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderVolumeOfHiAskVolume() : base("VolumeOfHiAskVolume") { }
  }

  /// <summary>
  ///   Наибольший объем на покупку из стакана
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FI Наибольший объем на покупку из стакана")]
  public class FinInfoReaderVolumeOfHiBidVolume : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderVolumeOfHiBidVolume() : base("VolumeOfHiBidVolume") { }
  }

  /// <summary>
  ///   Весь объем из стакана на продажу
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FI Весь объем из стакана на продажу")]
  public class FinInfoReaderAllAskVolume : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderAllAskVolume() : base("AllAskVolume") { }
  }

  /// <summary>
  ///   Весь объем из стакана на покупку
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FI Весь объем из стакана на покупку")]
  public class FinInfoReaderAllBidVolume : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderAllBidVolume() : base("AllBidVolume") { }
  }

  /// <summary>
  ///   Весь объем из стакана
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FI Весь объем из стакана")]
  public class FinInfoReaderAllDomVolume : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderAllDomVolume() : base("AllDomVolume") { }
  }

  /// <summary>
  ///   Дельта объема из стакана
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FI Дельта объема из стакана")]
  public class FinInfoReaderDeltaVolume : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderDeltaVolume() : base("DeltaVolume") { }
  }
  
  //ZIP
  // Здесь добовляем то что нам нужно (LastBarClosed) Тик закрытого бара - //summary
  /// <summary>
  ///   Тик закрытого бара LastBarClosed
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("FI Тик закрытого бара")]
  public class FinInfoReaderLastBarClosed : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderLastBarClosed() : base("LastBarClosed") { }
  }
  
  
  // Здесь добовляем то что нам нужно (STBP_0) Stakan Tik Buy Price - //summary
  /// <summary>
  ///   Цена из стакана на покупку STBP_0
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("STBP_0")]
  public class FinInfoReaderSTBP_0 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBP_0() : base("STBP_0") { }
  }
  
  // Здесь добовляем то что нам нужно (STBQ_0) Stakan Tik Buy Price - //summary
  /// <summary>
  ///   Объем из стакана на покупку STBQ_0
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("STBQL_0")]
  public class FinInfoReaderSTBQ_0 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBQ_0() : base("STBQ_0") { }
  }
  
  // Здесь добовляем то что нам нужно (STSP_0) Stakan Tik Buy Price - //summary
  /// <summary>
  ///   Цена из стакана на покупку STSP_0
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("STSPL_0")]
  public class FinInfoReaderSTSP_0 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSP_0() : base("STSP_0") { }
  }
  
  // Здесь добовляем то что нам нужно (STSQ_0) Stakan Tik Buy Price - //summary
  /// <summary>
  ///   Объем из стакана на покупку STSQ_0
  /// </summary>
  [HandlerCategory(Consts.Category), HandlerName("STSQL_0")]
  public class FinInfoReaderSTSQ_0 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSQ_0() : base("STSQ_0") { }
  }

// _1
//	STBP_1
  [HandlerCategory(Consts.Category), HandlerName("STBPL_1")]
  public class FinInfoReaderSTBP_1 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBP_1() : base("STBP_1") { }
  }

//	STBQ_1
  [HandlerCategory(Consts.Category), HandlerName("STBQL_1")]
  public class FinInfoReaderSTBQ_1 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBQ_1() : base("STBQ_1") { }
  }

//	STSP_1
  [HandlerCategory(Consts.Category), HandlerName("STSPL_1")]
  public class FinInfoReaderSTSP_1 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSP_1() : base("STSP_1") { }
  }

//	STSQ_1
  [HandlerCategory(Consts.Category), HandlerName("STSQL_1")]
  public class FinInfoReaderSTSQ_1 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSQ_1() : base("STSQ_1") { }
  }
  
// _2
//	STBP_2
  [HandlerCategory(Consts.Category), HandlerName("STBPL_2")]
  public class FinInfoReaderSTBP_2 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBP_2() : base("STBP_2") { }
  }

//	STBQ_2
  [HandlerCategory(Consts.Category), HandlerName("STBQL_2")]
  public class FinInfoReaderSTBQ_2 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBQ_2() : base("STBQ_2") { }
  }

//	STSP_2
  [HandlerCategory(Consts.Category), HandlerName("STSPL_2")]
  public class FinInfoReaderSTSP_2 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSP_2() : base("STSP_2") { }
  }

//	STSQ_2
  [HandlerCategory(Consts.Category), HandlerName("STSQL_2")]
  public class FinInfoReaderSTSQ_2 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSQ_2() : base("STSQ_2") { }
  }
  
// _3
// STBP_3
  [HandlerCategory(Consts.Category), HandlerName("STBPL_3")]
  public class FinInfoReaderSTBP_3 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBP_3() : base("STBP_3") { }
  }

// STBQ_3
  [HandlerCategory(Consts.Category), HandlerName("STBQL_3")]
  public class FinInfoReaderSTBQ_3 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBQ_3() : base("STBQ_3") { }
  }

// STSP_3
  [HandlerCategory(Consts.Category), HandlerName("STSPL_3")]
  public class FinInfoReaderSTSP_3 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSP_3() : base("STSP_3") { }
  }

// STSQ_3
  [HandlerCategory(Consts.Category), HandlerName("STSQL_3")]
  public class FinInfoReaderSTSQ_3 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSQ_3() : base("STSQ_3") { }
  }
  
// _4
// STBP_4
  [HandlerCategory(Consts.Category), HandlerName("STBPL_4")]
  public class FinInfoReaderSTBP_4 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBP_4() : base("STBP_4") { }
  }

// STBQ_4
  [HandlerCategory(Consts.Category), HandlerName("STBQL_4")]
  public class FinInfoReaderSTBQ_4 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBQ_4() : base("STBQ_4") { }
  }

// STSP_4
  [HandlerCategory(Consts.Category), HandlerName("STSPL_4")]
  public class FinInfoReaderSTSP_4 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSP_4() : base("STSP_4") { }
  }

// STSQ_4
  [HandlerCategory(Consts.Category), HandlerName("STSQL_4")]
  public class FinInfoReaderSTSQ_4 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSQ_4() : base("STSQ_4") { }
  }
  
// _5
// STBP_5
  [HandlerCategory(Consts.Category), HandlerName("STBPL_5")]
  public class FinInfoReaderSTBP_5 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBP_5() : base("STBP_5") { }
  }

// STBQ_5
  [HandlerCategory(Consts.Category), HandlerName("STBQL_5")]
  public class FinInfoReaderSTBQ_5 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBQ_5() : base("STBQ_5") { }
  }

// STSP_5
  [HandlerCategory(Consts.Category), HandlerName("STSPL_5")]
  public class FinInfoReaderSTSP_5 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSP_5() : base("STSP_5") { }
  }

// STSQ_5
  [HandlerCategory(Consts.Category), HandlerName("STSQL_5")]
  public class FinInfoReaderSTSQ_5 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSQ_5() : base("STSQ_5") { }
  }
  
// _6
// STBP_6
  [HandlerCategory(Consts.Category), HandlerName("STBPL_6")]
  public class FinInfoReaderSTBP_6 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBP_6() : base("STBP_6") { }
  }

// STBQ_6
  [HandlerCategory(Consts.Category), HandlerName("STBQL_6")]
  public class FinInfoReaderSTBQ_6 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBQ_6() : base("STBQ_6") { }
  }

// STSP_6
  [HandlerCategory(Consts.Category), HandlerName("STSPL_6")]
  public class FinInfoReaderSTSP_6 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSP_6() : base("STSP_6") { }
  }

// STSQ_6
  [HandlerCategory(Consts.Category), HandlerName("STSQL_6")]
  public class FinInfoReaderSTSQ_6 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSQ_6() : base("STSQ_6") { }
  }
  
// _7
// STBP_7
  [HandlerCategory(Consts.Category), HandlerName("STBPL_7")]
  public class FinInfoReaderSTBP_7 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBP_7() : base("STBP_7") { }
  }

// STBQ_7
  [HandlerCategory(Consts.Category), HandlerName("STBQL_7")]
  public class FinInfoReaderSTBQ_7 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBQ_7() : base("STBQ_7") { }
  }

// STSP_7
  [HandlerCategory(Consts.Category), HandlerName("STSPL_7")]
  public class FinInfoReaderSTSP_7 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSP_7() : base("STSP_7") { }
  }

// STSQ_7
  [HandlerCategory(Consts.Category), HandlerName("STSQL_7")]
  public class FinInfoReaderSTSQ_7 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSQ_7() : base("STSQ_7") { }
  }
  
// _8
// STBP_8
  [HandlerCategory(Consts.Category), HandlerName("STBPL_8")]
  public class FinInfoReaderSTBP_8 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBP_8() : base("STBP_8") { }
  }

// STBQ_8
  [HandlerCategory(Consts.Category), HandlerName("STBQL_8")]
  public class FinInfoReaderSTBQ_8 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBQ_8() : base("STBQ_8") { }
  }

// STSP_8
  [HandlerCategory(Consts.Category), HandlerName("STSPL_8")]
  public class FinInfoReaderSTSP_8 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSP_8() : base("STSP_8") { }
  }

// STSQ_8
  [HandlerCategory(Consts.Category), HandlerName("STSQL_8")]
  public class FinInfoReaderSTSQ_8 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSQ_8() : base("STSQ_8") { }
  }
  
// _9
// STBP_9
  [HandlerCategory(Consts.Category), HandlerName("STBPL_9")]
  public class FinInfoReaderSTBP_9 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBP_9() : base("STBP_9") { }
  }

// STBQ_9
  [HandlerCategory(Consts.Category), HandlerName("STBQL_9")]
  public class FinInfoReaderSTBQ_9 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBQ_9() : base("STBQ_9") { }
  }

// STSP_9
  [HandlerCategory(Consts.Category), HandlerName("STSPL_9")]
  public class FinInfoReaderSTSP_9 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSP_9() : base("STSP_9") { }
  }

// STSQ_9
  [HandlerCategory(Consts.Category), HandlerName("STSQL_9")]
  public class FinInfoReaderSTSQ_9 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSQ_9() : base("STSQ_9") { }
  }
  
// _10
// STBP_10
  [HandlerCategory(Consts.Category), HandlerName("STBPL_10")]
  public class FinInfoReaderSTBP_10 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBP_10() : base("STBP_10") { }
  }

// STBQ_10
  [HandlerCategory(Consts.Category), HandlerName("STBQL_10")]
  public class FinInfoReaderSTBQ_10 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBQ_10() : base("STBQ_10") { }
  }

// STSP_10
  [HandlerCategory(Consts.Category), HandlerName("STSPL_10")]
  public class FinInfoReaderSTSP_10 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSP_10() : base("STSP_10") { }
  }

// STSQ_10
  [HandlerCategory(Consts.Category), HandlerName("STSQL_10")]
  public class FinInfoReaderSTSQ_10 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSQ_10() : base("STSQ_10") { }
  }
  
// _11
// STBP_11
  [HandlerCategory(Consts.Category), HandlerName("STBPL_11")]
  public class FinInfoReaderSTBP_11 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBP_11() : base("STBP_11") { }
  }

// STBQ_11
  [HandlerCategory(Consts.Category), HandlerName("STBQL_11")]
  public class FinInfoReaderSTBQ_11 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBQ_11() : base("STBQ_11") { }
  }

// STSP_11
  [HandlerCategory(Consts.Category), HandlerName("STSPL_11")]
  public class FinInfoReaderSTSP_11 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSP_11() : base("STSP_11") { }
  }

// STSQ_11
  [HandlerCategory(Consts.Category), HandlerName("STSQL_11")]
  public class FinInfoReaderSTSQ_11 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSQ_11() : base("STSQ_11") { }
  }
   
// _12
// STBP_12
  [HandlerCategory(Consts.Category), HandlerName("STBPL_12")]
  public class FinInfoReaderSTBP_12 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBP_12() : base("STBP_12") { }
  }

// STBQ_12
  [HandlerCategory(Consts.Category), HandlerName("STBQL_12")]
  public class FinInfoReaderSTBQ_12 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBQ_12() : base("STBQ_12") { }
  }

// STSP_12
  [HandlerCategory(Consts.Category), HandlerName("STSPL_12")]
  public class FinInfoReaderSTSP_12 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSP_12() : base("STSP_12") { }
  }

// STSQ_12
  [HandlerCategory(Consts.Category), HandlerName("STSQL_12")]
  public class FinInfoReaderSTSQ_12 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSQ_12() : base("STSQ_12") { }
  }
  
// _13
// STBP_13
  [HandlerCategory(Consts.Category), HandlerName("STBPL_13")]
  public class FinInfoReaderSTBP_13 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBP_13() : base("STBP_13") { }
  }

// STBQ_13
  [HandlerCategory(Consts.Category), HandlerName("STBQL_13")]
  public class FinInfoReaderSTBQ_13 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBQ_13() : base("STBQ_13") { }
  }

// STSP_13
  [HandlerCategory(Consts.Category), HandlerName("STSPL_13")]
  public class FinInfoReaderSTSP_13 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSP_13() : base("STSP_13") { }
  }

// STSQ_13
  [HandlerCategory(Consts.Category), HandlerName("STSQL_13")]
  public class FinInfoReaderSTSQ_13 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSQ_13() : base("STSQ_13") { }
  }
  
// _14
// STBP_14
  [HandlerCategory(Consts.Category), HandlerName("STBPL_14")]
  public class FinInfoReaderSTBP_14 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBP_14() : base("STBP_14") { }
  }

// STBQ_14
  [HandlerCategory(Consts.Category), HandlerName("STBQL_14")]
  public class FinInfoReaderSTBQ_14 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBQ_14() : base("STBQ_14") { }
  }

// STSP_14
  [HandlerCategory(Consts.Category), HandlerName("STSPL_14")]
  public class FinInfoReaderSTSP_14 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSP_14() : base("STSP_14") { }
  }

// STSQ_14
  [HandlerCategory(Consts.Category), HandlerName("STSQL_14")]
  public class FinInfoReaderSTSQ_14 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSQ_14() : base("STSQ_14") { }
  }
  
// _15
// STBP_15
  [HandlerCategory(Consts.Category), HandlerName("STBPL_15")]
  public class FinInfoReaderSTBP_15 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBP_15() : base("STBP_15") { }
  }

// STBQ_15
  [HandlerCategory(Consts.Category), HandlerName("STBQL_15")]
  public class FinInfoReaderSTBQ_15 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBQ_15() : base("STBQ_15") { }
  }

// STSP_15
  [HandlerCategory(Consts.Category), HandlerName("STSPL_15")]
  public class FinInfoReaderSTSP_15 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSP_15() : base("STSP_15") { }
  }

// STSQ_15
  [HandlerCategory(Consts.Category), HandlerName("STSQL_15")]
  public class FinInfoReaderSTSQ_15 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSQ_15() : base("STSQ_15") { }
  }
  
// _16
// STBP_16
  [HandlerCategory(Consts.Category), HandlerName("STBPL_16")]
  public class FinInfoReaderSTBP_16 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBP_16() : base("STBP_16") { }
  }

// STBQ_16
  [HandlerCategory(Consts.Category), HandlerName("STBQL_16")]
  public class FinInfoReaderSTBQ_16 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBQ_16() : base("STBQ_16") { }
  }

// STSP_16
  [HandlerCategory(Consts.Category), HandlerName("STSPL_16")]
  public class FinInfoReaderSTSP_16 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSP_16() : base("STSP_16") { }
  }

// STSQ_16
  [HandlerCategory(Consts.Category), HandlerName("STSQL_16")]
  public class FinInfoReaderSTSQ_16 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSQ_16() : base("STSQ_16") { }
  }
  
// _17
// STBP_17
  [HandlerCategory(Consts.Category), HandlerName("STBPL_17")]
  public class FinInfoReaderSTBP_17 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBP_17() : base("STBP_17") { }
  }

// STBQ_17
  [HandlerCategory(Consts.Category), HandlerName("STBQL_17")]
  public class FinInfoReaderSTBQ_17 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBQ_17() : base("STBQ_17") { }
  }

// STSP_17
  [HandlerCategory(Consts.Category), HandlerName("STSPL_17")]
  public class FinInfoReaderSTSP_17 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSP_17() : base("STSP_17") { }
  }

// STSQ_17
  [HandlerCategory(Consts.Category), HandlerName("STSQL_17")]
  public class FinInfoReaderSTSQ_17 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSQ_17() : base("STSQ_17") { }
  }
  
// _18
// STBP_18
  [HandlerCategory(Consts.Category), HandlerName("STBPL_18")]
  public class FinInfoReaderSTBP_18 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBP_18() : base("STBP_18") { }
  }

// STBQ_18
  [HandlerCategory(Consts.Category), HandlerName("STBQL_18")]
  public class FinInfoReaderSTBQ_18 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBQ_18() : base("STBQ_18") { }
  }

// STSP_18
  [HandlerCategory(Consts.Category), HandlerName("STSPL_18")]
  public class FinInfoReaderSTSP_18 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSP_18() : base("STSP_18") { }
  }

// STSQ_18
  [HandlerCategory(Consts.Category), HandlerName("STSQL_18")]
  public class FinInfoReaderSTSQ_18 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSQ_18() : base("STSQ_18") { }
  }
  
// _19
// STBP_19
  [HandlerCategory(Consts.Category), HandlerName("STBPL_19")]
  public class FinInfoReaderSTBP_19 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBP_19() : base("STBP_19") { }
  }

// STBQ_19
  [HandlerCategory(Consts.Category), HandlerName("STBQL_19")]
  public class FinInfoReaderSTBQ_19 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBQ_19() : base("STBQ_19") { }
  }

// STSP_19
  [HandlerCategory(Consts.Category), HandlerName("STSPL_19")]
  public class FinInfoReaderSTSP_19 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSP_19() : base("STSP_19") { }
  }

// STSQ_19
  [HandlerCategory(Consts.Category), HandlerName("STSQL_19")]
  public class FinInfoReaderSTSQ_19 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSQ_19() : base("STSQ_19") { }
  }
  
// _20
// STBP_20
  [HandlerCategory(Consts.Category), HandlerName("STBPL_20")]
  public class FinInfoReaderSTBP_20 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBP_20() : base("STBP_20") { }
  }

// STBQ_20
  [HandlerCategory(Consts.Category), HandlerName("STBQL_20")]
  public class FinInfoReaderSTBQ_20 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBQ_20() : base("STBQ_20") { }
  }

// STSP_20
  [HandlerCategory(Consts.Category), HandlerName("STSPL_20")]
  public class FinInfoReaderSTSP_20 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSP_20() : base("STSP_20") { }
  }

// STSQ_20
  [HandlerCategory(Consts.Category), HandlerName("STSQL_20")]
  public class FinInfoReaderSTSQ_20 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSQ_20() : base("STSQ_20") { }
  }
    
// _21
// STBP_21
  [HandlerCategory(Consts.Category), HandlerName("STBPL_21")]
  public class FinInfoReaderSTBP_21 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBP_21() : base("STBP_21") { }
  }

// STBQ_21
  [HandlerCategory(Consts.Category), HandlerName("STBQL_21")]
  public class FinInfoReaderSTBQ_21 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBQ_21() : base("STBQ_21") { }
  }

// STSP_21
  [HandlerCategory(Consts.Category), HandlerName("STSPL_21")]
  public class FinInfoReaderSTSP_21 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSP_21() : base("STSP_21") { }
  }

// STSQ_21
  [HandlerCategory(Consts.Category), HandlerName("STSQL_21")]
  public class FinInfoReaderSTSQ_21 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSQ_21() : base("STSQ_21") { }
  }
    
// _22
// STBP_22
  [HandlerCategory(Consts.Category), HandlerName("STBPL_22")]
  public class FinInfoReaderSTBP_22 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBP_22() : base("STBP_22") { }
  }

// STBQ_22
  [HandlerCategory(Consts.Category), HandlerName("STBQL_22")]
  public class FinInfoReaderSTBQ_22 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBQ_22() : base("STBQ_22") { }
  }

// STSP_22
  [HandlerCategory(Consts.Category), HandlerName("STSPL_22")]
  public class FinInfoReaderSTSP_22 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSP_22() : base("STSP_22") { }
  }

// STSQ_22
  [HandlerCategory(Consts.Category), HandlerName("STSQL_22")]
  public class FinInfoReaderSTSQ_22 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSQ_22() : base("STSQ_22") { }
  }
    
// _23
// STBP_23
  [HandlerCategory(Consts.Category), HandlerName("STBPL_23")]
  public class FinInfoReaderSTBP_23 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBP_23() : base("STBP_23") { }
  }

// STBQ_23
  [HandlerCategory(Consts.Category), HandlerName("STBQL_23")]
  public class FinInfoReaderSTBQ_23 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBQ_23() : base("STBQ_23") { }
  }

// STSP_23
  [HandlerCategory(Consts.Category), HandlerName("STSPL_23")]
  public class FinInfoReaderSTSP_23 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSP_23() : base("STSP_23") { }
  }

// STSQ_23
  [HandlerCategory(Consts.Category), HandlerName("STSQL_23")]
  public class FinInfoReaderSTSQ_23 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSQ_23() : base("STSQ_23") { }
  }
    
// _24
// STBP_24
  [HandlerCategory(Consts.Category), HandlerName("STBPL_24")]
  public class FinInfoReaderSTBP_24 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBP_24() : base("STBP_24") { }
  }

// STBQ_24
  [HandlerCategory(Consts.Category), HandlerName("STBQL_24")]
  public class FinInfoReaderSTBQ_24 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBQ_24() : base("STBQ_24") { }
  }

// STSP_24
  [HandlerCategory(Consts.Category), HandlerName("STSPL_24")]
  public class FinInfoReaderSTSP_24 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSP_24() : base("STSP_24") { }
  }

// STSQ_24
  [HandlerCategory(Consts.Category), HandlerName("STSQL_24")]
  public class FinInfoReaderSTSQ_24 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSQ_24() : base("STSQ_24") { }
  }
    
// _25
// STBP_25
  [HandlerCategory(Consts.Category), HandlerName("STBPL_25")]
  public class FinInfoReaderSTBP_25 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBP_25() : base("STBP_25") { }
  }

// STBQ_25
  [HandlerCategory(Consts.Category), HandlerName("STBQL_25")]
  public class FinInfoReaderSTBQ_25 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBQ_25() : base("STBQ_25") { }
  }

// STSP_25
  [HandlerCategory(Consts.Category), HandlerName("STSPL_25")]
  public class FinInfoReaderSTSP_25 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSP_25() : base("STSP_25") { }
  }

// STSQ_25
  [HandlerCategory(Consts.Category), HandlerName("STSQL_25")]
  public class FinInfoReaderSTSQ_25 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSQ_25() : base("STSQ_25") { }
  }
  
// _26
// STBP_26
  [HandlerCategory(Consts.Category), HandlerName("STBPL_26")]
  public class FinInfoReaderSTBP_26 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBP_26() : base("STBP_26") { }
  }

// STBQ_26
  [HandlerCategory(Consts.Category), HandlerName("STBQL_26")]
  public class FinInfoReaderSTBQ_26 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBQ_26() : base("STBQ_26") { }
  }

// STSP_26
  [HandlerCategory(Consts.Category), HandlerName("STSPL_26")]
  public class FinInfoReaderSTSP_26 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSP_26() : base("STSP_26") { }
  }

// STSQ_26
  [HandlerCategory(Consts.Category), HandlerName("STSQL_26")]
  public class FinInfoReaderSTSQ_26 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSQ_26() : base("STSQ_26") { }
  }
  
// _27
// STBP_27
  [HandlerCategory(Consts.Category), HandlerName("STBPL_27")]
  public class FinInfoReaderSTBP_27 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBP_27() : base("STBP_27") { }
  }

// STBQ_27
  [HandlerCategory(Consts.Category), HandlerName("STBQL_27")]
  public class FinInfoReaderSTBQ_27 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBQ_27() : base("STBQ_27") { }
  }

// STSP_27
  [HandlerCategory(Consts.Category), HandlerName("STSPL_27")]
  public class FinInfoReaderSTSP_27 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSP_27() : base("STSP_27") { }
  }

// STSQ_27
  [HandlerCategory(Consts.Category), HandlerName("STSQL_27")]
  public class FinInfoReaderSTSQ_27 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSQ_27() : base("STSQ_27") { }
  }
  
// _28
// STBP_28
  [HandlerCategory(Consts.Category), HandlerName("STBPL_28")]
  public class FinInfoReaderSTBP_28 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBP_28() : base("STBP_28") { }
  }

// STBQ_28
  [HandlerCategory(Consts.Category), HandlerName("STBQL_28")]
  public class FinInfoReaderSTBQ_28 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBQ_28() : base("STBQ_28") { }
  }

// STSP_28
  [HandlerCategory(Consts.Category), HandlerName("STSPL_28")]
  public class FinInfoReaderSTSP_28 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSP_28() : base("STSP_28") { }
  }

// STSQ_28
  [HandlerCategory(Consts.Category), HandlerName("STSQL_28")]
  public class FinInfoReaderSTSQ_28 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSQ_28() : base("STSQ_28") { }
  }
  
// _29
// STBP_29
  [HandlerCategory(Consts.Category), HandlerName("STBPL_29")]
  public class FinInfoReaderSTBP_29 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBP_29() : base("STBP_29") { }
  }

// STBQ_29
  [HandlerCategory(Consts.Category), HandlerName("STBQL_29")]
  public class FinInfoReaderSTBQ_29 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBQ_29() : base("STBQ_29") { }
  }

// STSP_29
  [HandlerCategory(Consts.Category), HandlerName("STSPL_29")]
  public class FinInfoReaderSTSP_29 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSP_29() : base("STSP_29") { }
  }

// STSQ_29
  [HandlerCategory(Consts.Category), HandlerName("STSQL_29")]
  public class FinInfoReaderSTSQ_29 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSQ_29() : base("STSQ_29") { }
  }
  
// _30
// STBP_30
  [HandlerCategory(Consts.Category), HandlerName("STBPL_30")]
  public class FinInfoReaderSTBP_30 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBP_30() : base("STBP_30") { }
  }

// STBQ_30
  [HandlerCategory(Consts.Category), HandlerName("STBQL_30")]
  public class FinInfoReaderSTBQ_30 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBQ_30() : base("STBQ_30") { }
  }

// STSP_30
  [HandlerCategory(Consts.Category), HandlerName("STSPL_30")]
  public class FinInfoReaderSTSP_30 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSP_30() : base("STSP_30") { }
  }

// STSQ_30
  [HandlerCategory(Consts.Category), HandlerName("STSQL_30")]
  public class FinInfoReaderSTSQ_30 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSQ_30() : base("STSQ_30") { }
  }
  
// _31
// STBP_31
  [HandlerCategory(Consts.Category), HandlerName("STBPL_31")]
  public class FinInfoReaderSTBP_31 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBP_31() : base("STBP_31") { }
  }

// STBQ_31
  [HandlerCategory(Consts.Category), HandlerName("STBQL_31")]
  public class FinInfoReaderSTBQ_31 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBQ_31() : base("STBQ_31") { }
  }

// STSP_31
  [HandlerCategory(Consts.Category), HandlerName("STSPL_31")]
  public class FinInfoReaderSTSP_31 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSP_31() : base("STSP_31") { }
  }

// STSQ_31
  [HandlerCategory(Consts.Category), HandlerName("STSQL_31")]
  public class FinInfoReaderSTSQ_31 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSQ_31() : base("STSQ_31") { }
  }
  
// _32
// STBP_32
  [HandlerCategory(Consts.Category), HandlerName("STBPL_32")]
  public class FinInfoReaderSTBP_32 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBP_32() : base("STBP_32") { }
  }

// STBQ_32
  [HandlerCategory(Consts.Category), HandlerName("STBQL_32")]
  public class FinInfoReaderSTBQ_32 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBQ_32() : base("STBQ_32") { }
  }

// STSP_32
  [HandlerCategory(Consts.Category), HandlerName("STSPL_32")]
  public class FinInfoReaderSTSP_32 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSP_32() : base("STSP_32") { }
  }

// STSQ_32
  [HandlerCategory(Consts.Category), HandlerName("STSQL_32")]
  public class FinInfoReaderSTSQ_32 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSQ_32() : base("STSQ_32") { }
  }
  
// _33
// STBP_33
  [HandlerCategory(Consts.Category), HandlerName("STBPL_33")]
  public class FinInfoReaderSTBP_33 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBP_33() : base("STBP_33") { }
  }

// STBQ_33
  [HandlerCategory(Consts.Category), HandlerName("STBQL_33")]
  public class FinInfoReaderSTBQ_33 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBQ_33() : base("STBQ_33") { }
  }

// STSP_33
  [HandlerCategory(Consts.Category), HandlerName("STSPL_33")]
  public class FinInfoReaderSTSP_33 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSP_33() : base("STSP_33") { }
  }

// STSQ_33
  [HandlerCategory(Consts.Category), HandlerName("STSQL_33")]
  public class FinInfoReaderSTSQ_33 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSQ_33() : base("STSQ_33") { }
  }
  
// _34
// STBP_34
  [HandlerCategory(Consts.Category), HandlerName("STBPL_34")]
  public class FinInfoReaderSTBP_34 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBP_34() : base("STBP_34") { }
  }

// STBQ_34
  [HandlerCategory(Consts.Category), HandlerName("STBQL_34")]
  public class FinInfoReaderSTBQ_34 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBQ_34() : base("STBQ_34") { }
  }

// STSP_34
  [HandlerCategory(Consts.Category), HandlerName("STSPL_34")]
  public class FinInfoReaderSTSP_34 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSP_34() : base("STSP_34") { }
  }

// STSQ_34
  [HandlerCategory(Consts.Category), HandlerName("STSQL_34")]
  public class FinInfoReaderSTSQ_34 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSQ_34() : base("STSQ_34") { }
  }
    
// _35
// STBP_35
  [HandlerCategory(Consts.Category), HandlerName("STBPL_35")]
  public class FinInfoReaderSTBP_35 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBP_35() : base("STBP_35") { }
  }

// STBQ_35
  [HandlerCategory(Consts.Category), HandlerName("STBQL_35")]
  public class FinInfoReaderSTBQ_35 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBQ_35() : base("STBQ_35") { }
  }

// STSP_35
  [HandlerCategory(Consts.Category), HandlerName("STSPL_35")]
  public class FinInfoReaderSTSP_35 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSP_35() : base("STSP_35") { }
  }

// STSQ_35
  [HandlerCategory(Consts.Category), HandlerName("STSQL_35")]
  public class FinInfoReaderSTSQ_35 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSQ_35() : base("STSQ_35") { }
  }
    
// _36
// STBP_36
  [HandlerCategory(Consts.Category), HandlerName("STBPL_36")]
  public class FinInfoReaderSTBP_36 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBP_36() : base("STBP_36") { }
  }

// STBQ_36
  [HandlerCategory(Consts.Category), HandlerName("STBQL_36")]
  public class FinInfoReaderSTBQ_36 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBQ_36() : base("STBQ_36") { }
  }

// STSP_36
  [HandlerCategory(Consts.Category), HandlerName("STSPL_36")]
  public class FinInfoReaderSTSP_36 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSP_36() : base("STSP_36") { }
  }

// STSQ_36
  [HandlerCategory(Consts.Category), HandlerName("STSQL_36")]
  public class FinInfoReaderSTSQ_36 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSQ_36() : base("STSQ_36") { }
  }
    
// _37
// STBP_37
  [HandlerCategory(Consts.Category), HandlerName("STBPL_37")]
  public class FinInfoReaderSTBP_37 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBP_37() : base("STBP_37") { }
  }

// STBQ_37
  [HandlerCategory(Consts.Category), HandlerName("STBQL_37")]
  public class FinInfoReaderSTBQ_37 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBQ_37() : base("STBQ_37") { }
  }

// STSP_37
  [HandlerCategory(Consts.Category), HandlerName("STSPL_37")]
  public class FinInfoReaderSTSP_37 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSP_37() : base("STSP_37") { }
  }

// STSQ_37
  [HandlerCategory(Consts.Category), HandlerName("STSQL_37")]
  public class FinInfoReaderSTSQ_37 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSQ_37() : base("STSQ_37") { }
  }
    
// _38
// STBP_38
  [HandlerCategory(Consts.Category), HandlerName("STBPL_38")]
  public class FinInfoReaderSTBP_38 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBP_38() : base("STBP_38") { }
  }

// STBQ_38
  [HandlerCategory(Consts.Category), HandlerName("STBQL_38")]
  public class FinInfoReaderSTBQ_38 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBQ_38() : base("STBQ_38") { }
  }

// STSP_38
  [HandlerCategory(Consts.Category), HandlerName("STSPL_38")]
  public class FinInfoReaderSTSP_38 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSP_38() : base("STSP_38") { }
  }

// STSQ_38
  [HandlerCategory(Consts.Category), HandlerName("STSQL_38")]
  public class FinInfoReaderSTSQ_38 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSQ_38() : base("STSQ_38") { }
  }
    
// _39
// STBP_39
  [HandlerCategory(Consts.Category), HandlerName("STBPL_39")]
  public class FinInfoReaderSTBP_39 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBP_39() : base("STBP_39") { }
  }

// STBQ_39
  [HandlerCategory(Consts.Category), HandlerName("STBQL_39")]
  public class FinInfoReaderSTBQ_39 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBQ_39() : base("STBQ_39") { }
  }

// STSP_39
  [HandlerCategory(Consts.Category), HandlerName("STSPL_39")]
  public class FinInfoReaderSTSP_39 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSP_39() : base("STSP_39") { }
  }

// STSQ_39
  [HandlerCategory(Consts.Category), HandlerName("STSQL_39")]
  public class FinInfoReaderSTSQ_39 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSQ_39() : base("STSQ_39") { }
  }
  
// _40
// STBP_40
  [HandlerCategory(Consts.Category), HandlerName("STBPL_40")]
  public class FinInfoReaderSTBP_40 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBP_40() : base("STBP_40") { }
  }

// STBQ_40
  [HandlerCategory(Consts.Category), HandlerName("STBQL_40")]
  public class FinInfoReaderSTBQ_40 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBQ_40() : base("STBQ_40") { }
  }

// STSP_40
  [HandlerCategory(Consts.Category), HandlerName("STSPL_40")]
  public class FinInfoReaderSTSP_40 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSP_40() : base("STSP_40") { }
  }

// STSQ_40
  [HandlerCategory(Consts.Category), HandlerName("STSQL_40")]
  public class FinInfoReaderSTSQ_40 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSQ_40() : base("STSQ_40") { }
  }
  
// _41
// STBP_41
  [HandlerCategory(Consts.Category), HandlerName("STBPL_41")]
  public class FinInfoReaderSTBP_41 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBP_41() : base("STBP_41") { }
  }

// STBQ_41
  [HandlerCategory(Consts.Category), HandlerName("STBQL_41")]
  public class FinInfoReaderSTBQ_41 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBQ_41() : base("STBQ_41") { }
  }

// STSP_41
  [HandlerCategory(Consts.Category), HandlerName("STSPL_41")]
  public class FinInfoReaderSTSP_41 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSP_41() : base("STSP_41") { }
  }

// STSQ_41
  [HandlerCategory(Consts.Category), HandlerName("STSQL_41")]
  public class FinInfoReaderSTSQ_41 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSQ_41() : base("STSQ_41") { }
  }
  
// _42
// STBP_42
  [HandlerCategory(Consts.Category), HandlerName("STBPL_42")]
  public class FinInfoReaderSTBP_42 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBP_42() : base("STBP_42") { }
  }

// STBQ_42
  [HandlerCategory(Consts.Category), HandlerName("STBQL_42")]
  public class FinInfoReaderSTBQ_42 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBQ_42() : base("STBQ_42") { }
  }

// STSP_42
  [HandlerCategory(Consts.Category), HandlerName("STSPL_42")]
  public class FinInfoReaderSTSP_42 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSP_42() : base("STSP_42") { }
  }

// STSQ_42
  [HandlerCategory(Consts.Category), HandlerName("STSQL_42")]
  public class FinInfoReaderSTSQ_42 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSQ_42() : base("STSQ_42") { }
  }
  
// _43
// STBP_43
  [HandlerCategory(Consts.Category), HandlerName("STBPL_43")]
  public class FinInfoReaderSTBP_43 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBP_43() : base("STBP_43") { }
  }

// STBQ_43
  [HandlerCategory(Consts.Category), HandlerName("STBQL_43")]
  public class FinInfoReaderSTBQ_43 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBQ_43() : base("STBQ_43") { }
  }

// STSP_43
  [HandlerCategory(Consts.Category), HandlerName("STSPL_43")]
  public class FinInfoReaderSTSP_43 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSP_43() : base("STSP_43") { }
  }

// STSQ_43
  [HandlerCategory(Consts.Category), HandlerName("STSQL_43")]
  public class FinInfoReaderSTSQ_43 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSQ_43() : base("STSQ_43") { }
  }
  
// _44
// STBP_44
  [HandlerCategory(Consts.Category), HandlerName("STBPL_44")]
  public class FinInfoReaderSTBP_44 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBP_44() : base("STBP_44") { }
  }

// STBQ_44
  [HandlerCategory(Consts.Category), HandlerName("STBQL_44")]
  public class FinInfoReaderSTBQ_44 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBQ_44() : base("STBQ_44") { }
  }

// STSP_44
  [HandlerCategory(Consts.Category), HandlerName("STSPL_44")]
  public class FinInfoReaderSTSP_44 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSP_44() : base("STSP_44") { }
  }

// STSQ_44
  [HandlerCategory(Consts.Category), HandlerName("STSQL_44")]
  public class FinInfoReaderSTSQ_44 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSQ_44() : base("STSQ_44") { }
  }
  
// _45
// STBP_45
  [HandlerCategory(Consts.Category), HandlerName("STBPL_45")]
  public class FinInfoReaderSTBP_45 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBP_45() : base("STBP_45") { }
  }

// STBQ_45
  [HandlerCategory(Consts.Category), HandlerName("STBQL_45")]
  public class FinInfoReaderSTBQ_45 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBQ_45() : base("STBQ_45") { }
  }

// STSP_45
  [HandlerCategory(Consts.Category), HandlerName("STSPL_45")]
  public class FinInfoReaderSTSP_45 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSP_45() : base("STSP_45") { }
  }

// STSQ_45
  [HandlerCategory(Consts.Category), HandlerName("STSQL_45")]
  public class FinInfoReaderSTSQ_45 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSQ_45() : base("STSQ_45") { }
  }
  
// _46
// STBP_46
  [HandlerCategory(Consts.Category), HandlerName("STBPL_46")]
  public class FinInfoReaderSTBP_46 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBP_46() : base("STBP_46") { }
  }

// STBQ_46
  [HandlerCategory(Consts.Category), HandlerName("STBQL_46")]
  public class FinInfoReaderSTBQ_46 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBQ_46() : base("STBQ_46") { }
  }

// STSP_46
  [HandlerCategory(Consts.Category), HandlerName("STSPL_46")]
  public class FinInfoReaderSTSP_46 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSP_46() : base("STSP_46") { }
  }

// STSQ_46
  [HandlerCategory(Consts.Category), HandlerName("STSQL_46")]
  public class FinInfoReaderSTSQ_46 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSQ_46() : base("STSQ_46") { }
  }
  
// _47
// STBP_47
  [HandlerCategory(Consts.Category), HandlerName("STBPL_47")]
  public class FinInfoReaderSTBP_47 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBP_47() : base("STBP_47") { }
  }

// STBQ_47
  [HandlerCategory(Consts.Category), HandlerName("STBQL_47")]
  public class FinInfoReaderSTBQ_47 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBQ_47() : base("STBQ_47") { }
  }

// STSP_47
  [HandlerCategory(Consts.Category), HandlerName("STSPL_47")]
  public class FinInfoReaderSTSP_47 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSP_47() : base("STSP_47") { }
  }

// STSQ_47
  [HandlerCategory(Consts.Category), HandlerName("STSQL_47")]
  public class FinInfoReaderSTSQ_47 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSQ_47() : base("STSQ_47") { }
  }
  
// _48
// STBP_48
  [HandlerCategory(Consts.Category), HandlerName("STBPL_48")]
  public class FinInfoReaderSTBP_48 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBP_48() : base("STBP_48") { }
  }

// STBQ_48
  [HandlerCategory(Consts.Category), HandlerName("STBQL_48")]
  public class FinInfoReaderSTBQ_48 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBQ_48() : base("STBQ_48") { }
  }

// STSP_48
  [HandlerCategory(Consts.Category), HandlerName("STSPL_48")]
  public class FinInfoReaderSTSP_48 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSP_48() : base("STSP_48") { }
  }

// STSQ_48
  [HandlerCategory(Consts.Category), HandlerName("STSQL_48")]
  public class FinInfoReaderSTSQ_48 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSQ_48() : base("STSQ_48") { }
  }
    
// _49
// STBP_49
  [HandlerCategory(Consts.Category), HandlerName("STBPL_49")]
  public class FinInfoReaderSTBP_49 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBP_49() : base("STBP_49") { }
  }

// STBQ_49
  [HandlerCategory(Consts.Category), HandlerName("STBQL_49")]
  public class FinInfoReaderSTBQ_49 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTBQ_49() : base("STBQ_49") { }
  }

// STSP_49
  [HandlerCategory(Consts.Category), HandlerName("STSPL_49")]
  public class FinInfoReaderSTSP_49 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSP_49() : base("STSP_49") { }
  }

// STSQ_49
  [HandlerCategory(Consts.Category), HandlerName("STSQL_49")]
  public class FinInfoReaderSTSQ_49 : FinInfoReaderValue, IBar2DoubleHandler, IContextUses {
    public FinInfoReaderSTSQ_49() : base("STSQ_49") { }
  }
  
//  
  delegate void ActionEndOfBarFinInfo(int iBar, FinInfoBlock fb);

  sealed class FinInfoCache {
    const string Extension = ".fid";
    //const string LogExtension = ".log";
    const byte FileVer = 4;
    readonly Dictionary<string, List<FinInfoBlock>> _dict = new Dictionary<string, List<FinInfoBlock>>(4);
    // Оставлен статическим, чтобы разные скрипты не писали повторяющиеся значения в один и тот же файл
    static readonly Dictionary<string, FinInfoBlock> _lastFb = new Dictionary<string, FinInfoBlock>(4);
    static TimeZoneInfo _timeZone;
    /// <summary> Часовой пояс в котором работает биржа </summary>
    public static TimeZoneInfo ExchangeTimeZone { get { return _timeZone ?? (_timeZone = TimeZoneInfo.FindSystemTimeZoneById("Russian Standard Time")); } }

    public static void WriteFininfo(ISecurity security, string path, bool all, bool logging) {
      // текущие значения Fininfo
      var fi = security.FinInfo;
      if (fi == null || fi.LastUpdate.Ticks == 0 || !FinInfoContainsData(fi))
        return;

      var fb = new FinInfoBlock(fi);
      fb.GetDomValues(security);
     
      //StreamWriter logStream = null;
      //if (logging) {
      //  string logFName = Path.Combine(path, security.Symbol + LogExtension);
      //  try {
      //    logStream = new StreamWriter(logFName, true, Encoding.Default);
      //    //logStream.WriteLine();
      //  } catch {}
      //}

      //try {
        var key = security.Symbol + '_' + path;
        lock (_lastFb) {
          FinInfoBlock last;
          if (_lastFb.TryGetValue(key, out last)) {
            if (fb.LastUpdate < last.LastUpdate)
              return; // Время не может уменьшаться. В клиринг какие-то дубли через раз приходят
            var eq = fb.EqualsExceptLastUpdate(last); // Значения не изменились
            if (fb.LastUpdate == last.LastUpdate && eq)
              return; // Ничего не изменилось
            if (!all && eq)
              return; // Не писать повторы
          }

          using (FileStream fs = OpenAndPepareFile(Path.Combine(path, security.Symbol + Extension))) {
            if (last == null) {
              // Последнего значения нет - значит первый раз, попробовать прочитать
              long rec = (fs.Length - FinInfoBlock.HeaderSize)/FinInfoBlock.RecSize - 1;
              if (rec >= 0) {
                last = FinInfoBlock.ReadFininfoBlock(new BinaryReader(fs), rec);
                if (fb.LastUpdate < last.LastUpdate) {
                  _lastFb[key] = last;
                  return;
                }
                var eq = fb.EqualsExceptLastUpdate(last); // Значения не изменились
                if (fb.LastUpdate == last.LastUpdate && eq) {
                  _lastFb[key] = last;
                  return;
                }
                if (!all && eq) {
                  _lastFb[key] = last;
                  return;
                }
              }
            }
            _lastFb[key] = fb;

            // Записать новую запись
            long recCount = (fs.Length - FinInfoBlock.HeaderSize)/FinInfoBlock.RecSize;
            fs.Seek(FinInfoBlock.HeaderSize + recCount*FinInfoBlock.RecSize, SeekOrigin.Begin);
            using (var bw = new BinaryWriter(fs))
              fb.WriteRecord(bw);
          }
        }
      //} finally {
      //  if (logStream != null)
      //    logStream.Dispose();
      //}
    }

    static FileStream OpenAndPepareFile(string fname) {
      FileStream fs = File.Open(fname, FileMode.OpenOrCreate, FileAccess.ReadWrite, FileShare.Read);
      if (fs.Length > 0) {
        int v = fs.ReadByte();
        // Преобразовать версию файла, если надо
        if (v == 1 || v == 2 || v == 3) {
          fs.Close();  // Это позволит избежать лишнего открывания файла при обычном сценарии работы
          ConvertFileVersion(fname);
          fs = File.Open(fname, FileMode.OpenOrCreate, FileAccess.ReadWrite, FileShare.Read);
        } else if (v != FileVer)
          throw new ArgumentException("Ошибка в блоке записи данных FinInfo: неподходящий формат файла: " + fname);
      } else
        fs.WriteByte(FileVer); // Номер версии формата
      return fs;
    }

    //public IList<FinInfoBlock> GetFininfoBlocks(ISecurity security, IContext context, string path, bool log) {
    //  // Сохранить текущий FinInfoBlock
    //  WriteFininfo(security, path, false, log);

    //  // Построить выборку нужных FinInfoBlock по барам
    //  var bars = security.Bars;
    //  int barsCount = bars.Count;
    //  if (barsCount <= 0)
    //    return new FinInfoBlock[0];

    //  var key = GetKey(security, path);
    //  List<FinInfoBlock> fbList;
    //  lock (_dict) {
    //    if (!_dict.TryGetValue(key, out fbList))
    //      _dict[key] = fbList = new List<FinInfoBlock>(barsCount);

    //    // Обновить данные за левой и правой границами
    //    ReadFininfo(security, path, fbList);
    //  }

    //  var values = new FinInfoBlock[barsCount];
    //  int ifbLast = fbList.Count - 1;
    //  var fb = ifbLast >= 0 ? fbList[0] : new FinInfoBlock { LastUpdate = DateTime.MaxValue };
    //  var fbNext = ifbLast > 0 ? fbList[1] : new FinInfoBlock { LastUpdate = DateTime.MaxValue };
    //  var tf = GetTimeFrame(security.IntervalBase, security.Interval);

    //  // Пропустить свечи по которым нет данных FinInfoBlock
    //  int iBar = 0, iFb = 0;
    //  for (; iBar < barsCount; iBar++) {
    //    var barDate = bars[iBar].Date + tf;
    //    if (barDate >= fb.LastUpdate)
    //      break;
    //    values[iBar] = new FinInfoBlock { LastUpdate = new DateTime(barDate.Ticks - 1) };
    //  }

    //  // Оставшиеся - заполнить данными FinInfoBlock
    //  for (; iBar < barsCount; iBar++) {
    //    var barDate = bars[iBar].Date + tf;
    //    while (barDate >= fbNext.LastUpdate) {
    //      fb = fbNext;
    //      if (iFb < ifbLast)
    //        fbNext = fbList[++iFb];
    //      else
    //        fbNext = new FinInfoBlock { LastUpdate = DateTime.MaxValue };
    //    }
    //    values[iBar] = fb;
    //  }
    //  return values;
    //}

    // Оптимизированная по затратам памяти версия функции 
    public void GetFininfoBlocks(ISecurity security, IContext context, string path, bool log, ActionEndOfBarFinInfo barHandler) {
      // Сохранить текущий FinInfoBlock
      WriteFininfo(security, path, false, log);

      if (barHandler == null)
        return;

      // Построить выборку нужных FinInfoBlock по барам
      var bars = security.Bars;
      int barsCount = bars.Count;
      if (barsCount <= 0)
        return;

      var tf = GetTimeFrame(security.IntervalBase, security.Interval);
      var key = GetKey(security, path);
      List<FinInfoBlock> fbList;
      lock (_dict) {
        if (!_dict.TryGetValue(key, out fbList))
          _dict[key] = fbList = new List<FinInfoBlock>(barsCount);

        // Обновить данные за левой и правой границами
        var fname = Path.Combine(path, security.Symbol + Extension);
        ReadFininfo(bars, tf, fname, fbList);
      }

      int ifbLast = fbList.Count - 1;
      var fb = ifbLast >= 0 ? fbList[0] : new FinInfoBlock { LastUpdate = DateTime.MaxValue };
      var fbNext = ifbLast > 0 ? fbList[1] : new FinInfoBlock { LastUpdate = DateTime.MaxValue };

      // Пропустить свечи по которым нет данных FinInfoBlock
      FinInfoBlock fbBar = null;
      int iBar = 0, iFb = 0;
      for (; iBar < barsCount; iBar++) {
        var barDate = bars[iBar].Date + tf;
        if (barDate >= fb.LastUpdate)
          break;
        if (fbBar == null)
          fbBar = new FinInfoBlock();
        fbBar.LastUpdate = new DateTime(barDate.Ticks - 1);
        barHandler(iBar, fbBar); // Копию fbBar создавать нет необходимости
      }

      // Оставшиеся - заполнить данными FinInfoBlock
      for (; iBar < barsCount; iBar++) {
        var barDate = bars[iBar].Date + tf;
        while (barDate >= fbNext.LastUpdate) {
          fb = fbNext;
          if (iFb < ifbLast)
            fbNext = fbList[++iFb];
          else
            fbNext = new FinInfoBlock { LastUpdate = DateTime.MaxValue };
        }
        barHandler(iBar, fb);
      }
    }

    static string GetKey(ISecurity security, string path) {
      return security.Symbol + '_' + GetTimeFrameCode(security) + '_' + path;
      //return security.Symbol;
    }

    static void ReadFininfo(IReadOnlyList<IDataBar> bars, TimeSpan tf, string fname, List<FinInfoBlock> fbList) {
      int bCount = bars.Count;
      if (bCount <= 0)
        return;

      var fbListCount = fbList.Count;
      DateTime r1Left = bars[0].Date + tf;
      DateTime r1Right = fbListCount > 0 ? fbList[0].LastUpdate : DateTime.MinValue;
      bool r1Read = r1Left < r1Right;

      DateTime r2Left = fbListCount > 0 ? fbList[fbListCount - 1].LastUpdate : r1Left;
      DateTime r2Right = bars[bCount - 1].Date + tf;
      bool r2Read = r2Left < r2Right;

      if (!r1Read && !r2Read)
        return;

      // Прочитать из файла данные для двух диапазонов времени
      //  [r1Left ... r1Right) и (r2Left ... r2Right)

      if (!File.Exists(fname))
        return;

      using (var fs = OpenAndPepareFile(fname)) {
        int recCount = (int) ((fs.Length - FinInfoBlock.HeaderSize)/FinInfoBlock.RecSize);
        if (recCount <= 0)
          return;

        int curRec = 0, finishRec = recCount - 1, iBar = 0;
        FinInfoBlock fb, fbNext, lastFb = null;
        var br = new BinaryReader(fs);

        if (r1Read) {
          curRec = FindStartingRecord(br, r1Left, curRec, finishRec, FinInfoBlock.HeaderSize, FinInfoBlock.RecSize);
          fb = FinInfoBlock.ReadFininfoBlock(br, curRec);
          if (fb.LastUpdate < r1Right) {
            if (curRec < finishRec)
              fbNext = FinInfoBlock.ReadFininfoBlock(br, ++curRec);
            else
              fbNext = new FinInfoBlock { LastUpdate = DateTime.MaxValue };

            for (; iBar < bCount; iBar++) {
              var barDate = bars[iBar].Date + tf;
              if (barDate >= fb.LastUpdate || barDate >= r1Right)
                break;
            }

            int iList = 0;
            for (; iBar < bCount; iBar++) {
              var barDate = bars[iBar].Date + tf;
              if (barDate >= r1Right)
                break;
              while (barDate >= fbNext.LastUpdate) {
                var a = fb;  fb = fbNext;  fbNext = a;
                if (curRec >= finishRec) {
                  fbNext.LastUpdate = DateTime.MaxValue;
                  break;
                }
                FinInfoBlock.ReadFininfoBlock(br, fbNext);
                ++curRec;
              }

              if (lastFb == null || !lastFb.Equals(fb)) {
                // увеличить размер при необходимости. Стандартное поведение не подходит из-за удвоения размера нового массива
                if (fbList.Count == fbList.Capacity)
                  fbList.Capacity += 512;
                fbList.Insert(iList++, lastFb = fb.Clone());
              }
            }
          }
        }

        if (r2Read) {
          curRec = FindStartingRecord(br, r2Left, curRec, finishRec, FinInfoBlock.HeaderSize, FinInfoBlock.RecSize);
          fb = FinInfoBlock.ReadFininfoBlock(br, curRec);

          if (curRec < finishRec)
            fbNext = FinInfoBlock.ReadFininfoBlock(br, ++curRec);
          else
            fbNext = new FinInfoBlock { LastUpdate = DateTime.MaxValue };

          for (; iBar < bCount; iBar++) {
            var barDate = bars[iBar].Date + tf;
            if (barDate >= fb.LastUpdate)
              break;
          }

          lastFb = fbList.Count > 0 ? fbList[fbList.Count - 1] : null;

          for (; iBar < bCount; iBar++) {
            var barDate = bars[iBar].Date + tf;
            while (barDate >= fbNext.LastUpdate) {
              var a = fb;  fb = fbNext;  fbNext = a;
              if (curRec >= finishRec) {
                fbNext.LastUpdate = DateTime.MaxValue;
                break;
              }
              FinInfoBlock.ReadFininfoBlock(br, fbNext);
              ++curRec;
            }

            if (lastFb == null || !lastFb.Equals(fb)) {
              if (fbList.Count == fbList.Capacity)
                fbList.Capacity += 512;
              fbList.Add(lastFb = fb.Clone());
            }
          }
        }
      }
    }

    static int FindStartingRecord(BinaryReader br, DateTime dateTime, int startRec, int finishRec, byte headerSize, int recSize) {
      var date = TimeZoneInfo.ConvertTime(dateTime, ExchangeTimeZone, TimeZoneInfo.Utc); // В UTC
      var fs = br.BaseStream;
      
      fs.Seek(headerSize + (long) startRec * recSize, SeekOrigin.Begin);
      var currDate = DateTime.FromBinary(br.ReadInt64());
      if (date < currDate)
        return startRec;

      fs.Seek(headerSize + (long) finishRec * recSize, SeekOrigin.Begin);
      currDate = DateTime.FromBinary(br.ReadInt64());
      if (currDate < date)
        return finishRec;

      int rec = (int) (startRec + (finishRec - startRec) * 0.66);  // Первая итерация на начало последних 33%
      do {
        fs.Seek(headerSize + (long) rec * recSize, SeekOrigin.Begin);
        currDate = DateTime.FromBinary(br.ReadInt64());
        if (currDate < date)
          startRec = rec;
        else if (currDate > date)
          finishRec = rec;
        else
          return rec;
        rec = (finishRec + startRec) / 2;
      } while (startRec + 1 < finishRec);
      return startRec;
    }

    public static void ConvertFileVersion(string srcFile) {
      // Backup
      var date = DateTime.Now.ToString("_yyyyMMdd");
      string bfname = Path.ChangeExtension(srcFile + date, ".bak");
      int i = 1;
      while (File.Exists(bfname))
        bfname = Path.ChangeExtension(srcFile + date + '_' + i++, ".bak");
      File.Move(srcFile, bfname);

      // Convert
      try {
        using (var fs = File.Open(bfname, FileMode.Open, FileAccess.Read, FileShare.ReadWrite)) {
          var fVer = fs.ReadByte();

          using (var ws = File.Open(srcFile, FileMode.Create, FileAccess.Write, FileShare.Read)) {
            ws.WriteByte(FileVer);

            var br = new BinaryReader(fs);
            using (var bw = new BinaryWriter(ws)) {
              var fb = new FinInfoBlock();
              if (fVer == 1) {
                var fb1 = new FininfoBlock1();
                while (fs.Position < fs.Length && FininfoBlock1.ReadRecord(br, fb1))
                  fb.Set(fb1).WriteRecord(bw);

              } else if (fVer == 2) {
                var fb2 = new FininfoBlock2();
                while (fs.Position < fs.Length && FininfoBlock2.ReadRecord(br, fb2))
                  fb.Set(fb2).WriteRecord(bw);

              } else if (fVer == 3) {
                var fb3 = new FinInfoBlock3();
                while (fs.Position < fs.Length && FinInfoBlock3.ReadRecord(br, fb3))
                  fb.Set(fb3).WriteRecord(bw);
              }
            }
          }
        }
      } catch(EndOfStreamException) {}  // Не до конца записанный файл. Игнорировать
    }

    static bool FinInfoContainsData(FinInfo fi) {
      const double e = 1e-6;
      return Math.Abs(fi.Ask ?? 0) >= e ||
             Math.Abs(fi.AvgPrice ?? 0) >= e ||
             Math.Abs(fi.Bid ?? 0) >= e ||
             Math.Abs(fi.BuyCount ?? 0) >= e ||
             Math.Abs(fi.BuyDeposit ?? 0) >= e ||
             Math.Abs(fi.BuySqty ?? 0) >= e ||
             Math.Abs(fi.Change ?? 0) >= e ||
             Math.Abs(fi.ChangePercent ?? 0) >= e ||
             Math.Abs(fi.ClosePrice ?? 0) >= e ||
             Math.Abs(fi.HighBid ?? 0) >= e ||
             //fi.IsTradingEnabled !=
             Math.Abs(fi.LastPrice ?? 0) >= e ||
             Math.Abs(fi.LowOffer ?? 0) >= e ||
             Math.Abs(fi.MaxDeal ?? 0) >= e ||
             Math.Abs(fi.MinDeal ?? 0) >= e ||
             Math.Abs(fi.NoDeal ?? 0) >= e ||
             Math.Abs(fi.OpenInterest ?? 0) >= e ||
             Math.Abs(fi.OpenPrice ?? 0) >= e ||
             Math.Abs(fi.PrevPrice ?? 0) >= e ||
             Math.Abs(fi.Price2PrevAvg ?? 0) >= e ||
             Math.Abs(fi.PriceMax ?? 0) >= e ||
             Math.Abs(fi.PriceMin ?? 0) >= e ||
             Math.Abs(fi.SellCount ?? 0) >= e ||
             Math.Abs(fi.SellDeposit ?? 0) >= e ||
             Math.Abs(fi.SellSqty ?? 0) >= e ||
             Math.Abs(fi.StepPrice ?? 0) >= e ||
             Math.Abs(fi.SumQty ?? 0) >= e ||
             Math.Abs(fi.SumVolume ?? 0) >= e ||
             Math.Abs(fi.TheoreticalPrice ?? 0) >= e ||
             Math.Abs(fi.Volatility ?? 0) >= e;
      		 
    }

    static string GetTimeFrameCode(ISecurity security) {
      string tf = security.Interval.ToString();
      switch (security.IntervalBase) {
        case DataIntervals.DAYS:
          tf += 'd';
          break;
        case DataIntervals.MINUTE:
          tf += 'm';
          break;
        case DataIntervals.SECONDS:
          tf += 's';
          break;
        case DataIntervals.TICK:
          tf += 't';
          break;
      }
      return tf;
    }

    static TimeSpan GetTimeFrame(DataIntervals intervalBase, int interval) {
      TimeSpan tf;
      switch (intervalBase) {
        case DataIntervals.DAYS:
          tf = new TimeSpan(interval, 0, 0, 0);
          break;
        case DataIntervals.MINUTE:
          tf = new TimeSpan(0, 0, interval, 0);
          break;
        case DataIntervals.SECONDS:
          tf = new TimeSpan(0, 0, 0, interval);
          break;
        case DataIntervals.TICK:
          tf = new TimeSpan(0, 0, 0, 0, interval);
          break;
        default:
          throw new ArgumentException("Неизвестный DataIntervals в GetTimeFrame");
      }
      return tf;
    }
  }

  class FininfoBlock1 {
    public DateTime LastUpdate { get; set; }
    public double Ask { get; set; }
    public double Bid { get; set; }
    public double BuyCount { get; set; }
    public double BuySqty { get; set; }
    public double SellCount { get; set; }
    public double SellSqty { get; set; }
    public double TheoriticPrice { get; set; }
    public double Volatility { get; set; }

    public const int HeaderSize = sizeof(byte);
    public const int RecSize = sizeof(long) + 8 * sizeof(double);

    public static bool ReadRecord(BinaryReader br, FininfoBlock1 fb) {
      if (br.BaseStream.Length - br.BaseStream.Position < RecSize)
        return false;
      fb.LastUpdate = TimeZoneInfo.ConvertTime(DateTime.FromBinary(br.ReadInt64()), TimeZoneInfo.Utc, FinInfoCache.ExchangeTimeZone);
      fb.Ask = br.ReadDouble();
      fb.Bid = br.ReadDouble();
      fb.BuyCount = br.ReadDouble();
      fb.BuySqty = br.ReadDouble();
      fb.SellCount = br.ReadDouble();
      fb.SellSqty = br.ReadDouble();
      fb.TheoriticPrice = br.ReadDouble();
      fb.Volatility = br.ReadDouble();
      return true;
    }
  }

  class FininfoBlock2 {
    public DateTime LastUpdate { get; set; }
    public float Ask { get; set; }
    public float Bid { get; set; }
    public float BuyCount { get; set; }
    public float BuySqty { get; set; }
    public float SellCount { get; set; }
    public float SellSqty { get; set; }
    public float TheoriticPrice { get; set; }
    public float Volatility { get; set; }

    public const int HeaderSize = sizeof(byte);
    public const int RecSize = sizeof(long) + 8 * sizeof(float);

    public static bool ReadRecord(BinaryReader br, FininfoBlock2 fb) {
      if (br.BaseStream.Length - br.BaseStream.Position < RecSize)
        return false;
      fb.LastUpdate = TimeZoneInfo.ConvertTime(DateTime.FromBinary(br.ReadInt64()), TimeZoneInfo.Utc, FinInfoCache.ExchangeTimeZone);
      fb.Ask = br.ReadSingle();
      fb.Bid = br.ReadSingle();
      fb.BuyCount = br.ReadSingle();
      fb.BuySqty = br.ReadSingle();
      fb.SellCount = br.ReadSingle();
      fb.SellSqty = br.ReadSingle();
      fb.TheoriticPrice = br.ReadSingle();
      fb.Volatility = br.ReadSingle();
      return true;
    }
  }

  class FinInfoBlock3 {
    public DateTime LastUpdate { get; set; }
    public DateTime LocalDate { get; set; }
    public float Ask { get; set; }
    public float AvgPrice { get; set; }
    public float Bid { get; set; }
    public float BuyCount { get; set; }
    public float BuyDeposit { get; set; }
    public float BuySqty { get; set; }
    public float Change { get; set; }
    public float ChangePercent { get; set; }
    public float ClosePrice { get; set; }
    public float HighBid { get; set; }
    public bool IsTradingEnabled { get; set; }
    public float LastPrice { get; set; }
    public float LowOffer { get; set; }
    public float MaxDeal { get; set; }
    public float MinDeal { get; set; }
    public float NoDeal { get; set; }
    public float OpenPrice { get; set; }
    public float PrevPrice { get; set; }
    public float Price2PrevAvg { get; set; }
    public float PriceMax { get; set; }
    public float PriceMin { get; set; }
    public float SellCount { get; set; }
    public float SellDeposit { get; set; }
    public float SellSqty { get; set; }
    public float StepPrice { get; set; }
    public float SumQty { get; set; }
    public float SumVolume { get; set; }
    public float TheoreticPrice { get; set; }
    public float Volatility { get; set; }

    public const int HeaderSize = sizeof(byte);
    public const int RecSize = 2*sizeof(long) + 28*sizeof(float) + sizeof(bool); // По числу сохраняемых параметров

    public static bool ReadRecord(BinaryReader br, FinInfoBlock3 fb) {
      if (br.BaseStream.Length - br.BaseStream.Position < RecSize)
        return false;
      fb.LastUpdate = TimeZoneInfo.ConvertTime(DateTime.FromBinary(br.ReadInt64()), TimeZoneInfo.Utc, FinInfoCache.ExchangeTimeZone);
      fb.LocalDate = TimeZoneInfo.ConvertTime(DateTime.FromBinary(br.ReadInt64()), TimeZoneInfo.Utc, FinInfoCache.ExchangeTimeZone);
      fb.Ask = br.ReadSingle();
      fb.AvgPrice = br.ReadSingle();
      fb.Bid = br.ReadSingle();
      fb.BuyCount = br.ReadSingle();
      fb.BuyDeposit = br.ReadSingle();
      fb.BuySqty = br.ReadSingle();
      fb.Change = br.ReadSingle();
      fb.ChangePercent = br.ReadSingle();
      fb.ClosePrice = br.ReadSingle();
      fb.HighBid = br.ReadSingle();
      fb.IsTradingEnabled = br.ReadBoolean();
      fb.LastPrice = br.ReadSingle();
      fb.LowOffer = br.ReadSingle();
      fb.MaxDeal = br.ReadSingle();
      fb.MinDeal = br.ReadSingle();
      fb.NoDeal = br.ReadSingle();
      fb.OpenPrice = br.ReadSingle();
      fb.PrevPrice = br.ReadSingle();
      fb.Price2PrevAvg = br.ReadSingle();
      fb.PriceMax = br.ReadSingle();
      fb.PriceMin = br.ReadSingle();
      fb.SellCount = br.ReadSingle();
      fb.SellDeposit = br.ReadSingle();
      fb.SellSqty = br.ReadSingle();
      fb.StepPrice = br.ReadSingle();
      fb.SumQty = br.ReadSingle();
      fb.SumVolume = br.ReadSingle();
      fb.TheoreticPrice = br.ReadSingle();
      fb.Volatility = br.ReadSingle();
      return true;
    }
  }

  class FinInfoBlock {
    /// <summary> Время последнего обновления (серверное), биржевое время <see cref="FinInfoCache.ExchangeTimeZone"/> </summary>
    public DateTime LastUpdate { get; set; }
    /// <summary> Локальное время получения данных, приведенное к биржевому часовому поясу <see cref="FinInfoCache.ExchangeTimeZone"/> </summary>
    public DateTime LocalDate { get; set; }
    /// <summary> Время последней сделки </summary>
    public DateTime LastTradeUpdate { get; set; }
    /// <summary> Лучшая (минимальная) цена продажи </summary>
    public float Ask { get; set; }
    /// <summary> Средняя цена за сессию </summary>
    public float AvgPrice { get; set; }
    /// <summary> Лучшая (максимальная) цена покупки </summary>
    public float Bid { get; set; }
    /// <summary> Количество заявок на покупку в очереди торговой системы </summary>
    public float BuyCount { get; set; }
    /// <summary> Гарантийные обязательства покупателя, в рублях </summary>
    public float BuyDeposit { get; set; }
    /// <summary> Объем всех заявок на покупку в очереди торговой системы, выраженный в единицах ценных бумаг </summary>
    public float BuySqty { get; set; }
    /// <summary> Изменение (абсолютное) </summary>
    public float Change { get; set; }
    /// <summary> Изменение (процентное) </summary>
    public float ChangePercent { get; set; }
    /// <summary> Цена закрытия сессии (прошлой) </summary>
    public float ClosePrice { get; set; }
    /// <summary> Наибольшая цена спроса в течение торговой сессии </summary>
    public float HighBid { get; set; }
    /// <summary> Торги по инструменту разрешены или нет </summary>
    public bool IsTradingEnabled { get; set; }
    /// <summary> Цена последней сделки </summary>
    public float LastPrice { get; set; }
    /// <summary> Наименьшая цена предложения в течение торговой сессии </summary>
    public float LowOffer { get; set; }
    /// <summary> Максимальная цена сделки за сессию </summary>
    public float MaxDeal { get; set; }
    /// <summary> Минимальная цена сделки за сессию </summary>
    public float MinDeal { get; set; }
    /// <summary> Количество сделок за торговый день </summary>
    public float NoDeal { get; set; }
    /// <summary> Открытый интерес </summary>
    public float OpenInterest { get; set; }
    /// <summary> Цена открытия сессии </summary>
    public float OpenPrice { get; set; }
    /// <summary> Цена последней сделки предыдущего торгового дня </summary>
    public float PrevPrice { get; set; }
    /// <summary> Цена последней сделки к оценке предыдущего дня </summary>
    public float Price2PrevAvg { get; set; }
    /// <summary> Верхний лимит цены </summary>
    public float PriceMax { get; set; }
    /// <summary> Нижний лимит цены </summary>
    public float PriceMin { get; set; }
    /// <summary> Количество заявок на продажу в очереди торговой системы </summary>
    public float SellCount { get; set; }
    /// <summary> Гарантийные обязательства продавца, в рублях </summary>
    public float SellDeposit { get; set; }
    /// <summary> Объем всех заявок на продажу в очереди торговой системы, выраженный в единицах ценных бумаг </summary>
    public float SellSqty { get; set; }
    //Состояние торговой сессии по инструменту 
    //public string StatusDescription { get; set; }
    /// <summary> Стоимость шага цены </summary>
    public float StepPrice { get; set; }
    /// <summary> Объем совершенных сделок, выраженный в единицах ценных бумаг </summary>
    public float SumQty { get; set; }
    /// <summary> Объем совершенных сделок, выраженный в рублях. </summary>
    public float SumVolume { get; set; }
    /// <summary> Теоретическая (Расчетная) цена опциона </summary>
    public float TheoreticPrice { get; set; }
    /// <summary> Волатильность финансового инструмента, % </summary>
    public float Volatility { get; set; }

    /// <summary> Объем по лучшаей цене продажи </summary>
    public float AskVolume { get; set; }
    /// <summary> Объем по лучшаей цене покупки </summary>
    public float BidVolume { get; set; }

    /// <summary> Цена из стакана наибольшего объема на продажу </summary>
    public float PriceOfHiAskVolume { get; set; }
    /// <summary> Цена из стакана наибольшего объема на покупку </summary>
    public float PriceOfHiBidVolume { get; set; }
    /// <summary> Наибольший объем из стакана на продажу </summary>
    public float VolumeOfHiAskVolume { get; set; }
    /// <summary> Наибольший объем из стакана на покупку </summary>
    public float VolumeOfHiBidVolume { get; set; }
    /// <summary> Весь объем из стакана на продажу </summary>
    public float AllAskVolume { get; set; }
    /// <summary> Весь объем из стакана на покупку </summary>
    public float AllBidVolume { get; set; }

    /// <summary> Дельта заявок </summary>
    public float DeltaCount { get { return BuyCount - SellCount; } }
    /// <summary> Дельта объема </summary>
    public float DeltaSqty { get { return BuySqty - SellSqty; } }
    /// <summary> Средний объем заявки на покупку </summary>
    public float AvgBuy { get { return BuyCount != 0 ? BuySqty/BuyCount : 0; } }
    /// <summary> Средний объем заявки на продажу </summary>
    public float AvgSell { get { return SellCount != 0 ? SellSqty/SellCount : 0; } }
    /// <summary> Весь объем из стакана </summary>
    public float AllDomVolume { get { return AllAskVolume + AllBidVolume; } }
    /// <summary> Дельта объема из стакана </summary>
    public float DeltaVolume { get { return AllAskVolume - AllBidVolume; } }
    
    //Тик закрытого бара
    public float LastBarClosed { get; set; }
    
    // Здесь добовляем то что нам нужно (STBP_0) Stakan Tik Buy Price - //summary
    /// <summary> Stakan Tik Buy Price </summary>
// _0    
    public float STBP_0 { get; set; }
    public float STBQ_0 { get; set; }
    public float STSP_0 { get; set; }
    public float STSQ_0 { get; set; }
    
// _1
    public float STBP_1 { get; set; }
    public float STBQ_1 { get; set; }
    public float STSP_1 { get; set; }
    public float STSQ_1 { get; set; }
    
// _2
    public float STBP_2 { get; set; }
    public float STBQ_2 { get; set; }
    public float STSP_2 { get; set; }
    public float STSQ_2 { get; set; }
    
// _3
    public float STBP_3 { get; set; }
    public float STBQ_3 { get; set; }
    public float STSP_3 { get; set; }
    public float STSQ_3 { get; set; }
    
// _4
    public float STBP_4 { get; set; }
    public float STBQ_4 { get; set; }
    public float STSP_4 { get; set; }
    public float STSQ_4 { get; set; }
    
// _5
    public float STBP_5 { get; set; }
    public float STBQ_5 { get; set; }
    public float STSP_5 { get; set; }
    public float STSQ_5 { get; set; }
    
// _6
    public float STBP_6 { get; set; }
    public float STBQ_6 { get; set; }
    public float STSP_6 { get; set; }
    public float STSQ_6 { get; set; }
    
// _7
    public float STBP_7 { get; set; }
    public float STBQ_7 { get; set; }
    public float STSP_7 { get; set; }
    public float STSQ_7 { get; set; }
    
// _8
    public float STBP_8 { get; set; }
    public float STBQ_8 { get; set; }
    public float STSP_8 { get; set; }
    public float STSQ_8 { get; set; }
    
// _9
    public float STBP_9 { get; set; }
    public float STBQ_9 { get; set; }
    public float STSP_9 { get; set; }
    public float STSQ_9 { get; set; }
    
// _10
    public float STBP_10 { get; set; }
    public float STBQ_10 { get; set; }
    public float STSP_10 { get; set; }
    public float STSQ_10 { get; set; }
    
// _11
    public float STBP_11 { get; set; }
    public float STBQ_11 { get; set; }
    public float STSP_11 { get; set; }
    public float STSQ_11 { get; set; }
    
// _12
    public float STBP_12 { get; set; }
    public float STBQ_12 { get; set; }
    public float STSP_12 { get; set; }
    public float STSQ_12 { get; set; }
    
// _13
    public float STBP_13 { get; set; }
    public float STBQ_13 { get; set; }
    public float STSP_13 { get; set; }
    public float STSQ_13 { get; set; }
    
// _14
    public float STBP_14 { get; set; }
    public float STBQ_14 { get; set; }
    public float STSP_14 { get; set; }
    public float STSQ_14 { get; set; }
    
// _15
    public float STBP_15 { get; set; }
    public float STBQ_15 { get; set; }
    public float STSP_15 { get; set; }
    public float STSQ_15 { get; set; }
    
// _16
    public float STBP_16 { get; set; }
    public float STBQ_16 { get; set; }
    public float STSP_16 { get; set; }
    public float STSQ_16 { get; set; }
    
// _17
    public float STBP_17 { get; set; }
    public float STBQ_17 { get; set; }
    public float STSP_17 { get; set; }
    public float STSQ_17 { get; set; }
    
// _18
    public float STBP_18 { get; set; }
    public float STBQ_18 { get; set; }
    public float STSP_18 { get; set; }
    public float STSQ_18 { get; set; }
    
// _19
    public float STBP_19 { get; set; }
    public float STBQ_19 { get; set; }
    public float STSP_19 { get; set; }
    public float STSQ_19 { get; set; }
    
// _20
    public float STBP_20 { get; set; }
    public float STBQ_20 { get; set; }
    public float STSP_20 { get; set; }
    public float STSQ_20 { get; set; }
    
// _21
    public float STBP_21 { get; set; }
    public float STBQ_21 { get; set; }
    public float STSP_21 { get; set; }
    public float STSQ_21 { get; set; }
    
// _22
    public float STBP_22 { get; set; }
    public float STBQ_22 { get; set; }
    public float STSP_22 { get; set; }
    public float STSQ_22 { get; set; }
    
// _23
    public float STBP_23 { get; set; }
    public float STBQ_23 { get; set; }
    public float STSP_23 { get; set; }
    public float STSQ_23 { get; set; }
    
// _24
    public float STBP_24 { get; set; }
    public float STBQ_24 { get; set; }
    public float STSP_24 { get; set; }
    public float STSQ_24 { get; set; }
    
// _25
    public float STBP_25 { get; set; }
    public float STBQ_25 { get; set; }
    public float STSP_25 { get; set; }
    public float STSQ_25 { get; set; }
    
// _26
    public float STBP_26 { get; set; }
    public float STBQ_26 { get; set; }
    public float STSP_26 { get; set; }
    public float STSQ_26 { get; set; }
    
// _27
    public float STBP_27 { get; set; }
    public float STBQ_27 { get; set; }
    public float STSP_27 { get; set; }
    public float STSQ_27 { get; set; }
    
// _28
    public float STBP_28 { get; set; }
    public float STBQ_28 { get; set; }
    public float STSP_28 { get; set; }
    public float STSQ_28 { get; set; }
    
// _29
    public float STBP_29 { get; set; }
    public float STBQ_29 { get; set; }
    public float STSP_29 { get; set; }
    public float STSQ_29 { get; set; }
    
// _30
    public float STBP_30 { get; set; }
    public float STBQ_30 { get; set; }
    public float STSP_30 { get; set; }
    public float STSQ_30 { get; set; }
   
// _31
    public float STBP_31 { get; set; }
    public float STBQ_31 { get; set; }
    public float STSP_31 { get; set; }
    public float STSQ_31 { get; set; }
    
// _32
    public float STBP_32 { get; set; }
    public float STBQ_32 { get; set; }
    public float STSP_32 { get; set; }
    public float STSQ_32 { get; set; }
    
// _33
    public float STBP_33 { get; set; }
    public float STBQ_33 { get; set; }
    public float STSP_33 { get; set; }
    public float STSQ_33 { get; set; }
    
// _34
    public float STBP_34 { get; set; }
    public float STBQ_34 { get; set; }
    public float STSP_34 { get; set; }
    public float STSQ_34 { get; set; }
    
// _35
    public float STBP_35 { get; set; }
    public float STBQ_35 { get; set; }
    public float STSP_35 { get; set; }
    public float STSQ_35 { get; set; }
    
// _36
    public float STBP_36 { get; set; }
    public float STBQ_36 { get; set; }
    public float STSP_36 { get; set; }
    public float STSQ_36 { get; set; }
    
// _37
    public float STBP_37 { get; set; }
    public float STBQ_37 { get; set; }
    public float STSP_37 { get; set; }
    public float STSQ_37 { get; set; }
    
// _38
    public float STBP_38 { get; set; }
    public float STBQ_38 { get; set; }
    public float STSP_38 { get; set; }
    public float STSQ_38 { get; set; }
    
// _39
    public float STBP_39 { get; set; }
    public float STBQ_39 { get; set; }
    public float STSP_39 { get; set; }
    public float STSQ_39 { get; set; }
    
// _40
    public float STBP_40 { get; set; }
    public float STBQ_40 { get; set; }
    public float STSP_40 { get; set; }
    public float STSQ_40 { get; set; }
    
    
// _41
    public float STBP_41 { get; set; }
    public float STBQ_41 { get; set; }
    public float STSP_41 { get; set; }
    public float STSQ_41 { get; set; }
    
// _42
    public float STBP_42 { get; set; }
    public float STBQ_42 { get; set; }
    public float STSP_42 { get; set; }
    public float STSQ_42 { get; set; }
    
// _43
    public float STBP_43 { get; set; }
    public float STBQ_43 { get; set; }
    public float STSP_43 { get; set; }
    public float STSQ_43 { get; set; }
    
// _44
    public float STBP_44 { get; set; }
    public float STBQ_44 { get; set; }
    public float STSP_44 { get; set; }
    public float STSQ_44 { get; set; }
    
// _45
    public float STBP_45 { get; set; }
    public float STBQ_45 { get; set; }
    public float STSP_45 { get; set; }
    public float STSQ_45 { get; set; }
    
// _46
    public float STBP_46 { get; set; }
    public float STBQ_46 { get; set; }
    public float STSP_46 { get; set; }
    public float STSQ_46 { get; set; }
    
// _47
    public float STBP_47 { get; set; }
    public float STBQ_47 { get; set; }
    public float STSP_47 { get; set; }
    public float STSQ_47 { get; set; }
    
// _48
    public float STBP_48 { get; set; }
    public float STBQ_48 { get; set; }
    public float STSP_48 { get; set; }
    public float STSQ_48 { get; set; }
    
// _49
    public float STBP_49 { get; set; }
    public float STBQ_49 { get; set; }
    public float STSP_49 { get; set; }
    public float STSQ_49 { get; set; }

    public const int HeaderSize = sizeof(byte);
    public const int RecSize = 3*sizeof(long) + 238 * sizeof(float) + sizeof(bool);  // По числу сохраняемых параметров

    public FinInfoBlock() { }

    public FinInfoBlock(FinInfo fi) {
      LastUpdate = new DateTime(fi.LastUpdate.Ticks); // Копия - специально
      LocalDate = TimeZoneInfo.ConvertTime(DateTime.UtcNow, TimeZoneInfo.Utc, FinInfoCache.ExchangeTimeZone);  // Из UTC в биржевое время
      LastTradeUpdate = new DateTime(fi.LastTradeUpdate.HasValue ? fi.LastTradeUpdate.Value.Ticks : 0);
      Ask = (float) (fi.Ask ?? 0);
      AvgPrice = (float) (fi.AvgPrice ?? 0);
      Bid = (float) (fi.Bid ?? 0);
      BuyCount = (float) (fi.BuyCount ?? 0);
      BuyDeposit = (float) (fi.BuyDeposit ?? 0);
      BuySqty = (float) (fi.BuySqty ?? 0);
      Change = (float) (fi.Change ?? 0);
      ChangePercent = (float) (fi.ChangePercent ?? 0);
      ClosePrice = (float) (fi.ClosePrice ?? 0);
      HighBid = (float) (fi.HighBid ?? 0);
      IsTradingEnabled = fi.IsTradingEnabled;
      LastPrice = (float) (fi.LastPrice ?? 0);
      LowOffer = (float) (fi.LowOffer ?? 0);
      MaxDeal = (float) (fi.MaxDeal ?? 0);
      MinDeal = (float) (fi.MinDeal ?? 0);
      NoDeal = (long) (fi.NoDeal ?? 0);
      OpenInterest = (long) (fi.OpenInterest ?? 0);
      OpenPrice = (float) (fi.OpenPrice ?? 0);
      PrevPrice = (float) (fi.PrevPrice ?? 0);
      Price2PrevAvg = (float) (fi.Price2PrevAvg ?? 0);
      PriceMax = (float) (fi.PriceMax ?? 0);
      PriceMin = (float) (fi.PriceMin ?? 0);
      SellCount = (float) (fi.SellCount ?? 0);
      SellDeposit = (float) (fi.SellDeposit ?? 0);
      SellSqty = (float) (fi.SellSqty ?? 0);
      StepPrice = (float) (fi.StepPrice ?? 0);
      SumQty = (float) (fi.SumQty ?? 0);
      SumVolume = (float) (fi.SumVolume ?? 0);
      TheoreticPrice = (float) (fi.TheoreticalPrice ?? 0);
      Volatility = (float) (fi.Volatility ?? 0);
      //AskVolume
      //BidVolume
      //PriceOfHiAskVolume 
      //PriceOfHiBidVolume 
      //VolumeOfHiAskVolume
      //VolumeOfHiBidVolume
      //AllAskVolume 
      //AllBidVolume 
      
      //Тик закрытого бара
      //LastBarClosed
      // Здесь добовляем то что нам нужно (STBP_0) Stakan Tik Buy Price - //
      //STBP_0
      //STBQ_0
      //STSP_0
      //STSQ_0
      //STBP_1
      //STBQ_1
      //STSP_1
      //STSQ_1
      //STBP_2 - STBP_49
    }

    public FinInfoBlock Set(FininfoBlock1 fb) {
      LocalDate = LastUpdate = new DateTime(fb.LastUpdate.Ticks); // Копия - специально
      LastTradeUpdate = new DateTime();
      Ask = (float) fb.Ask;
      AvgPrice = 0;
      Bid = (float) fb.Bid;
      BuyCount = (float) fb.BuyCount;
      BuyDeposit = 0;
      BuySqty = (float) fb.BuySqty;
      Change = 0;
      ChangePercent = 0;
      ClosePrice = 0;
      HighBid = 0;
      IsTradingEnabled = true;
      LastPrice = 0;
      LowOffer = 0;
      MaxDeal = 0;
      MinDeal = 0;
      NoDeal = 0;
      OpenInterest = 0;
      OpenPrice = 0;
      PrevPrice = 0;
      Price2PrevAvg = 0;
      PriceMax = 0;
      PriceMin = 0;
      SellCount = (float) fb.SellCount;
      SellDeposit = 0;
      SellSqty = (float) fb.SellSqty;
      StepPrice = 0;
      SumQty = 0;
      SumVolume = 0;
      TheoreticPrice = (float) fb.TheoriticPrice;
      Volatility = (float) fb.Volatility;
      AskVolume = 0;
      BidVolume = 0;
      PriceOfHiAskVolume = 0;
      PriceOfHiBidVolume = 0;
      VolumeOfHiAskVolume = 0;
      VolumeOfHiBidVolume = 0;
      AllAskVolume = 0;
      AllBidVolume = 0;
      
      LastBarClosed = 0;
      // Здесь добовляем то что нам нужно (STBP_0) Stakan Tik Buy Price - 0
      STBP_0 = 0;
      STBQ_0 = 0;
      STSP_0 = 0;
      STSQ_0 = 0;
      STBP_1 = 0;
      STBQ_1 = 0;
      STSP_1 = 0;
      STSQ_1 = 0;
      STBP_2 = 0;
      STBQ_2 = 0;
      STSP_2 = 0;
      STSQ_2 = 0;
      STBP_3 = 0;
      STBQ_3 = 0;
      STSP_3 = 0;
      STSQ_3 = 0;
      STBP_4 = 0;
      STBQ_4 = 0;
      STSP_4 = 0;
      STSQ_4 = 0;
      STBP_5 = 0;
      STBQ_5 = 0;
      STSP_5 = 0;
      STSQ_5 = 0;
      STBP_6 = 0;
      STBQ_6 = 0;
      STSP_6 = 0;
      STSQ_6 = 0;
      STBP_7 = 0;
      STBQ_7 = 0;
      STSP_7 = 0;
      STSQ_7 = 0;
      STBP_8 = 0;
      STBQ_8 = 0;
      STSP_8 = 0;
      STSQ_8 = 0;
      STBP_9 = 0;
      STBQ_9 = 0;
      STSP_9 = 0;
      STSQ_9 = 0;
      STBP_10 = 0;
      STBQ_10 = 0;
      STSP_10 = 0;
      STSQ_10 = 0;
      STBP_11 = 0;
      STBQ_11 = 0;
      STSP_11 = 0;
      STSQ_11 = 0;
      STBP_12 = 0;
      STBQ_12 = 0;
      STSP_12 = 0;
      STSQ_12 = 0;
      STBP_13 = 0;
      STBQ_13 = 0;
      STSP_13 = 0;
      STSQ_13 = 0;
      STBP_14 = 0;
      STBQ_14 = 0;
      STSP_14 = 0;
      STSQ_14 = 0;
      STBP_15 = 0;
      STBQ_15 = 0;
      STSP_15 = 0;
      STSQ_15 = 0;
      STBP_16 = 0;
      STBQ_16 = 0;
      STSP_16 = 0;
      STSQ_16 = 0;
      STBP_17 = 0;
      STBQ_17 = 0;
      STSP_17 = 0;
      STSQ_17 = 0;
      STBP_18 = 0;
      STBQ_18 = 0;
      STSP_18 = 0;
      STSQ_18 = 0;
      STBP_19 = 0;
      STBQ_19 = 0;
      STSP_19 = 0;
      STSQ_19 = 0;
      STBP_20 = 0;
      STBQ_20 = 0;
      STSP_20 = 0;
      STSQ_20 = 0;
      STBP_21 = 0;
      STBQ_21 = 0;
      STSP_21 = 0;
      STSQ_21 = 0;
      STBP_22 = 0;
      STBQ_22 = 0;
      STSP_22 = 0;
      STSQ_22 = 0;
      STBP_23 = 0;
      STBQ_23 = 0;
      STSP_23 = 0;
      STSQ_23 = 0;
      STBP_24 = 0;
      STBQ_24 = 0;
      STSP_24 = 0;
      STSQ_24 = 0;
      STBP_25 = 0;
      STBQ_25 = 0;
      STSP_25 = 0;
      STSQ_25 = 0;
      STBP_26 = 0;
      STBQ_26 = 0;
      STSP_26 = 0;
      STSQ_26 = 0;
      STBP_27 = 0;
      STBQ_27 = 0;
      STSP_27 = 0;
      STSQ_27 = 0;
      STBP_28 = 0;
      STBQ_28 = 0;
      STSP_28 = 0;
      STSQ_28 = 0;
      STBP_29 = 0;
      STBQ_29 = 0;
      STSP_29 = 0;
      STSQ_29 = 0;
      STBP_30 = 0;
      STBQ_30 = 0;
      STSP_30 = 0;
      STSQ_30 = 0;
      STBP_31 = 0;
      STBQ_31 = 0;
      STSP_31 = 0;
      STSQ_31 = 0;
      STBP_32 = 0;
      STBQ_32 = 0;
      STSP_32 = 0;
      STSQ_32 = 0;
      STBP_33 = 0;
      STBQ_33 = 0;
      STSP_33 = 0;
      STSQ_33 = 0;
      STBP_34 = 0;
      STBQ_34 = 0;
      STSP_34 = 0;
      STSQ_34 = 0;
      STBP_35 = 0;
      STBQ_35 = 0;
      STSP_35 = 0;
      STSQ_35 = 0;
      STBP_36 = 0;
      STBQ_36 = 0;
      STSP_36 = 0;
      STSQ_36 = 0;
      STBP_37 = 0;
      STBQ_37 = 0;
      STSP_37 = 0;
      STSQ_37 = 0;
      STBP_38 = 0;
      STBQ_38 = 0;
      STSP_38 = 0;
      STSQ_38 = 0;
      STBP_39 = 0;
      STBQ_39 = 0;
      STSP_39 = 0;
      STSQ_39 = 0;
      STBP_40 = 0;
      STBQ_40 = 0;
      STSP_40 = 0;
      STSQ_40 = 0;
      STBP_41 = 0;
      STBQ_41 = 0;
      STSP_41 = 0;
      STSQ_41 = 0;
      STBP_42 = 0;
      STBQ_42 = 0;
      STSP_42 = 0;
      STSQ_42 = 0;
      STBP_43 = 0;
      STBQ_43 = 0;
      STSP_43 = 0;
      STSQ_43 = 0;
      STBP_44 = 0;
      STBQ_44 = 0;
      STSP_44 = 0;
      STSQ_44 = 0;
      STBP_45 = 0;
      STBQ_45 = 0;
      STSP_45 = 0;
      STSQ_45 = 0;
      STBP_46 = 0;
      STBQ_46 = 0;
      STSP_46 = 0;
      STSQ_46 = 0;
      STBP_47 = 0;
      STBQ_47 = 0;
      STSP_47 = 0;
      STSQ_47 = 0;
      STBP_48 = 0;
      STBQ_48 = 0;
      STSP_48 = 0;
      STSQ_48 = 0;
      STBP_49 = 0;
      STBQ_49 = 0;
      STSP_49 = 0;
      STSQ_49 = 0;
      return this;
    }

    public FinInfoBlock Set(FininfoBlock2 fb) {
      LocalDate = LastUpdate = new DateTime(fb.LastUpdate.Ticks); // Копия - специально
      LastTradeUpdate = new DateTime();
      Ask = fb.Ask;
      AvgPrice = 0;
      Bid = fb.Bid;
      BuyCount = fb.BuyCount;
      BuyDeposit = 0;
      BuySqty = fb.BuySqty;
      Change = 0;
      ChangePercent = 0;
      ClosePrice = 0;
      HighBid = 0;
      IsTradingEnabled = true;
      LastPrice = 0;
      LowOffer = 0;
      MaxDeal = 0;
      MinDeal = 0;
      NoDeal = 0;
      OpenInterest = 0;
      OpenPrice = 0;
      PrevPrice = 0;
      Price2PrevAvg = 0;
      PriceMax = 0;
      PriceMin = 0;
      SellCount = fb.SellCount;
      SellDeposit = 0;
      SellSqty = fb.SellSqty;
      StepPrice = 0;
      SumQty = 0;
      SumVolume = 0;
      TheoreticPrice = fb.TheoriticPrice;
      Volatility = fb.Volatility;
      AskVolume = 0;
      BidVolume = 0;
      PriceOfHiAskVolume = 0;
      PriceOfHiBidVolume = 0; 
      VolumeOfHiAskVolume = 0;
      VolumeOfHiBidVolume = 0;
      AllAskVolume = 0;
      AllBidVolume = 0;
      
      LastBarClosed = 0;
      // Здесь добовляем то что нам нужно (STBP_0) Stakan Tik Buy Price - 0
      STBP_0 = 0;
      STBQ_0 = 0;
      STSP_0 = 0;
      STSQ_0 = 0;
      STBP_1 = 0;
      STBQ_1 = 0;
      STSP_1 = 0;
      STSQ_1 = 0;
      STBP_2 = 0;
      STBQ_2 = 0;
      STSP_2 = 0;
      STSQ_2 = 0;
      STBP_3 = 0;
      STBQ_3 = 0;
      STSP_3 = 0;
      STSQ_3 = 0;
      STBP_4 = 0;
      STBQ_4 = 0;
      STSP_4 = 0;
      STSQ_4 = 0;
      STBP_5 = 0;
      STBQ_5 = 0;
      STSP_5 = 0;
      STSQ_5 = 0;
      STBP_6 = 0;
      STBQ_6 = 0;
      STSP_6 = 0;
      STSQ_6 = 0;
      STBP_7 = 0;
      STBQ_7 = 0;
      STSP_7 = 0;
      STSQ_7 = 0;
      STBP_8 = 0;
      STBQ_8 = 0;
      STSP_8 = 0;
      STSQ_8 = 0;
      STBP_9 = 0;
      STBQ_9 = 0;
      STSP_9 = 0;
      STSQ_9 = 0;
      STBP_10 = 0;
      STBQ_10 = 0;
      STSP_10 = 0;
      STSQ_10 = 0;
      STBP_11 = 0;
      STBQ_11 = 0;
      STSP_11 = 0;
      STSQ_11 = 0;
      STBP_12 = 0;
      STBQ_12 = 0;
      STSP_12 = 0;
      STSQ_12 = 0;
      STBP_13 = 0;
      STBQ_13 = 0;
      STSP_13 = 0;
      STSQ_13 = 0;
      STBP_14 = 0;
      STBQ_14 = 0;
      STSP_14 = 0;
      STSQ_14 = 0;
      STBP_15 = 0;
      STBQ_15 = 0;
      STSP_15 = 0;
      STSQ_15 = 0;
      STBP_16 = 0;
      STBQ_16 = 0;
      STSP_16 = 0;
      STSQ_16 = 0;
      STBP_17 = 0;
      STBQ_17 = 0;
      STSP_17 = 0;
      STSQ_17 = 0;
      STBP_18 = 0;
      STBQ_18 = 0;
      STSP_18 = 0;
      STSQ_18 = 0;
      STBP_19 = 0;
      STBQ_19 = 0;
      STSP_19 = 0;
      STSQ_19 = 0;
      STBP_20 = 0;
      STBQ_20 = 0;
      STSP_20 = 0;
      STSQ_20 = 0;
      STBP_21 = 0;
      STBQ_21 = 0;
      STSP_21 = 0;
      STSQ_21 = 0;
      STBP_22 = 0;
      STBQ_22 = 0;
      STSP_22 = 0;
      STSQ_22 = 0;
      STBP_23 = 0;
      STBQ_23 = 0;
      STSP_23 = 0;
      STSQ_23 = 0;
      STBP_24 = 0;
      STBQ_24 = 0;
      STSP_24 = 0;
      STSQ_24 = 0;
      STBP_25 = 0;
      STBQ_25 = 0;
      STSP_25 = 0;
      STSQ_25 = 0;
      STBP_26 = 0;
      STBQ_26 = 0;
      STSP_26 = 0;
      STSQ_26 = 0;
      STBP_27 = 0;
      STBQ_27 = 0;
      STSP_27 = 0;
      STSQ_27 = 0;
      STBP_28 = 0;
      STBQ_28 = 0;
      STSP_28 = 0;
      STSQ_28 = 0;
      STBP_29 = 0;
      STBQ_29 = 0;
      STSP_29 = 0;
      STSQ_29 = 0;
      STBP_30 = 0;
      STBQ_30 = 0;
      STSP_30 = 0;
      STSQ_30 = 0;
      STBP_31 = 0;
      STBQ_31 = 0;
      STSP_31 = 0;
      STSQ_31 = 0;
      STBP_32 = 0;
      STBQ_32 = 0;
      STSP_32 = 0;
      STSQ_32 = 0;
      STBP_33 = 0;
      STBQ_33 = 0;
      STSP_33 = 0;
      STSQ_33 = 0;
      STBP_34 = 0;
      STBQ_34 = 0;
      STSP_34 = 0;
      STSQ_34 = 0;
      STBP_35 = 0;
      STBQ_35 = 0;
      STSP_35 = 0;
      STSQ_35 = 0;
      STBP_36 = 0;
      STBQ_36 = 0;
      STSP_36 = 0;
      STSQ_36 = 0;
      STBP_37 = 0;
      STBQ_37 = 0;
      STSP_37 = 0;
      STSQ_37 = 0;
      STBP_38 = 0;
      STBQ_38 = 0;
      STSP_38 = 0;
      STSQ_38 = 0;
      STBP_39 = 0;
      STBQ_39 = 0;
      STSP_39 = 0;
      STSQ_39 = 0;
      STBP_40 = 0;
      STBQ_40 = 0;
      STSP_40 = 0;
      STSQ_40 = 0;
      STBP_41 = 0;
      STBQ_41 = 0;
      STSP_41 = 0;
      STSQ_41 = 0;
      STBP_42 = 0;
      STBQ_42 = 0;
      STSP_42 = 0;
      STSQ_42 = 0;
      STBP_43 = 0;
      STBQ_43 = 0;
      STSP_43 = 0;
      STSQ_43 = 0;
      STBP_44 = 0;
      STBQ_44 = 0;
      STSP_44 = 0;
      STSQ_44 = 0;
      STBP_45 = 0;
      STBQ_45 = 0;
      STSP_45 = 0;
      STSQ_45 = 0;
      STBP_46 = 0;
      STBQ_46 = 0;
      STSP_46 = 0;
      STSQ_46 = 0;
      STBP_47 = 0;
      STBQ_47 = 0;
      STSP_47 = 0;
      STSQ_47 = 0;
      STBP_48 = 0;
      STBQ_48 = 0;
      STSP_48 = 0;
      STSQ_48 = 0;
      STBP_49 = 0;
      STBQ_49 = 0;
      STSP_49 = 0;
      STSQ_49 = 0;
      return this;
    }

    public FinInfoBlock Set(FinInfoBlock3 fb) {
      LocalDate = new DateTime(fb.LocalDate.Ticks);
      LastUpdate = new DateTime(fb.LastUpdate.Ticks);
      LastTradeUpdate = new DateTime();
      Ask = fb.Ask;
      AvgPrice = fb.AvgPrice;
      Bid = fb.Bid;
      BuyCount = fb.BuyCount;
      BuyDeposit = fb.BuyDeposit;
      BuySqty = fb.BuySqty;
      Change = fb.Change;
      ChangePercent = fb.ChangePercent;
      ClosePrice = fb.ClosePrice;
      HighBid = fb.HighBid;
      IsTradingEnabled = fb.IsTradingEnabled;
      LastPrice = fb.LastPrice;
      LowOffer = fb.LowOffer;
      MaxDeal = fb.MaxDeal;
      MinDeal = fb.MinDeal;
      NoDeal = fb.NoDeal;
      OpenInterest = 0;
      OpenPrice = fb.OpenPrice;
      PrevPrice = fb.PrevPrice;
      Price2PrevAvg = fb.Price2PrevAvg;
      PriceMax = fb.PriceMax;
      PriceMin = fb.PriceMin;
      SellCount = fb.SellCount;
      SellDeposit = fb.SellDeposit;
      SellSqty = fb.SellSqty;
      StepPrice = fb.StepPrice;
      SumQty = fb.SumQty;
      SumVolume = fb.SumVolume;
      TheoreticPrice = fb.TheoreticPrice;
      Volatility = fb.Volatility;
      AskVolume = 0;
      BidVolume = 0;
      PriceOfHiAskVolume = 0;
      PriceOfHiBidVolume = 0;
      VolumeOfHiAskVolume = 0;
      VolumeOfHiBidVolume = 0;
      AllAskVolume = 0;
      AllBidVolume = 0;
      
      LastBarClosed = 0;
      // Здесь добовляем то что нам нужно (STBP_0) Stakan Tik Buy Price - 0
      STBP_0 = 0;
      STBQ_0 = 0;
      STSP_0 = 0;
      STSQ_0 = 0;
      STBP_1 = 0;
      STBQ_1 = 0;
      STSP_1 = 0;
      STSQ_1 = 0;
      STBP_2 = 0;
      STBQ_2 = 0;
      STSP_2 = 0;
      STSQ_2 = 0;
      STBP_3 = 0;
      STBQ_3 = 0;
      STSP_3 = 0;
      STSQ_3 = 0;
      STBP_4 = 0;
      STBQ_4 = 0;
      STSP_4 = 0;
      STSQ_4 = 0;
      STBP_5 = 0;
      STBQ_5 = 0;
      STSP_5 = 0;
      STSQ_5 = 0;
      STBP_6 = 0;
      STBQ_6 = 0;
      STSP_6 = 0;
      STSQ_6 = 0;
      STBP_7 = 0;
      STBQ_7 = 0;
      STSP_7 = 0;
      STSQ_7 = 0;
      STBP_8 = 0;
      STBQ_8 = 0;
      STSP_8 = 0;
      STSQ_8 = 0;
      STBP_9 = 0;
      STBQ_9 = 0;
      STSP_9 = 0;
      STSQ_9 = 0;
      STBP_10 = 0;
      STBQ_10 = 0;
      STSP_10 = 0;
      STSQ_10 = 0;
      STBP_11 = 0;
      STBQ_11 = 0;
      STSP_11 = 0;
      STSQ_11 = 0;
      STBP_12 = 0;
      STBQ_12 = 0;
      STSP_12 = 0;
      STSQ_12 = 0;
      STBP_13 = 0;
      STBQ_13 = 0;
      STSP_13 = 0;
      STSQ_13 = 0;
      STBP_14 = 0;
      STBQ_14 = 0;
      STSP_14 = 0;
      STSQ_14 = 0;
      STBP_15 = 0;
      STBQ_15 = 0;
      STSP_15 = 0;
      STSQ_15 = 0;
      STBP_16 = 0;
      STBQ_16 = 0;
      STSP_16 = 0;
      STSQ_16 = 0;
      STBP_17 = 0;
      STBQ_17 = 0;
      STSP_17 = 0;
      STSQ_17 = 0;
      STBP_18 = 0;
      STBQ_18 = 0;
      STSP_18 = 0;
      STSQ_18 = 0;
      STBP_19 = 0;
      STBQ_19 = 0;
      STSP_19 = 0;
      STSQ_19 = 0;
      STBP_20 = 0;
      STBQ_20 = 0;
      STSP_20 = 0;
      STSQ_20 = 0;
      STBP_21 = 0;
      STBQ_21 = 0;
      STSP_21 = 0;
      STSQ_21 = 0;
      STBP_22 = 0;
      STBQ_22 = 0;
      STSP_22 = 0;
      STSQ_22 = 0;
      STBP_23 = 0;
      STBQ_23 = 0;
      STSP_23 = 0;
      STSQ_23 = 0;
      STBP_24 = 0;
      STBQ_24 = 0;
      STSP_24 = 0;
      STSQ_24 = 0;
      STBP_25 = 0;
      STBQ_25 = 0;
      STSP_25 = 0;
      STSQ_25 = 0;
      STBP_26 = 0;
      STBQ_26 = 0;
      STSP_26 = 0;
      STSQ_26 = 0;
      STBP_27 = 0;
      STBQ_27 = 0;
      STSP_27 = 0;
      STSQ_27 = 0;
      STBP_28 = 0;
      STBQ_28 = 0;
      STSP_28 = 0;
      STSQ_28 = 0;
      STBP_29 = 0;
      STBQ_29 = 0;
      STSP_29 = 0;
      STSQ_29 = 0;
      STBP_30 = 0;
      STBQ_30 = 0;
      STSP_30 = 0;
      STSQ_30 = 0;
      STBP_31 = 0;
      STBQ_31 = 0;
      STSP_31 = 0;
      STSQ_31 = 0;
      STBP_32 = 0;
      STBQ_32 = 0;
      STSP_32 = 0;
      STSQ_32 = 0;
      STBP_33 = 0;
      STBQ_33 = 0;
      STSP_33 = 0;
      STSQ_33 = 0;
      STBP_34 = 0;
      STBQ_34 = 0;
      STSP_34 = 0;
      STSQ_34 = 0;
      STBP_35 = 0;
      STBQ_35 = 0;
      STSP_35 = 0;
      STSQ_35 = 0;
      STBP_36 = 0;
      STBQ_36 = 0;
      STSP_36 = 0;
      STSQ_36 = 0;
      STBP_37 = 0;
      STBQ_37 = 0;
      STSP_37 = 0;
      STSQ_37 = 0;
      STBP_38 = 0;
      STBQ_38 = 0;
      STSP_38 = 0;
      STSQ_38 = 0;
      STBP_39 = 0;
      STBQ_39 = 0;
      STSP_39 = 0;
      STSQ_39 = 0;
      STBP_40 = 0;
      STBQ_40 = 0;
      STSP_40 = 0;
      STSQ_40 = 0;
      STBP_41 = 0;
      STBQ_41 = 0;
      STSP_41 = 0;
      STSQ_41 = 0;
      STBP_42 = 0;
      STBQ_42 = 0;
      STSP_42 = 0;
      STSQ_42 = 0;
      STBP_43 = 0;
      STBQ_43 = 0;
      STSP_43 = 0;
      STSQ_43 = 0;
      STBP_44 = 0;
      STBQ_44 = 0;
      STSP_44 = 0;
      STSQ_44 = 0;
      STBP_45 = 0;
      STBQ_45 = 0;
      STSP_45 = 0;
      STSQ_45 = 0;
      STBP_46 = 0;
      STBQ_46 = 0;
      STSP_46 = 0;
      STSQ_46 = 0;
      STBP_47 = 0;
      STBQ_47 = 0;
      STSP_47 = 0;
      STSQ_47 = 0;
      STBP_48 = 0;
      STBQ_48 = 0;
      STSP_48 = 0;
      STSQ_48 = 0;
      STBP_49 = 0;
      STBQ_49 = 0;
      STSP_49 = 0;
      STSQ_49 = 0;
      return this;
    }

    public bool EqualsExceptLastUpdate(FinInfoBlock fi) {
      const float e = 1e-6f;

      return Math.Abs(Ask - fi.Ask) < e &&
             Math.Abs(AvgPrice - fi.AvgPrice) < e &&
             Math.Abs(Bid - fi.Bid) < e &&
             Math.Abs(BuyCount - fi.BuyCount) < e &&
             Math.Abs(BuyDeposit - fi.BuyDeposit) < e &&
             Math.Abs(BuySqty - fi.BuySqty) < e &&
             Math.Abs(Change - fi.Change) < e &&
             Math.Abs(ChangePercent - fi.ChangePercent) < e &&
             Math.Abs(ClosePrice - fi.ClosePrice) < e &&
             Math.Abs(HighBid - fi.HighBid) < e &&
             IsTradingEnabled == fi.IsTradingEnabled &&
             Math.Abs(LastPrice - fi.LastPrice) < e &&
             Math.Abs(LowOffer - fi.LowOffer) < e &&
             Math.Abs(MaxDeal - fi.MaxDeal) < e &&
             Math.Abs(MinDeal - fi.MinDeal) < e &&
             Math.Abs(NoDeal - fi.NoDeal) < e &&
             Math.Abs(OpenInterest - fi.OpenInterest) < e &&
             Math.Abs(OpenPrice - fi.OpenPrice) < e &&
             Math.Abs(PrevPrice - fi.PrevPrice) < e &&
             Math.Abs(Price2PrevAvg - fi.Price2PrevAvg) < e &&
             Math.Abs(PriceMax - fi.PriceMax) < e &&
             Math.Abs(PriceMin - fi.PriceMin) < e &&
             Math.Abs(SellCount - fi.SellCount) < e &&
             Math.Abs(SellDeposit - fi.SellDeposit) < e &&
             Math.Abs(SellSqty - fi.SellSqty) < e &&
             Math.Abs(StepPrice - fi.StepPrice) < e &&
             Math.Abs(SumQty - fi.SumQty) < e &&
             Math.Abs(SumVolume - fi.SumVolume) < e &&
             Math.Abs(TheoreticPrice - fi.TheoreticPrice) < e &&
             Math.Abs(Volatility - fi.Volatility) < e &&

             Math.Abs(AskVolume - fi.AskVolume) < e &&
             Math.Abs(BidVolume - fi.BidVolume) < e &&
             Math.Abs(PriceOfHiAskVolume - fi.PriceOfHiAskVolume) < e &&
             Math.Abs(PriceOfHiBidVolume - fi.PriceOfHiBidVolume) < e &&
             Math.Abs(VolumeOfHiAskVolume - fi.VolumeOfHiAskVolume) < e &&
             Math.Abs(VolumeOfHiBidVolume - fi.VolumeOfHiBidVolume) < e &&
             Math.Abs(AllAskVolume - fi.AllAskVolume) < e &&
             Math.Abs(AllBidVolume - fi.AllBidVolume) < e &&
      		 
      	     Math.Abs(LastBarClosed - fi.LastBarClosed) < e &&
      	     // Здесь добовляем то что нам нужно (STBP_0) Stakan Tik Buy Price - Math
      		 Math.Abs(STBP_0 - fi.STBP_0) < e &&
      	     Math.Abs(STBQ_0 - fi.STBQ_0) < e &&
      	     Math.Abs(STSP_0 - fi.STSP_0) < e &&
      	     Math.Abs(STSQ_0 - fi.STSQ_0) < e &&
      	 	 Math.Abs(STBP_1 - fi.STBP_1) < e &&
      	     Math.Abs(STBQ_1 - fi.STBQ_1) < e &&
      	     Math.Abs(STSP_1 - fi.STSP_1) < e &&
      	     Math.Abs(STSQ_1 - fi.STSQ_1) < e &&
      		 Math.Abs(STBP_2 - fi.STBP_2) < e &&
      	     Math.Abs(STBQ_2 - fi.STBQ_2) < e &&
      	     Math.Abs(STSP_2 - fi.STSP_2) < e &&
      	     Math.Abs(STSQ_2 - fi.STSQ_2) < e &&
      	     Math.Abs(STBP_3 - fi.STBP_3) < e &&
      	     Math.Abs(STBQ_3 - fi.STBQ_3) < e &&
      	     Math.Abs(STSP_3 - fi.STSP_3) < e &&
      	     Math.Abs(STSQ_3 - fi.STSQ_3) < e &&
      	     Math.Abs(STBP_4 - fi.STBP_4) < e &&
      	     Math.Abs(STBQ_4 - fi.STBQ_4) < e &&
      	     Math.Abs(STSP_4 - fi.STSP_4) < e &&
      	     Math.Abs(STSQ_4 - fi.STSQ_4) < e &&
      	     Math.Abs(STBP_5 - fi.STBP_5) < e &&
      	     Math.Abs(STBQ_5 - fi.STBQ_5) < e &&
      	     Math.Abs(STSP_5 - fi.STSP_5) < e &&
      	     Math.Abs(STSQ_5 - fi.STSQ_5) < e &&
      	     Math.Abs(STBP_6 - fi.STBP_6) < e &&
      	     Math.Abs(STBQ_6 - fi.STBQ_6) < e &&
      	     Math.Abs(STSP_6 - fi.STSP_6) < e &&
      	     Math.Abs(STSQ_6 - fi.STSQ_6) < e &&
      	     Math.Abs(STBP_7 - fi.STBP_7) < e &&
      	     Math.Abs(STBQ_7 - fi.STBQ_7) < e &&
      	     Math.Abs(STSP_7 - fi.STSP_7) < e &&
      	     Math.Abs(STSQ_7 - fi.STSQ_7) < e &&
      	     Math.Abs(STBP_8 - fi.STBP_8) < e &&
      	     Math.Abs(STBQ_8 - fi.STBQ_8) < e &&
      	     Math.Abs(STSP_8 - fi.STSP_8) < e &&
      	     Math.Abs(STSQ_8 - fi.STSQ_8) < e &&
      	     Math.Abs(STBP_9 - fi.STBP_9) < e &&
      	     Math.Abs(STBQ_9 - fi.STBQ_9) < e &&
      	     Math.Abs(STSP_9 - fi.STSP_9) < e &&
      	     Math.Abs(STSQ_9 - fi.STSQ_9) < e &&
      	     Math.Abs(STBP_10 - fi.STBP_10) < e &&
      	     Math.Abs(STBQ_10 - fi.STBQ_10) < e &&
      	     Math.Abs(STSP_10 - fi.STSP_10) < e &&
      	     Math.Abs(STSQ_10 - fi.STSQ_10) < e &&
      	     Math.Abs(STBP_11 - fi.STBP_11) < e &&
      	     Math.Abs(STBQ_11 - fi.STBQ_11) < e &&
      	     Math.Abs(STSP_11 - fi.STSP_11) < e &&
      	     Math.Abs(STSQ_11 - fi.STSQ_11) < e &&
      	     Math.Abs(STBP_12 - fi.STBP_12) < e &&
      	     Math.Abs(STBQ_12 - fi.STBQ_12) < e &&
      	     Math.Abs(STSP_12 - fi.STSP_12) < e &&
      	     Math.Abs(STSQ_12 - fi.STSQ_12) < e &&
      	     Math.Abs(STBP_13 - fi.STBP_13) < e &&
      	     Math.Abs(STBQ_13 - fi.STBQ_13) < e &&
      	     Math.Abs(STSP_13 - fi.STSP_13) < e &&
      	     Math.Abs(STSQ_13 - fi.STSQ_13) < e &&
      	     Math.Abs(STBP_14 - fi.STBP_14) < e &&
      	     Math.Abs(STBQ_14 - fi.STBQ_14) < e &&
      	     Math.Abs(STSP_14 - fi.STSP_14) < e &&
      	     Math.Abs(STSQ_14 - fi.STSQ_14) < e &&
      	     Math.Abs(STBP_15 - fi.STBP_15) < e &&
      	     Math.Abs(STBQ_15 - fi.STBQ_15) < e &&
      	     Math.Abs(STSP_15 - fi.STSP_15) < e &&
      	     Math.Abs(STSQ_15 - fi.STSQ_15) < e &&
      	     Math.Abs(STBP_16 - fi.STBP_16) < e &&
      	     Math.Abs(STBQ_16 - fi.STBQ_16) < e &&
      	     Math.Abs(STSP_16 - fi.STSP_16) < e &&
      	     Math.Abs(STSQ_16 - fi.STSQ_16) < e &&
      	     Math.Abs(STBP_17 - fi.STBP_17) < e &&
      	     Math.Abs(STBQ_17 - fi.STBQ_17) < e &&
      	     Math.Abs(STSP_17 - fi.STSP_17) < e &&
      	     Math.Abs(STSQ_17 - fi.STSQ_17) < e &&
      	     Math.Abs(STBP_18 - fi.STBP_18) < e &&
      	     Math.Abs(STBQ_18 - fi.STBQ_18) < e &&
      	     Math.Abs(STSP_18 - fi.STSP_18) < e &&
      	     Math.Abs(STSQ_18 - fi.STSQ_18) < e &&
      	     Math.Abs(STBP_19 - fi.STBP_19) < e &&
      	     Math.Abs(STBQ_19 - fi.STBQ_19) < e &&
      	     Math.Abs(STSP_19 - fi.STSP_19) < e &&
      	     Math.Abs(STSQ_19 - fi.STSQ_19) < e &&
       	     Math.Abs(STBP_20 - fi.STBP_20) < e &&
      	     Math.Abs(STBQ_20 - fi.STBQ_20) < e &&
      	     Math.Abs(STSP_20 - fi.STSP_20) < e &&
      	     Math.Abs(STSQ_20 - fi.STSQ_20) < e &&
      	     Math.Abs(STBP_21 - fi.STBP_21) < e &&
      	     Math.Abs(STBQ_21 - fi.STBQ_21) < e &&
      	     Math.Abs(STSP_21 - fi.STSP_21) < e &&
      	     Math.Abs(STSQ_21 - fi.STSQ_21) < e &&
      	     Math.Abs(STBP_22 - fi.STBP_22) < e &&
      	     Math.Abs(STBQ_22 - fi.STBQ_22) < e &&
      	     Math.Abs(STSP_22 - fi.STSP_22) < e &&
      	     Math.Abs(STSQ_22 - fi.STSQ_22) < e &&
      	     Math.Abs(STBP_23 - fi.STBP_23) < e &&
      	     Math.Abs(STBQ_23 - fi.STBQ_23) < e &&
      	     Math.Abs(STSP_23 - fi.STSP_23) < e &&
      	     Math.Abs(STSQ_23 - fi.STSQ_23) < e &&
      	     Math.Abs(STBP_24 - fi.STBP_24) < e &&
      	     Math.Abs(STBQ_24 - fi.STBQ_24) < e &&
      	     Math.Abs(STSP_24 - fi.STSP_24) < e &&
      	     Math.Abs(STSQ_24 - fi.STSQ_24) < e &&
      	     Math.Abs(STBP_25 - fi.STBP_25) < e &&
      	     Math.Abs(STBQ_25 - fi.STBQ_25) < e &&
      	     Math.Abs(STSP_25 - fi.STSP_25) < e &&
      	     Math.Abs(STSQ_25 - fi.STSQ_25) < e &&
      	     Math.Abs(STBP_26 - fi.STBP_26) < e &&
      	     Math.Abs(STBQ_26 - fi.STBQ_26) < e &&
      	     Math.Abs(STSP_26 - fi.STSP_26) < e &&
      	     Math.Abs(STSQ_26 - fi.STSQ_26) < e &&
      	     Math.Abs(STBP_27 - fi.STBP_27) < e &&
      	     Math.Abs(STBQ_27 - fi.STBQ_27) < e &&
      	     Math.Abs(STSP_27 - fi.STSP_27) < e &&
      	     Math.Abs(STSQ_27 - fi.STSQ_27) < e &&
      	     Math.Abs(STBP_28 - fi.STBP_28) < e &&
      	     Math.Abs(STBQ_28 - fi.STBQ_28) < e &&
      	     Math.Abs(STSP_28 - fi.STSP_28) < e &&
      	     Math.Abs(STSQ_28 - fi.STSQ_28) < e &&
      	     Math.Abs(STBP_29 - fi.STBP_29) < e &&
      	     Math.Abs(STBQ_29 - fi.STBQ_29) < e &&
      	     Math.Abs(STSP_29 - fi.STSP_29) < e &&
      	     Math.Abs(STSQ_29 - fi.STSQ_29) < e &&
      	     Math.Abs(STBP_30 - fi.STBP_30) < e &&
      	     Math.Abs(STBQ_30 - fi.STBQ_30) < e &&
      	     Math.Abs(STSP_30 - fi.STSP_30) < e &&
      	     Math.Abs(STSQ_30 - fi.STSQ_30) < e &&
      	     Math.Abs(STBP_31 - fi.STBP_31) < e &&
      	     Math.Abs(STBQ_31 - fi.STBQ_31) < e &&
      	     Math.Abs(STSP_31 - fi.STSP_31) < e &&
      	     Math.Abs(STSQ_31 - fi.STSQ_31) < e &&
      	     Math.Abs(STBP_32 - fi.STBP_32) < e &&
      	     Math.Abs(STBQ_32 - fi.STBQ_32) < e &&
      	     Math.Abs(STSP_32 - fi.STSP_32) < e &&
      	     Math.Abs(STSQ_32 - fi.STSQ_32) < e &&
      	     Math.Abs(STBP_33 - fi.STBP_33) < e &&
      	     Math.Abs(STBQ_33 - fi.STBQ_33) < e &&
      	     Math.Abs(STSP_33 - fi.STSP_33) < e &&
      	     Math.Abs(STSQ_33 - fi.STSQ_33) < e &&
      	     Math.Abs(STBP_34 - fi.STBP_34) < e &&
      	     Math.Abs(STBQ_34 - fi.STBQ_34) < e &&
      	     Math.Abs(STSP_34 - fi.STSP_34) < e &&
      	     Math.Abs(STSQ_34 - fi.STSQ_34) < e &&
      	     Math.Abs(STBP_35 - fi.STBP_35) < e &&
      	     Math.Abs(STBQ_35 - fi.STBQ_35) < e &&
      	     Math.Abs(STSP_35 - fi.STSP_35) < e &&
      	     Math.Abs(STSQ_35 - fi.STSQ_35) < e &&
      	     Math.Abs(STBP_36 - fi.STBP_36) < e &&
      	     Math.Abs(STBQ_36 - fi.STBQ_36) < e &&
      	     Math.Abs(STSP_36 - fi.STSP_36) < e &&
      	     Math.Abs(STSQ_36 - fi.STSQ_36) < e &&
      	     Math.Abs(STBP_37 - fi.STBP_37) < e &&
      	     Math.Abs(STBQ_37 - fi.STBQ_37) < e &&
      	     Math.Abs(STSP_37 - fi.STSP_37) < e &&
      	     Math.Abs(STSQ_37 - fi.STSQ_37) < e &&
      	     Math.Abs(STBP_38 - fi.STBP_38) < e &&
      	     Math.Abs(STBQ_38 - fi.STBQ_38) < e &&
      	     Math.Abs(STSP_38 - fi.STSP_38) < e &&
      	     Math.Abs(STSQ_38 - fi.STSQ_38) < e &&
      	     Math.Abs(STBP_39 - fi.STBP_39) < e &&
      	     Math.Abs(STBQ_39 - fi.STBQ_39) < e &&
      	     Math.Abs(STSP_39 - fi.STSP_39) < e &&
      	     Math.Abs(STSQ_39 - fi.STSQ_39) < e &&
      	     Math.Abs(STBP_40 - fi.STBP_40) < e &&
      	     Math.Abs(STBQ_40 - fi.STBQ_40) < e &&
      	     Math.Abs(STSP_40 - fi.STSP_40) < e &&
      	     Math.Abs(STSQ_40 - fi.STSQ_40) < e &&
      	     Math.Abs(STBP_41 - fi.STBP_41) < e &&
      	     Math.Abs(STBQ_41 - fi.STBQ_41) < e &&
      	     Math.Abs(STSP_41 - fi.STSP_41) < e &&
      	     Math.Abs(STSQ_41 - fi.STSQ_41) < e &&
      	     Math.Abs(STBP_42 - fi.STBP_42) < e &&
      	     Math.Abs(STBQ_42 - fi.STBQ_42) < e &&
      	     Math.Abs(STSP_42 - fi.STSP_42) < e &&
      	     Math.Abs(STSQ_42 - fi.STSQ_42) < e &&
      	     Math.Abs(STBP_43 - fi.STBP_43) < e &&
      	     Math.Abs(STBQ_43 - fi.STBQ_43) < e &&
      	     Math.Abs(STSP_43 - fi.STSP_43) < e &&
      	     Math.Abs(STSQ_43 - fi.STSQ_43) < e &&
      	     Math.Abs(STBP_44 - fi.STBP_44) < e &&
      	     Math.Abs(STBQ_44 - fi.STBQ_44) < e &&
      	     Math.Abs(STSP_44 - fi.STSP_44) < e &&
      	     Math.Abs(STSQ_44 - fi.STSQ_44) < e &&
      	     Math.Abs(STBP_45 - fi.STBP_45) < e &&
      	     Math.Abs(STBQ_45 - fi.STBQ_45) < e &&
      	     Math.Abs(STSP_45 - fi.STSP_45) < e &&
      	     Math.Abs(STSQ_45 - fi.STSQ_45) < e &&
      	     Math.Abs(STBP_46 - fi.STBP_46) < e &&
      	     Math.Abs(STBQ_46 - fi.STBQ_46) < e &&
      	     Math.Abs(STSP_46 - fi.STSP_46) < e &&
      	     Math.Abs(STSQ_46 - fi.STSQ_46) < e &&
      	     Math.Abs(STBP_47 - fi.STBP_47) < e &&
      	     Math.Abs(STBQ_47 - fi.STBQ_47) < e &&
      	     Math.Abs(STSP_47 - fi.STSP_47) < e &&
      	     Math.Abs(STSQ_47 - fi.STSQ_47) < e &&
      	     Math.Abs(STBP_48 - fi.STBP_48) < e &&
      	     Math.Abs(STBQ_48 - fi.STBQ_48) < e &&
      	     Math.Abs(STSP_48 - fi.STSP_48) < e &&
      	     Math.Abs(STSQ_48 - fi.STSQ_48) < e &&
      	     Math.Abs(STBP_49 - fi.STBP_49) < e &&
      	     Math.Abs(STBQ_49 - fi.STBQ_49) < e &&
      	     Math.Abs(STSP_49 - fi.STSP_49) < e &&
      	     Math.Abs(STSQ_49 - fi.STSQ_49) < e
      		 ;
    }

    public FinInfoBlock Clone() { return (FinInfoBlock) MemberwiseClone(); }

    public void WriteRecord(BinaryWriter bw) {
      bw.Write(TimeZoneInfo.ConvertTime(LastUpdate, FinInfoCache.ExchangeTimeZone, TimeZoneInfo.Utc).ToBinary());
      bw.Write(TimeZoneInfo.ConvertTime(LocalDate, FinInfoCache.ExchangeTimeZone, TimeZoneInfo.Utc).ToBinary());
      bw.Write(TimeZoneInfo.ConvertTime(LastTradeUpdate, FinInfoCache.ExchangeTimeZone, TimeZoneInfo.Utc).ToBinary());
      bw.Write(Ask);
      bw.Write(AvgPrice);
      bw.Write(Bid);
      bw.Write(BuyCount);
      bw.Write(BuyDeposit);
      bw.Write(BuySqty);
      bw.Write(Change);
      bw.Write(ChangePercent);
      bw.Write(ClosePrice);
      bw.Write(HighBid);
      bw.Write(IsTradingEnabled);
      bw.Write(LastPrice);
      bw.Write(LowOffer);
      bw.Write(MaxDeal);
      bw.Write(MinDeal);
      bw.Write(NoDeal);
      bw.Write(OpenInterest);
      bw.Write(OpenPrice);
      bw.Write(PrevPrice);
      bw.Write(Price2PrevAvg);
      bw.Write(PriceMax);
      bw.Write(PriceMin);
      bw.Write(SellCount);
      bw.Write(SellDeposit);
      bw.Write(SellSqty);
      bw.Write(StepPrice);
      bw.Write(SumQty);
      bw.Write(SumVolume);
      bw.Write(TheoreticPrice);
      bw.Write(Volatility);
      bw.Write(AskVolume);
      bw.Write(BidVolume);
      bw.Write(PriceOfHiAskVolume);
      bw.Write(PriceOfHiBidVolume);
      bw.Write(VolumeOfHiAskVolume);
      bw.Write(VolumeOfHiBidVolume);
      bw.Write(AllAskVolume);
      bw.Write(AllBidVolume);
      
      bw.Write(LastBarClosed);
      // Здесь добовляем то что нам нужно (STBP_0) Stakan Tik Buy Price - Write
      bw.Write(STBP_0);
      bw.Write(STBQ_0);
      bw.Write(STSP_0);
      bw.Write(STSQ_0);
      bw.Write(STBP_1);
      bw.Write(STBQ_1);
      bw.Write(STSP_1);
      bw.Write(STSQ_1);
      bw.Write(STBP_2);
      bw.Write(STBQ_2);
      bw.Write(STSP_2);
      bw.Write(STSQ_2);
      bw.Write(STBP_3);
      bw.Write(STBQ_3);
      bw.Write(STSP_3);
      bw.Write(STSQ_3);
      bw.Write(STBP_4);
      bw.Write(STBQ_4);
      bw.Write(STSP_4);
      bw.Write(STSQ_4);
      bw.Write(STBP_5);
      bw.Write(STBQ_5);
      bw.Write(STSP_5);
      bw.Write(STSQ_5);
      bw.Write(STBP_6);
      bw.Write(STBQ_6);
      bw.Write(STSP_6);
      bw.Write(STSQ_6);
      bw.Write(STBP_7);
      bw.Write(STBQ_7);
      bw.Write(STSP_7);
      bw.Write(STSQ_7);
      bw.Write(STBP_8);
      bw.Write(STBQ_8);
      bw.Write(STSP_8);
      bw.Write(STSQ_8);
      bw.Write(STBP_9);
      bw.Write(STBQ_9);
      bw.Write(STSP_9);
      bw.Write(STSQ_9);
      bw.Write(STBP_10);
      bw.Write(STBQ_10);
      bw.Write(STSP_10);
      bw.Write(STSQ_10);
      bw.Write(STBP_11);
      bw.Write(STBQ_11);
      bw.Write(STSP_11);
      bw.Write(STSQ_11);
      bw.Write(STBP_12);
      bw.Write(STBQ_12);
      bw.Write(STSP_12);
      bw.Write(STSQ_12);
      bw.Write(STBP_13);
      bw.Write(STBQ_13);
      bw.Write(STSP_13);
      bw.Write(STSQ_13);
      bw.Write(STBP_14);
      bw.Write(STBQ_14);
      bw.Write(STSP_14);
      bw.Write(STSQ_14);
      bw.Write(STBP_15);
      bw.Write(STBQ_15);
      bw.Write(STSP_15);
      bw.Write(STSQ_15);
      bw.Write(STBP_16);
      bw.Write(STBQ_16);
      bw.Write(STSP_16);
      bw.Write(STSQ_16);
      bw.Write(STBP_17);
      bw.Write(STBQ_17);
      bw.Write(STSP_17);
      bw.Write(STSQ_17);
      bw.Write(STBP_18);
      bw.Write(STBQ_18);
      bw.Write(STSP_18);
      bw.Write(STSQ_18);
      bw.Write(STBP_19);
      bw.Write(STBQ_19);
      bw.Write(STSP_19);
      bw.Write(STSQ_19);
      bw.Write(STBP_20);
      bw.Write(STBQ_20);
      bw.Write(STSP_20);
      bw.Write(STSQ_20);
      bw.Write(STBP_21);
      bw.Write(STBQ_21);
      bw.Write(STSP_21);
      bw.Write(STSQ_21);
      bw.Write(STBP_22);
      bw.Write(STBQ_22);
      bw.Write(STSP_22);
      bw.Write(STSQ_22);
      bw.Write(STBP_23);
      bw.Write(STBQ_23);
      bw.Write(STSP_23);
      bw.Write(STSQ_23);
      bw.Write(STBP_24);
      bw.Write(STBQ_24);
      bw.Write(STSP_24);
      bw.Write(STSQ_24);
      bw.Write(STBP_25);
      bw.Write(STBQ_25);
      bw.Write(STSP_25);
      bw.Write(STSQ_25);
      bw.Write(STBP_26);
      bw.Write(STBQ_26);
      bw.Write(STSP_26);
      bw.Write(STSQ_26);
      bw.Write(STBP_27);
      bw.Write(STBQ_27);
      bw.Write(STSP_27);
      bw.Write(STSQ_27);
      bw.Write(STBP_28);
      bw.Write(STBQ_28);
      bw.Write(STSP_28);
      bw.Write(STSQ_28);
      bw.Write(STBP_29);
      bw.Write(STBQ_29);
      bw.Write(STSP_29);
      bw.Write(STSQ_29);
      bw.Write(STBP_30);
      bw.Write(STBQ_30);
      bw.Write(STSP_30);
      bw.Write(STSQ_30);
      bw.Write(STBP_31);
      bw.Write(STBQ_31);
      bw.Write(STSP_31);
      bw.Write(STSQ_31);
      bw.Write(STBP_32);
      bw.Write(STBQ_32);
      bw.Write(STSP_32);
      bw.Write(STSQ_32);
      bw.Write(STBP_33);
      bw.Write(STBQ_33);
      bw.Write(STSP_33);
      bw.Write(STSQ_33);
      bw.Write(STBP_34);
      bw.Write(STBQ_34);
      bw.Write(STSP_34);
      bw.Write(STSQ_34);
      bw.Write(STBP_35);
      bw.Write(STBQ_35);
      bw.Write(STSP_35);
      bw.Write(STSQ_35);
      bw.Write(STBP_36);
      bw.Write(STBQ_36);
      bw.Write(STSP_36);
      bw.Write(STSQ_36);
      bw.Write(STBP_37);
      bw.Write(STBQ_37);
      bw.Write(STSP_37);
      bw.Write(STSQ_37);
      bw.Write(STBP_38);
      bw.Write(STBQ_38);
      bw.Write(STSP_38);
      bw.Write(STSQ_38);
      bw.Write(STBP_39);
      bw.Write(STBQ_39);
      bw.Write(STSP_39);
      bw.Write(STSQ_39);
      bw.Write(STBP_40);
      bw.Write(STBQ_40);
      bw.Write(STSP_40);
      bw.Write(STSQ_40);
      bw.Write(STBP_41);
      bw.Write(STBQ_41);
      bw.Write(STSP_41);
      bw.Write(STSQ_41);
      bw.Write(STBP_42);
      bw.Write(STBQ_42);
      bw.Write(STSP_42);
      bw.Write(STSQ_42);
      bw.Write(STBP_43);
      bw.Write(STBQ_43);
      bw.Write(STSP_43);
      bw.Write(STSQ_43);
      bw.Write(STBP_44);
      bw.Write(STBQ_44);
      bw.Write(STSP_44);
      bw.Write(STSQ_44);
      bw.Write(STBP_45);
      bw.Write(STBQ_45);
      bw.Write(STSP_45);
      bw.Write(STSQ_45);
      bw.Write(STBP_46);
      bw.Write(STBQ_46);
      bw.Write(STSP_46);
      bw.Write(STSQ_46);
      bw.Write(STBP_47);
      bw.Write(STBQ_47);
      bw.Write(STSP_47);
      bw.Write(STSQ_47);
      bw.Write(STBP_48);
      bw.Write(STBQ_48);
      bw.Write(STSP_48);
      bw.Write(STSQ_48);
      bw.Write(STBP_49);
      bw.Write(STBQ_49);
      bw.Write(STSP_49);
      bw.Write(STSQ_49);      
    }

    public static FinInfoBlock ReadFininfoBlock(BinaryReader br, long rec) {
      br.BaseStream.Seek(HeaderSize + rec * RecSize, SeekOrigin.Begin);
      var fb = new FinInfoBlock();
      ReadFininfoBlock(br, fb);
      return fb;
    }

    static readonly DateTime _zeroDate = new DateTime(2000, 1, 1); // Ранеше жизни не было

    public static bool ReadFininfoBlock(BinaryReader br, FinInfoBlock fb) {
      if (br.BaseStream.Length - br.BaseStream.Position < RecSize)
        return false;
      fb.LastUpdate = TimeZoneInfo.ConvertTime(DateTime.FromBinary(br.ReadInt64()), TimeZoneInfo.Utc, FinInfoCache.ExchangeTimeZone);
      fb.LocalDate = TimeZoneInfo.ConvertTime(DateTime.FromBinary(br.ReadInt64()), TimeZoneInfo.Utc, FinInfoCache.ExchangeTimeZone);
      fb.LastTradeUpdate = TimeZoneInfo.ConvertTime(DateTime.FromBinary(br.ReadInt64()), TimeZoneInfo.Utc, FinInfoCache.ExchangeTimeZone);
      if (fb.LastTradeUpdate < _zeroDate)
        fb.LastTradeUpdate = new DateTime();
      fb.Ask = br.ReadSingle();
      fb.AvgPrice = br.ReadSingle();
      fb.Bid = br.ReadSingle();
      fb.BuyCount = br.ReadSingle();
      fb.BuyDeposit = br.ReadSingle();
      fb.BuySqty = br.ReadSingle();
      fb.Change = br.ReadSingle();
      fb.ChangePercent = br.ReadSingle();
      fb.ClosePrice = br.ReadSingle();
      fb.HighBid = br.ReadSingle();
      fb.IsTradingEnabled = br.ReadBoolean();
      fb.LastPrice = br.ReadSingle();
      fb.LowOffer = br.ReadSingle();
      fb.MaxDeal = br.ReadSingle();
      fb.MinDeal = br.ReadSingle();
      fb.NoDeal = br.ReadSingle();
      fb.OpenInterest = br.ReadSingle();
      fb.OpenPrice = br.ReadSingle();
      fb.PrevPrice = br.ReadSingle();
      fb.Price2PrevAvg = br.ReadSingle();
      fb.PriceMax = br.ReadSingle();
      fb.PriceMin = br.ReadSingle();
      fb.SellCount = br.ReadSingle();
      fb.SellDeposit = br.ReadSingle();
      fb.SellSqty = br.ReadSingle();
      fb.StepPrice = br.ReadSingle();
      fb.SumQty = br.ReadSingle();
      fb.SumVolume = br.ReadSingle();
      fb.TheoreticPrice = br.ReadSingle();
      fb.Volatility = br.ReadSingle();
      fb.AskVolume = br.ReadSingle();
      fb.BidVolume = br.ReadSingle();
      fb.PriceOfHiAskVolume = br.ReadSingle();
      fb.PriceOfHiBidVolume = br.ReadSingle();
      fb.VolumeOfHiAskVolume = br.ReadSingle();
      fb.VolumeOfHiBidVolume = br.ReadSingle();
      fb.AllAskVolume = br.ReadSingle();
      fb.AllBidVolume = br.ReadSingle();
      
      fb.LastBarClosed = br.ReadSingle();
      // Здесь добовляем то что нам нужно (STBP_0) Stakan Tik Buy Price - Read
	  fb.STBP_0 = br.ReadSingle();
	  fb.STBQ_0 = br.ReadSingle();
	  fb.STSP_0 = br.ReadSingle();
	  fb.STSQ_0 = br.ReadSingle();
	  fb.STBP_1 = br.ReadSingle();
	  fb.STBQ_1 = br.ReadSingle();
	  fb.STSP_1 = br.ReadSingle();
	  fb.STSQ_1 = br.ReadSingle();
      fb.STBP_2 = br.ReadSingle();
      fb.STBQ_2 = br.ReadSingle();
      fb.STSP_2 = br.ReadSingle();
      fb.STSQ_2 = br.ReadSingle();
      fb.STBP_3 = br.ReadSingle();
      fb.STBQ_3 = br.ReadSingle();
      fb.STSP_3 = br.ReadSingle();
      fb.STSQ_3 = br.ReadSingle();
      fb.STBP_4 = br.ReadSingle();
      fb.STBQ_4 = br.ReadSingle();
      fb.STSP_4 = br.ReadSingle();
      fb.STSQ_4 = br.ReadSingle();
      fb.STBP_5 = br.ReadSingle();
      fb.STBQ_5 = br.ReadSingle();
      fb.STSP_5 = br.ReadSingle();
      fb.STSQ_5 = br.ReadSingle();
      fb.STBP_6 = br.ReadSingle();
      fb.STBQ_6 = br.ReadSingle();
      fb.STSP_6 = br.ReadSingle();
      fb.STSQ_6 = br.ReadSingle();
      fb.STBP_7 = br.ReadSingle();
      fb.STBQ_7 = br.ReadSingle();
      fb.STSP_7 = br.ReadSingle();
      fb.STSQ_7 = br.ReadSingle();
      fb.STBP_8 = br.ReadSingle();
      fb.STBQ_8 = br.ReadSingle();
      fb.STSP_8 = br.ReadSingle();
      fb.STSQ_8 = br.ReadSingle();
      fb.STBP_9 = br.ReadSingle();
      fb.STBQ_9 = br.ReadSingle();
      fb.STSP_9 = br.ReadSingle();
      fb.STSQ_9 = br.ReadSingle();
      fb.STBP_10 = br.ReadSingle();
      fb.STBQ_10 = br.ReadSingle();
      fb.STSP_10 = br.ReadSingle();
      fb.STSQ_10 = br.ReadSingle();
      fb.STBP_11 = br.ReadSingle();
      fb.STBQ_11 = br.ReadSingle();
      fb.STSP_11 = br.ReadSingle();
      fb.STSQ_11 = br.ReadSingle();
      fb.STBP_12 = br.ReadSingle();
      fb.STBQ_12 = br.ReadSingle();
      fb.STSP_12 = br.ReadSingle();
      fb.STSQ_12 = br.ReadSingle();
      fb.STBP_13 = br.ReadSingle();
      fb.STBQ_13 = br.ReadSingle();
      fb.STSP_13 = br.ReadSingle();
      fb.STSQ_13 = br.ReadSingle();
      fb.STBP_14 = br.ReadSingle();
      fb.STBQ_14 = br.ReadSingle();
      fb.STSP_14 = br.ReadSingle();
      fb.STSQ_14 = br.ReadSingle();
      fb.STBP_15 = br.ReadSingle();
      fb.STBQ_15 = br.ReadSingle();
      fb.STSP_15 = br.ReadSingle();
      fb.STSQ_15 = br.ReadSingle();
      fb.STBP_16 = br.ReadSingle();
      fb.STBQ_16 = br.ReadSingle();
      fb.STSP_16 = br.ReadSingle();
      fb.STSQ_16 = br.ReadSingle();
      fb.STBP_17 = br.ReadSingle();
      fb.STBQ_17 = br.ReadSingle();
      fb.STSP_17 = br.ReadSingle();
      fb.STSQ_17 = br.ReadSingle();
      fb.STBP_18 = br.ReadSingle();
      fb.STBQ_18 = br.ReadSingle();
      fb.STSP_18 = br.ReadSingle();
      fb.STSQ_18 = br.ReadSingle();
      fb.STBP_19 = br.ReadSingle();
      fb.STBQ_19 = br.ReadSingle();
      fb.STSP_19 = br.ReadSingle();
      fb.STSQ_19 = br.ReadSingle();
      fb.STBP_20 = br.ReadSingle();
      fb.STBQ_20 = br.ReadSingle();
      fb.STSP_20 = br.ReadSingle();
      fb.STSQ_20 = br.ReadSingle();
      fb.STBP_21 = br.ReadSingle();
      fb.STBQ_21 = br.ReadSingle();
      fb.STSP_21 = br.ReadSingle();
      fb.STSQ_21 = br.ReadSingle();
      fb.STBP_22 = br.ReadSingle();
      fb.STBQ_22 = br.ReadSingle();
      fb.STSP_22 = br.ReadSingle();
      fb.STSQ_22 = br.ReadSingle();
      fb.STBP_23 = br.ReadSingle();
      fb.STBQ_23 = br.ReadSingle();
      fb.STSP_23 = br.ReadSingle();
      fb.STSQ_23 = br.ReadSingle();
      fb.STBP_24 = br.ReadSingle();
      fb.STBQ_24 = br.ReadSingle();
      fb.STSP_24 = br.ReadSingle();
      fb.STSQ_24 = br.ReadSingle();
      fb.STBP_25 = br.ReadSingle();
      fb.STBQ_25 = br.ReadSingle();
      fb.STSP_25 = br.ReadSingle();
      fb.STSQ_25 = br.ReadSingle();
      fb.STBP_26 = br.ReadSingle();
      fb.STBQ_26 = br.ReadSingle();
      fb.STSP_26 = br.ReadSingle();
      fb.STSQ_26 = br.ReadSingle();
      fb.STBP_27 = br.ReadSingle();
      fb.STBQ_27 = br.ReadSingle();
      fb.STSP_27 = br.ReadSingle();
      fb.STSQ_27 = br.ReadSingle();
      fb.STBP_28 = br.ReadSingle();
      fb.STBQ_28 = br.ReadSingle();
      fb.STSP_28 = br.ReadSingle();
      fb.STSQ_28 = br.ReadSingle();
      fb.STBP_29 = br.ReadSingle();
      fb.STBQ_29 = br.ReadSingle();
      fb.STSP_29 = br.ReadSingle();
      fb.STSQ_29 = br.ReadSingle();
      fb.STBP_30 = br.ReadSingle();
      fb.STBQ_30 = br.ReadSingle();
      fb.STSP_30 = br.ReadSingle();
      fb.STSQ_30 = br.ReadSingle();
      fb.STBP_31 = br.ReadSingle();
      fb.STBQ_31 = br.ReadSingle();
      fb.STSP_31 = br.ReadSingle();
      fb.STSQ_31 = br.ReadSingle();
      fb.STBP_32 = br.ReadSingle();
      fb.STBQ_32 = br.ReadSingle();
      fb.STSP_32 = br.ReadSingle();
      fb.STSQ_32 = br.ReadSingle();
      fb.STBP_33 = br.ReadSingle();
      fb.STBQ_33 = br.ReadSingle();
      fb.STSP_33 = br.ReadSingle();
      fb.STSQ_33 = br.ReadSingle();
      fb.STBP_34 = br.ReadSingle();
      fb.STBQ_34 = br.ReadSingle();
      fb.STSP_34 = br.ReadSingle();
      fb.STSQ_34 = br.ReadSingle();
      fb.STBP_35 = br.ReadSingle();
      fb.STBQ_35 = br.ReadSingle();
      fb.STSP_35 = br.ReadSingle();
      fb.STSQ_35 = br.ReadSingle();
      fb.STBP_36 = br.ReadSingle();
      fb.STBQ_36 = br.ReadSingle();
      fb.STSP_36 = br.ReadSingle();
      fb.STSQ_36 = br.ReadSingle();
      fb.STBP_37 = br.ReadSingle();
      fb.STBQ_37 = br.ReadSingle();
      fb.STSP_37 = br.ReadSingle();
      fb.STSQ_37 = br.ReadSingle();
      fb.STBP_38 = br.ReadSingle();
      fb.STBQ_38 = br.ReadSingle();
      fb.STSP_38 = br.ReadSingle();
      fb.STSQ_38 = br.ReadSingle();
      fb.STBP_39 = br.ReadSingle();
      fb.STBQ_39 = br.ReadSingle();
      fb.STSP_39 = br.ReadSingle();
      fb.STSQ_39 = br.ReadSingle();
      fb.STBP_40 = br.ReadSingle();
      fb.STBQ_40 = br.ReadSingle();
      fb.STSP_40 = br.ReadSingle();
      fb.STSQ_40 = br.ReadSingle();
      fb.STBP_41 = br.ReadSingle();
      fb.STBQ_41 = br.ReadSingle();
      fb.STSP_41 = br.ReadSingle();
      fb.STSQ_41 = br.ReadSingle();
      fb.STBP_42 = br.ReadSingle();
      fb.STBQ_42 = br.ReadSingle();
      fb.STSP_42 = br.ReadSingle();
      fb.STSQ_42 = br.ReadSingle();
      fb.STBP_43 = br.ReadSingle();
      fb.STBQ_43 = br.ReadSingle();
      fb.STSP_43 = br.ReadSingle();
      fb.STSQ_43 = br.ReadSingle();
      fb.STBP_44 = br.ReadSingle();
      fb.STBQ_44 = br.ReadSingle();
      fb.STSP_44 = br.ReadSingle();
      fb.STSQ_44 = br.ReadSingle();
      fb.STBP_45 = br.ReadSingle();
      fb.STBQ_45 = br.ReadSingle();
      fb.STSP_45 = br.ReadSingle();
      fb.STSQ_45 = br.ReadSingle();
      fb.STBP_46 = br.ReadSingle();
      fb.STBQ_46 = br.ReadSingle();
      fb.STSP_46 = br.ReadSingle();
      fb.STSQ_46 = br.ReadSingle();
      fb.STBP_47 = br.ReadSingle();
      fb.STBQ_47 = br.ReadSingle();
      fb.STSP_47 = br.ReadSingle();
      fb.STSQ_47 = br.ReadSingle();
      fb.STBP_48 = br.ReadSingle();
      fb.STBQ_48 = br.ReadSingle();
      fb.STSP_48 = br.ReadSingle();
      fb.STSQ_48 = br.ReadSingle();
      fb.STBP_49 = br.ReadSingle();
      fb.STBQ_49 = br.ReadSingle();
      fb.STSP_49 = br.ReadSingle();
      fb.STSQ_49 = br.ReadSingle();	  
      return true;
    }

    public void GetDomValues(ISecurity security) {
      {
        var asks = security.GetSellQueue(0);
        if (asks != null && asks.Count > 0) {
          var qd = asks[0];
          Ask = (float) qd.Price;
          AskVolume = (float) qd.Quantity;
         
         // Здесь добовляем то что нам нужно Stakan Tik Sell Price
          STSP_0 = (float) qd.Price;
          STSQ_0 = (float) qd.Quantity;
         //
          
         double maxVol = 0, price = 0, sum = 0;
          foreach (var line in asks) {
            if (maxVol < line.Quantity) {
              maxVol = line.Quantity;
              price = line.Price;
            }
            sum += line.Quantity;
          }
          PriceOfHiAskVolume = (float) price;
          VolumeOfHiAskVolume = (float) maxVol;
          AllAskVolume = (float) sum;
        }
      }
    	
      var bids = security.GetBuyQueue(0);
      if (bids != null && bids.Count > 0) {
        var qd = bids[0];
        Bid = (float) qd.Price;
        BidVolume = (float) qd.Quantity;
		
        //LastBarClosed = (float) qd.LastUpdate.Ticks;
        LastBarClosed = (float) security.Bars.Count;
                
        // Здесь добовляем то что нам нужно Stakan Tik Buy Price
		STBP_0 = (float) qd.Price;
        STBQ_0 = (float) qd.Quantity;
        //
        
        double maxVol = 0, price = 0, sum = 0;
        foreach (var line in bids) {
          if (maxVol < line.Quantity) {
            maxVol = line.Quantity;
            price = line.Price;
          }
          sum += line.Quantity;
        }
        PriceOfHiBidVolume = (float) price;
        VolumeOfHiBidVolume = (float) maxVol;
        AllBidVolume = (float) sum;
      }
      
      // Здесь добовляем то что нам нужно Stakan Tik Sell Price 1
      var asks_1 = security.GetSellQueue(1);
      if (asks_1 != null && asks_1.Count > 0) {
          var qd = asks_1[1];
          
          STSP_1 = (float) qd.Price;
          STSQ_1 = (float) qd.Quantity;       }
      
      var bids_1 = security.GetBuyQueue(1);
      if (bids_1 != null && bids_1.Count > 0) {
          var qd = bids_1[1];
       
          STBP_1 = (float) qd.Price;
          STBQ_1 = (float) qd.Quantity;       }
      //
      // Здесь добовляем то что нам нужно Stakan Tik Sell Price 2
      var asks_2 = security.GetSellQueue(2);
      if (asks_2 != null && asks_2.Count > 0) {
          var qd = asks_2[2];
          
          STSP_2 = (float) qd.Price;
          STSQ_2 = (float) qd.Quantity;       }
      
      var bids_2 = security.GetBuyQueue(2);
      if (bids_2 != null && bids_2.Count > 0) {
          var qd = bids_2[2];
       
          STBP_2 = (float) qd.Price;
          STBQ_2 = (float) qd.Quantity;       }
      //
// Здесь добовляем то что нам нужно Stakan Tik Sell Price 3
      var asks_3 = security.GetSellQueue(3);
      if (asks_3 != null && asks_3.Count > 0) {
          var qd = asks_3[3];
          
          STSP_3 = (float) qd.Price;
          STSQ_3 = (float) qd.Quantity;       }
      
      var bids_3 = security.GetBuyQueue(3);
      if (bids_3 != null && bids_3.Count > 0) {
          var qd = bids_3[3];
       
          STBP_3 = (float) qd.Price;
          STBQ_3 = (float) qd.Quantity;       }
      //
// Здесь добовляем то что нам нужно Stakan Tik Sell Price 4
      var asks_4 = security.GetSellQueue(4);
      if (asks_4 != null && asks_4.Count > 0) {
          var qd = asks_4[4];
          
          STSP_4 = (float) qd.Price;
          STSQ_4 = (float) qd.Quantity;       }
      
      var bids_4 = security.GetBuyQueue(4);
      if (bids_4 != null && bids_4.Count > 0) {
          var qd = bids_4[4];
       
          STBP_4 = (float) qd.Price;
          STBQ_4 = (float) qd.Quantity;       }
      //
 // Здесь добовляем то что нам нужно Stakan Tik Sell Price 5
      var asks_5 = security.GetSellQueue(5);
      if (asks_5 != null && asks_5.Count > 0) {
          var qd = asks_5[5];
          
          STSP_5 = (float) qd.Price;
          STSQ_5 = (float) qd.Quantity;       }
      
      var bids_5 = security.GetBuyQueue(5);
      if (bids_5 != null && bids_5.Count > 0) {
          var qd = bids_5[5];
       
          STBP_5 = (float) qd.Price;
          STBQ_5 = (float) qd.Quantity;       }
      //
// Здесь добовляем то что нам нужно Stakan Tik Sell Price 6
      var asks_6 = security.GetSellQueue(6);
      if (asks_6 != null && asks_6.Count > 0) {
          var qd = asks_6[6];
          
          STSP_6 = (float) qd.Price;
          STSQ_6 = (float) qd.Quantity;       }
      
      var bids_6 = security.GetBuyQueue(6);
      if (bids_6 != null && bids_6.Count > 0) {
          var qd = bids_6[6];
       
          STBP_6 = (float) qd.Price;
          STBQ_6 = (float) qd.Quantity;       }
      //
// Здесь добовляем то что нам нужно Stakan Tik Sell Price 7
      var asks_7 = security.GetSellQueue(7);
      if (asks_7 != null && asks_7.Count > 0) {
          var qd = asks_7[7];
          
          STSP_7 = (float) qd.Price;
          STSQ_7 = (float) qd.Quantity;       }
      
      var bids_7 = security.GetBuyQueue(7);
      if (bids_7 != null && bids_7.Count > 0) {
          var qd = bids_7[7];
       
          STBP_7 = (float) qd.Price;
          STBQ_7 = (float) qd.Quantity;       }
      //
// Здесь добовляем то что нам нужно Stakan Tik Sell Price 8
      var asks_8 = security.GetSellQueue(8);
      if (asks_8 != null && asks_8.Count > 0) {
          var qd = asks_8[8];
          
          STSP_8 = (float) qd.Price;
          STSQ_8 = (float) qd.Quantity;       }
      
      var bids_8 = security.GetBuyQueue(8);
      if (bids_8 != null && bids_8.Count > 0) {
          var qd = bids_8[8];
       
          STBP_8 = (float) qd.Price;
          STBQ_8 = (float) qd.Quantity;       }
      //
 // Здесь добовляем то что нам нужно Stakan Tik Sell Price 9
      var asks_9 = security.GetSellQueue(9);
      if (asks_9 != null && asks_9.Count > 0) {
          var qd = asks_9[9];
          
          STSP_9 = (float) qd.Price;
          STSQ_9 = (float) qd.Quantity;       }
      
      var bids_9 = security.GetBuyQueue(9);
      if (bids_9 != null && bids_9.Count > 0) {
          var qd = bids_9[9];
       
          STBP_9 = (float) qd.Price;
          STBQ_9 = (float) qd.Quantity;       }
      //
// Здесь добовляем то что нам нужно Stakan Tik Sell Price 10
      var asks_10 = security.GetSellQueue(10);
      if (asks_10 != null && asks_10.Count > 0) {
          var qd = asks_10[10];
          
          STSP_10 = (float) qd.Price;
          STSQ_10 = (float) qd.Quantity;       }
      
      var bids_10 = security.GetBuyQueue(10);
      if (bids_10 != null && bids_10.Count > 0) {
          var qd = bids_10[10];
       
          STBP_10 = (float) qd.Price;
          STBQ_10 = (float) qd.Quantity;       }
      //
// Здесь добовляем то что нам нужно Stakan Tik Sell Price 11
      var asks_11 = security.GetSellQueue(11);
      if (asks_11 != null && asks_11.Count > 0) {
          var qd = asks_11[11];
          
          STSP_11 = (float) qd.Price;
          STSQ_11 = (float) qd.Quantity;       }
      
      var bids_11 = security.GetBuyQueue(11);
      if (bids_11 != null && bids_11.Count > 0) {
          var qd = bids_11[11];
       
          STBP_11 = (float) qd.Price;
          STBQ_11 = (float) qd.Quantity;       }
      //
// Здесь добовляем то что нам нужно Stakan Tik Sell Price 12
      var asks_12 = security.GetSellQueue(12);
      if (asks_12 != null && asks_12.Count > 0) {
          var qd = asks_12[12];
          
          STSP_12 = (float) qd.Price;
          STSQ_12 = (float) qd.Quantity;       }
      
      var bids_12 = security.GetBuyQueue(12);
      if (bids_12 != null && bids_12.Count > 0) {
          var qd = bids_12[12];
       
          STBP_12 = (float) qd.Price;
          STBQ_12 = (float) qd.Quantity;       }
      //
 // Здесь добовляем то что нам нужно Stakan Tik Sell Price 13
      var asks_13 = security.GetSellQueue(13);
      if (asks_13 != null && asks_13.Count > 0) {
          var qd = asks_13[13];
          
          STSP_13 = (float) qd.Price;
          STSQ_13 = (float) qd.Quantity;       }
      
      var bids_13 = security.GetBuyQueue(13);
      if (bids_13 != null && bids_13.Count > 0) {
          var qd = bids_13[13];
       
          STBP_13 = (float) qd.Price;
          STBQ_13 = (float) qd.Quantity;       }
      //
// Здесь добовляем то что нам нужно Stakan Tik Sell Price 14
      var asks_14 = security.GetSellQueue(14);
      if (asks_14 != null && asks_14.Count > 0) {
          var qd = asks_14[14];
          
          STSP_14 = (float) qd.Price;
          STSQ_14 = (float) qd.Quantity;       }
      
      var bids_14 = security.GetBuyQueue(14);
      if (bids_14 != null && bids_14.Count > 0) {
          var qd = bids_14[14];
       
          STBP_14 = (float) qd.Price;
          STBQ_14 = (float) qd.Quantity;       }
      //
// Здесь добовляем то что нам нужно Stakan Tik Sell Price 15
      var asks_15 = security.GetSellQueue(15);
      if (asks_15 != null && asks_15.Count > 0) {
          var qd = asks_15[15];
          
          STSP_15 = (float) qd.Price;
          STSQ_15 = (float) qd.Quantity;       }
      
      var bids_15 = security.GetBuyQueue(15);
      if (bids_15 != null && bids_15.Count > 0) {
          var qd = bids_15[15];
       
          STBP_15 = (float) qd.Price;
          STBQ_15 = (float) qd.Quantity;       }
      //
// Здесь добовляем то что нам нужно Stakan Tik Sell Price 16
      var asks_16 = security.GetSellQueue(16);
      if (asks_16 != null && asks_16.Count > 0) {
          var qd = asks_16[16];
          
          STSP_16 = (float) qd.Price;
          STSQ_16 = (float) qd.Quantity;       }
      
      var bids_16 = security.GetBuyQueue(16);
      if (bids_16 != null && bids_16.Count > 0) {
          var qd = bids_16[16];
       
          STBP_16 = (float) qd.Price;
          STBQ_16 = (float) qd.Quantity;       }
      //
 // Здесь добовляем то что нам нужно Stakan Tik Sell Price 17
      var asks_17 = security.GetSellQueue(17);
      if (asks_17 != null && asks_17.Count > 0) {
          var qd = asks_17[17];
          
          STSP_17 = (float) qd.Price;
          STSQ_17 = (float) qd.Quantity;       }
      
      var bids_17 = security.GetBuyQueue(17);
      if (bids_17 != null && bids_17.Count > 0) {
          var qd = bids_17[17];
       
          STBP_17 = (float) qd.Price;
          STBQ_17 = (float) qd.Quantity;       }
      //
// Здесь добовляем то что нам нужно Stakan Tik Sell Price 18
      var asks_18 = security.GetSellQueue(18);
      if (asks_18 != null && asks_18.Count > 0) {
          var qd = asks_18[18];
          
          STSP_18 = (float) qd.Price;
          STSQ_18 = (float) qd.Quantity;       }
      
      var bids_18 = security.GetBuyQueue(18);
      if (bids_18 != null && bids_18.Count > 0) {
          var qd = bids_18[18];
       
          STBP_18 = (float) qd.Price;
          STBQ_18 = (float) qd.Quantity;       }
      //
// Здесь добовляем то что нам нужно Stakan Tik Sell Price 19
      var asks_19 = security.GetSellQueue(19);
      if (asks_19 != null && asks_19.Count > 0) {
          var qd = asks_19[19];
          
          STSP_19 = (float) qd.Price;
          STSQ_19 = (float) qd.Quantity;       }
      
      var bids_19 = security.GetBuyQueue(19);
      if (bids_19 != null && bids_19.Count > 0) {
          var qd = bids_19[19];
       
          STBP_19 = (float) qd.Price;
          STBQ_19 = (float) qd.Quantity;       }
      //
// Здесь добовляем то что нам нужно Stakan Tik Sell Price 20
      var asks_20 = security.GetSellQueue(20);
      if (asks_20 != null && asks_20.Count > 0) {
          var qd = asks_20[20];
          
          STSP_20 = (float) qd.Price;
          STSQ_20 = (float) qd.Quantity;       }
      
      var bids_20 = security.GetBuyQueue(20);
      if (bids_20 != null && bids_20.Count > 0) {
          var qd = bids_20[20];
       
          STBP_20 = (float) qd.Price;
          STBQ_20 = (float) qd.Quantity;       }
      //
 // Здесь добовляем то что нам нужно Stakan Tik Sell Price 21
      var asks_21 = security.GetSellQueue(21);
      if (asks_21 != null && asks_21.Count > 0) {
          var qd = asks_21[21];
          
          STSP_21 = (float) qd.Price;
          STSQ_21 = (float) qd.Quantity;       }
      
      var bids_21 = security.GetBuyQueue(21);
      if (bids_21 != null && bids_21.Count > 0) {
          var qd = bids_21[21];
       
          STBP_21 = (float) qd.Price;
          STBQ_21 = (float) qd.Quantity;       }
      //
// Здесь добовляем то что нам нужно Stakan Tik Sell Price 22
      var asks_22 = security.GetSellQueue(22);
      if (asks_22 != null && asks_22.Count > 0) {
          var qd = asks_22[22];
          
          STSP_22 = (float) qd.Price;
          STSQ_22 = (float) qd.Quantity;       }
      
      var bids_22 = security.GetBuyQueue(22);
      if (bids_22 != null && bids_22.Count > 0) {
          var qd = bids_22[22];
       
          STBP_22 = (float) qd.Price;
          STBQ_22 = (float) qd.Quantity;       }
      //
// Здесь добовляем то что нам нужно Stakan Tik Sell Price 23
      var asks_23 = security.GetSellQueue(23);
      if (asks_23 != null && asks_23.Count > 0) {
          var qd = asks_23[23];
          
          STSP_23 = (float) qd.Price;
          STSQ_23 = (float) qd.Quantity;       }
      
      var bids_23 = security.GetBuyQueue(23);
      if (bids_23 != null && bids_23.Count > 0) {
          var qd = bids_23[23];
       
          STBP_23 = (float) qd.Price;
          STBQ_23 = (float) qd.Quantity;       }
      //
// Здесь добовляем то что нам нужно Stakan Tik Sell Price 24
      var asks_24 = security.GetSellQueue(24);
      if (asks_24 != null && asks_24.Count > 0) {
          var qd = asks_24[24];
          
          STSP_24 = (float) qd.Price;
          STSQ_24 = (float) qd.Quantity;       }
      
      var bids_24 = security.GetBuyQueue(24);
      if (bids_24 != null && bids_24.Count > 0) {
          var qd = bids_24[24];
       
          STBP_24 = (float) qd.Price;
          STBQ_24 = (float) qd.Quantity;       }
      //
 // Здесь добовляем то что нам нужно Stakan Tik Sell Price 25
      var asks_25 = security.GetSellQueue(25);
      if (asks_25 != null && asks_25.Count > 0) {
          var qd = asks_25[25];
          
          STSP_25 = (float) qd.Price;
          STSQ_25 = (float) qd.Quantity;       }
      
      var bids_25 = security.GetBuyQueue(25);
      if (bids_25 != null && bids_25.Count > 0) {
          var qd = bids_25[25];
       
          STBP_25 = (float) qd.Price;
          STBQ_25 = (float) qd.Quantity;       }
      //
// Здесь добовляем то что нам нужно Stakan Tik Sell Price 26
      var asks_26 = security.GetSellQueue(26);
      if (asks_26 != null && asks_26.Count > 0) {
          var qd = asks_26[26];
          
          STSP_26 = (float) qd.Price;
          STSQ_26 = (float) qd.Quantity;       }
      
      var bids_26 = security.GetBuyQueue(26);
      if (bids_26 != null && bids_26.Count > 0) {
          var qd = bids_26[26];
       
          STBP_26 = (float) qd.Price;
          STBQ_26 = (float) qd.Quantity;       }
      //
// Здесь добовляем то что нам нужно Stakan Tik Sell Price 27
      var asks_27 = security.GetSellQueue(27);
      if (asks_27 != null && asks_27.Count > 0) {
          var qd = asks_27[27];
          
          STSP_27 = (float) qd.Price;
          STSQ_27 = (float) qd.Quantity;       }
      
      var bids_27 = security.GetBuyQueue(27);
      if (bids_27 != null && bids_27.Count > 0) {
          var qd = bids_27[27];
       
          STBP_27 = (float) qd.Price;
          STBQ_27 = (float) qd.Quantity;       }
      //
// Здесь добовляем то что нам нужно Stakan Tik Sell Price 28
      var asks_28 = security.GetSellQueue(28);
      if (asks_28 != null && asks_28.Count > 0) {
          var qd = asks_28[28];
          
          STSP_28 = (float) qd.Price;
          STSQ_28 = (float) qd.Quantity;       }
      
      var bids_28 = security.GetBuyQueue(28);
      if (bids_28 != null && bids_28.Count > 0) {
          var qd = bids_28[28];
       
          STBP_28 = (float) qd.Price;
          STBQ_28 = (float) qd.Quantity;       }
      //
 // Здесь добовляем то что нам нужно Stakan Tik Sell Price 29
      var asks_29 = security.GetSellQueue(29);
      if (asks_29 != null && asks_29.Count > 0) {
          var qd = asks_29[29];
          
          STSP_29 = (float) qd.Price;
          STSQ_29 = (float) qd.Quantity;       }
      
      var bids_29 = security.GetBuyQueue(29);
      if (bids_29 != null && bids_29.Count > 0) {
          var qd = bids_29[29];
       
          STBP_29 = (float) qd.Price;
          STBQ_29 = (float) qd.Quantity;       }
      //
// Здесь добовляем то что нам нужно Stakan Tik Sell Price 30
      var asks_30 = security.GetSellQueue(30);
      if (asks_30 != null && asks_30.Count > 0) {
          var qd = asks_30[30];
          
          STSP_30 = (float) qd.Price;
          STSQ_30 = (float) qd.Quantity;       }
      
      var bids_30 = security.GetBuyQueue(30);
      if (bids_30 != null && bids_30.Count > 0) {
          var qd = bids_30[30];
       
          STBP_30 = (float) qd.Price;
          STBQ_30 = (float) qd.Quantity;       }
      //
// Здесь добовляем то что нам нужно Stakan Tik Sell Price 31
      var asks_31 = security.GetSellQueue(31);
      if (asks_31 != null && asks_31.Count > 0) {
          var qd = asks_31[31];
          
          STSP_31 = (float) qd.Price;
          STSQ_31 = (float) qd.Quantity;       }
      
      var bids_31 = security.GetBuyQueue(31);
      if (bids_31 != null && bids_31.Count > 0) {
          var qd = bids_31[31];
       
          STBP_31 = (float) qd.Price;
          STBQ_31 = (float) qd.Quantity;       }
      //
// Здесь добовляем то что нам нужно Stakan Tik Sell Price 32
      var asks_32 = security.GetSellQueue(32);
      if (asks_32 != null && asks_32.Count > 0) {
          var qd = asks_32[32];
          
          STSP_32 = (float) qd.Price;
          STSQ_32 = (float) qd.Quantity;       }
      
      var bids_32 = security.GetBuyQueue(32);
      if (bids_32 != null && bids_32.Count > 0) {
          var qd = bids_32[32];
       
          STBP_32 = (float) qd.Price;
          STBQ_32 = (float) qd.Quantity;       }
      //
 // Здесь добовляем то что нам нужно Stakan Tik Sell Price 33
      var asks_33 = security.GetSellQueue(33);
      if (asks_33 != null && asks_33.Count > 0) {
          var qd = asks_33[33];
          
          STSP_33 = (float) qd.Price;
          STSQ_33 = (float) qd.Quantity;       }
      
      var bids_33 = security.GetBuyQueue(33);
      if (bids_33 != null && bids_33.Count > 0) {
          var qd = bids_33[33];
       
          STBP_33 = (float) qd.Price;
          STBQ_33 = (float) qd.Quantity;       }
      //
// Здесь добовляем то что нам нужно Stakan Tik Sell Price 34
      var asks_34 = security.GetSellQueue(34);
      if (asks_34 != null && asks_34.Count > 0) {
          var qd = asks_34[34];
          
          STSP_34 = (float) qd.Price;
          STSQ_34 = (float) qd.Quantity;       }
      
      var bids_34 = security.GetBuyQueue(34);
      if (bids_34 != null && bids_34.Count > 0) {
          var qd = bids_34[34];
       
          STBP_34 = (float) qd.Price;
          STBQ_34 = (float) qd.Quantity;       }
      //
// Здесь добовляем то что нам нужно Stakan Tik Sell Price 35
      var asks_35 = security.GetSellQueue(35);
      if (asks_35 != null && asks_35.Count > 0) {
          var qd = asks_35[35];
          
          STSP_35 = (float) qd.Price;
          STSQ_35 = (float) qd.Quantity;       }
      
      var bids_35 = security.GetBuyQueue(35);
      if (bids_35 != null && bids_35.Count > 0) {
          var qd = bids_35[35];
       
          STBP_35 = (float) qd.Price;
          STBQ_35 = (float) qd.Quantity;       }
      //
// Здесь добовляем то что нам нужно Stakan Tik Sell Price 36
      var asks_36 = security.GetSellQueue(36);
      if (asks_36 != null && asks_36.Count > 0) {
          var qd = asks_36[36];
          
          STSP_36 = (float) qd.Price;
          STSQ_36 = (float) qd.Quantity;       }
      
      var bids_36 = security.GetBuyQueue(36);
      if (bids_36 != null && bids_36.Count > 0) {
          var qd = bids_36[36];
       
          STBP_36 = (float) qd.Price;
          STBQ_36 = (float) qd.Quantity;       }
      //
 // Здесь добовляем то что нам нужно Stakan Tik Sell Price 37
      var asks_37 = security.GetSellQueue(37);
      if (asks_37 != null && asks_37.Count > 0) {
          var qd = asks_37[37];
          
          STSP_37 = (float) qd.Price;
          STSQ_37 = (float) qd.Quantity;       }
      
      var bids_37 = security.GetBuyQueue(37);
      if (bids_37 != null && bids_37.Count > 0) {
          var qd = bids_37[37];
       
          STBP_37 = (float) qd.Price;
          STBQ_37 = (float) qd.Quantity;       }
      //
// Здесь добовляем то что нам нужно Stakan Tik Sell Price 38
      var asks_38 = security.GetSellQueue(38);
      if (asks_38 != null && asks_38.Count > 0) {
          var qd = asks_38[38];
          
          STSP_38 = (float) qd.Price;
          STSQ_38 = (float) qd.Quantity;       }
      
      var bids_38 = security.GetBuyQueue(38);
      if (bids_38 != null && bids_38.Count > 0) {
          var qd = bids_38[38];
       
          STBP_38 = (float) qd.Price;
          STBQ_38 = (float) qd.Quantity;       }
      //
// Здесь добовляем то что нам нужно Stakan Tik Sell Price 39
      var asks_39 = security.GetSellQueue(39);
      if (asks_39 != null && asks_39.Count > 0) {
          var qd = asks_39[39];
          
          STSP_39 = (float) qd.Price;
          STSQ_39 = (float) qd.Quantity;       }
      
      var bids_39 = security.GetBuyQueue(39);
      if (bids_39 != null && bids_39.Count > 0) {
          var qd = bids_39[39];
       
          STBP_39 = (float) qd.Price;
          STBQ_39 = (float) qd.Quantity;       }
      //
// Здесь добовляем то что нам нужно Stakan Tik Sell Price 40
      var asks_40 = security.GetSellQueue(40);
      if (asks_40 != null && asks_40.Count > 0) {
          var qd = asks_40[40];
          
          STSP_40 = (float) qd.Price;
          STSQ_40 = (float) qd.Quantity;       }
      
      var bids_40 = security.GetBuyQueue(40);
      if (bids_40 != null && bids_40.Count > 0) {
          var qd = bids_40[40];
       
          STBP_40 = (float) qd.Price;
          STBQ_40 = (float) qd.Quantity;       }
      //
 // Здесь добовляем то что нам нужно Stakan Tik Sell Price 41
      var asks_41 = security.GetSellQueue(41);
      if (asks_41 != null && asks_41.Count > 0) {
          var qd = asks_41[41];
          
          STSP_41 = (float) qd.Price;
          STSQ_41 = (float) qd.Quantity;       }
      
      var bids_41 = security.GetBuyQueue(41);
      if (bids_41 != null && bids_41.Count > 0) {
          var qd = bids_41[41];
       
          STBP_41 = (float) qd.Price;
          STBQ_41 = (float) qd.Quantity;       }
      //
// Здесь добовляем то что нам нужно Stakan Tik Sell Price 42
      var asks_42 = security.GetSellQueue(42);
      if (asks_42 != null && asks_42.Count > 0) {
          var qd = asks_42[42];
          
          STSP_42 = (float) qd.Price;
          STSQ_42 = (float) qd.Quantity;       }
      
      var bids_42 = security.GetBuyQueue(42);
      if (bids_42 != null && bids_42.Count > 0) {
          var qd = bids_42[42];
       
          STBP_42 = (float) qd.Price;
          STBQ_42 = (float) qd.Quantity;       }
      //
// Здесь добовляем то что нам нужно Stakan Tik Sell Price 43
      var asks_43 = security.GetSellQueue(43);
      if (asks_43 != null && asks_43.Count > 0) {
          var qd = asks_43[43];
          
          STSP_43 = (float) qd.Price;
          STSQ_43 = (float) qd.Quantity;       }
      
      var bids_43 = security.GetBuyQueue(43);
      if (bids_43 != null && bids_43.Count > 0) {
          var qd = bids_43[43];
       
          STBP_43 = (float) qd.Price;
          STBQ_43 = (float) qd.Quantity;       }
      //
// Здесь добовляем то что нам нужно Stakan Tik Sell Price 44
      var asks_44 = security.GetSellQueue(44);
      if (asks_44 != null && asks_44.Count > 0) {
          var qd = asks_44[44];
          
          STSP_44 = (float) qd.Price;
          STSQ_44 = (float) qd.Quantity;       }
      
      var bids_44 = security.GetBuyQueue(44);
      if (bids_44 != null && bids_44.Count > 0) {
          var qd = bids_44[44];
       
          STBP_44 = (float) qd.Price;
          STBQ_44 = (float) qd.Quantity;       }
      //
 // Здесь добовляем то что нам нужно Stakan Tik Sell Price 45
      var asks_45 = security.GetSellQueue(45);
      if (asks_45 != null && asks_45.Count > 0) {
          var qd = asks_45[45];
          
          STSP_45 = (float) qd.Price;
          STSQ_45 = (float) qd.Quantity;       }
      
      var bids_45 = security.GetBuyQueue(45);
      if (bids_45 != null && bids_45.Count > 0) {
          var qd = bids_45[45];
       
          STBP_45 = (float) qd.Price;
          STBQ_45 = (float) qd.Quantity;       }
      //
// Здесь добовляем то что нам нужно Stakan Tik Sell Price 46
      var asks_46 = security.GetSellQueue(46);
      if (asks_46 != null && asks_46.Count > 0) {
          var qd = asks_46[46];
          
          STSP_46 = (float) qd.Price;
          STSQ_46 = (float) qd.Quantity;       }
      
      var bids_46 = security.GetBuyQueue(46);
      if (bids_46 != null && bids_46.Count > 0) {
          var qd = bids_46[46];
       
          STBP_46 = (float) qd.Price;
          STBQ_46 = (float) qd.Quantity;       }
      //
// Здесь добовляем то что нам нужно Stakan Tik Sell Price 47
      var asks_47 = security.GetSellQueue(47);
      if (asks_47 != null && asks_47.Count > 0) {
          var qd = asks_47[47];
          
          STSP_47 = (float) qd.Price;
          STSQ_47 = (float) qd.Quantity;       }
      
      var bids_47 = security.GetBuyQueue(47);
      if (bids_47 != null && bids_47.Count > 0) {
          var qd = bids_47[47];
       
          STBP_47 = (float) qd.Price;
          STBQ_47 = (float) qd.Quantity;       }
      //
// Здесь добовляем то что нам нужно Stakan Tik Sell Price 48
      var asks_48 = security.GetSellQueue(48);
      if (asks_48 != null && asks_48.Count > 0) {
          var qd = asks_48[48];
          
          STSP_48 = (float) qd.Price;
          STSQ_48 = (float) qd.Quantity;       }
      
      var bids_48 = security.GetBuyQueue(48);
      if (bids_48 != null && bids_48.Count > 0) {
          var qd = bids_48[48];
       
          STBP_48 = (float) qd.Price;
          STBQ_48 = (float) qd.Quantity;       }
      //
 // Здесь добовляем то что нам нужно Stakan Tik Sell Price 49
      var asks_49 = security.GetSellQueue(49);
      if (asks_49 != null && asks_49.Count > 0) {
          var qd = asks_49[49];
          
          STSP_49 = (float) qd.Price;
          STSQ_49 = (float) qd.Quantity;       }
      
      var bids_49 = security.GetBuyQueue(49);
      if (bids_49 != null && bids_49.Count > 0) {
          var qd = bids_49[49];
       
          STBP_49 = (float) qd.Price;
          STBQ_49 = (float) qd.Quantity;       }
      //
  
    }  
  } 
}
